/*
        Copyright (C) 2001,2002 Encore Software Ltd.
        All rights reserved.
        Encore Software Ltd.
        6F Leo Complex
        44 & 45 Residency Cross Road
        Bangalore 560025 (INDIA)
        Tel: +91-80-558 0405 / 558 8257, Fax: +91-80-558 7690
        Email:  vasuki@ncoretech.com
*/
/*************************************************************************/
/*
**  Filename:    t38.c
**
**  Description:
**    This file contains the functions which form the fax relay according
**    to T.38 recommondations.
**
**
**  Public Functions :
**-------------------
** FAXRELAY_ESL_T38OpenChannel
**    Initializes the relay data structure by clearing all the varialbles
**    and then setting them to their initial values.
**    All the pointers are intialized to their respective buffers.
**    State machine is set appropriately based on the mode of the call
**    [ emitting relay or receiving relay ] by setting the state machine
**    function pointer.
void FAXRELAY_ESL_T38_AlgEnter(tFAXRELAY_ESL_T38 *ptT38Params,
**    Packets received from the IP network through UART are analysed.
**    Messages if any sent by the data pump are processed.
**    The state machine is executed.
**    Packets if formed by the state machine are sent to the IP network.
**
**  Private Functions :
**--------------------
** WaitForCedPkt :
**    State machine is set to wait for CED packet from the IP network.
**    On reception of CED pkt, a message (mDP_ANSWER) is sent to Data pump
**    asking it to transmit answer tone to the remote fax machine.
**    State machine is reset to wait for response from DP
** WaitForCngPkt :
**    State machine is set to wait for CNG packet from the IP network.
**    On reception of CNG pkt, messages mDP_DIALDURATION and mDP_TONEDIAL are
**    sent to Data pump asking it to dial the number of the fax terminal.
**    State machine is reset to wait for tone end response from DP
** WaitForToneEndResp :
**    State machine is set to wait for response from DP (data pump)
**    indicating the end of dial procedure.
**    On reception of which, DP is set to originate mode by sending the
**    message mDP_ORIGINATE.
**    State machine is reset to wait for detection of answer tone from DP.
** WaitForConnMsg :
**    State machine is set to wait for response indicating detection of 
**    answer tone from DP.
**    The indicator packet CED, is packed to be sent to IP network.
**    State machine is reset to initialize DP to reciver in low speed (V.21)
**    HDLC mode.
** WaitForConnMsgAns :
**    State machine is set to wait for response indicating end of transmision 
**    of answer tone from DP to the fax terminal.
**    State machine is reset to initialize DP to reciver in low speed (V.21)
**    HDLC mode.
**    State machine is reset to wait for status flags for further settings.
** SetDpModeHdlcLSRx :
**    DP is initialized to reciver in low speed (V.21) HDLC mode which will
**    be the default mode of configuration.
**    State machine is reset to wait for status flags for further settings.
** WaitForSettingMode :
**    State machine is set to wait for status flags to be set either by DP
**    messages or by packets received from the IP network.
**    The state machine is configured according to the status flags.
**
** SetDpModeHdlcTx:
**    To set the DP mode to transmit in HDLC mode at a given speed.
** WaitForConnMsgToTxFlgs :
**    In high speed HDLC mode of operation, the relay waits for connect
**    message from DP and then the DP is configured to send preamble flags.
** TxPreamble :
**    Relay waits for iPreambleTimer amount of time for the DP to send 
**    preamble flags to the fax terminal, after which state machine is set
**    to check for availabilty of HDLC data from the IP network
** ChkHdlcDataAvl :
**    Availabilty of HDLC data packet from the IP network is checked.
**    A time out timer started.
**    When data is obtained, it is copied to spoofing buffer for further
**    processing.
** CopyHdlcDataToSpoofingBuf :
**    HDLC frames are byte reversed and copied into a buffer which will be
**    transfered to the DP on reception of HDLC frame end indicator packet.
**    Information of correctness of the HDLC frame is also sent to the DP
**    through messages ( FCSOK and FCSBAD)
** TxHdlcDataToDp :
**    The data bytes in the HDLC frame are put to the message parameter 
**    buffer and then updated to the queue by calling  a function.
** SendCRP :
**    The HDLC frame "CRP" is sent to the DP through a message, on reception
**    of which, the DP modulates it sends to the fax terminal
** SendMsgToHost :
**    Whenever a message is available, it is put in the InMessage Queue.
** SendTxLvlMsgToDp :
**    Send the Tx level message to the datapump. 
** SendTxRxMsgToDp :
**    Send the message to set the Tx / Rx mode to the datapump. 
** CopyDataToBeTx :
**    Copy data from the input packet to the buffer pointed to by DataBufPtr
** GetMsgFromHost :
**    Get messages from the host and put them in the message parameter
**    buffer. Read/analyse the message to set the relevant flags.
** ProcessMessages :
**    Analyse the message and set the relevant flags. 
** T38PktT30Ind :
**    Packetize the indicator field.
** ReorderUDPPkts :
**    To reorder the UDP packets received from the IP network    
** DePacketize :
**    To de-packetize the packets received from the IP network and alalyse
**    them by a function call 
** AnalyzeInputPacket : 
**    The input packet is analyzed. If it is a control packet some flags are
**    set, if it is a data packet the data is copied.
** ConfigDpInHSLTrnModeTx :
**    The data pump is configured in high speed Tx mode with long train,
**    in case of V.17 mode
** ConfigDpInHSSTrnModeTx :
**    The data pump is configured in high speed Tx mode with short train, 
**    in case of V.17 mode
** ConfigDpInHSModeTx :
**    The mode of data is chosen to be MSB to LSB. This is because the DP
**    works in little endian mode and the T.38 is big endian. 
**    Datapump is set to high speed transmision. The state fn ptr is set
**    to TxHSDataToDp.
** ConfigDpInHdlcModeTx :
**    The mode of data is chosen to be HDLC.
**    Then the datapump is set to Hdlc Tx with the given dRate.
** ConfigDpInHdlcModeRx :
**    The mode of data is chosen to be HDLC.
**    Then the datapump is set to Hdlc Rx with the given dRate.
** ConfigDpInHSLTrnModeRx :
**    The mode of data is chosen to be MSB to LSB.
**    The data pump is configured in high speed Rx mode with long train
** ConfigDpInHSSTrnModeRx :
**    The mode of data is chosen to be MSB to LSB.
**    The data pump is configured in high speed Rx mode with short train
** InitHighSpeedRx :
**    Initialize the variables used by T.38 to receive high speed data. 
** RecvTrn :
**    State machine waits till it receives Trn message from datapump
**    and then pack the Trn Indicator packet.
**    The state fn ptr is set to PackHSDataPkt.
** RecvLTCF :
**    State machine waits till it receives Trn message from datapump
**    and then pack the Trn Indicator packet.
**    The state fn ptr is set to PackHSDataPkt, since the TCF 
**    data is what will follow.
** RecvTCF :
**    State machine waits till it receives Trn message from datapump
**    and then pack the Trn Indicator packet.
**    The state fn ptr is set to PackHSDataPkt, since the TCF 
**    data is what will follow.
** WaitForEndOfHSData :
**    The state machine waits till no carrier message is received from the DP
**    Then the number of zeros in TCF is evaluated and the local response is set.
**    The DP and the relay are set to the default mode of Low speed reception.
** PackHdlcDataPkt :
**    Packetization of HDLC frames is done both in High speed and low speed 
**    modes.
** T38PktT30Data :
**    Mode of packet transfer (TCP or UDP ) is checked and the Packets are 
**    formed by reading the data from a buffer accrodingly
** ProcErrInHdlcDataRx :
**    This function is used to form the CRC status reporting packet.
** PackEndOfHdlcFrame(int TypeOfMsg);
**    A packet having information of End of HDLC frame, along with the 
**    correctness of the HDLC frame is formed.
** PackHdlcSigEnd :
**    A packet having information of end of signal is formed .
** PackHSDataPkt :
**    State machine waits for the minimum number of elements (data bytes)
**    to be received from the DP for packetization. SIGEND information is
**    packet on reception of No carrier message from DP.
** PackHSData :
**    Packs the high speed data packet for non ecm. 
** CallPktForRemainingHSData :
**    Packs the high speed data bytes when a no carrier occurs.
** InitHdlcTx :
**    Initialization of the variables required for Hdlc remodulation along
**    with resetting of data buffer pointers
** InitHdlcTxVar :
**    Initialization of the variables required for Hdlc remodulation
** ProcNSFFrameTx :
**    Editing the FIF fields.     
** ProcDCSFrameTx :
**    Processing the DCS frame to identify the data mode. The data pump
**    is set for sending the TCF in the indentified mode (rate)
** ProcDCSFrV27Tx :
**    Initializing the variables for remodulation with V27 info.
** ProcDCSFrV29Tx :
**    Initializing the variables for remodulation with V29 info.
** ProcDCSFrV17Tx :
**    Initializing the variables for remodulation with V17 info.
** ProcEndHSDataTx :
**    DP is configured to end fax transmision/reception.
**    State machine is set to wait for the response from DP stating that
**    it has ended fax TxRx.  
** SendEndMsgToDp :
**    A message is sent to DP indicating it to end fax transmision/reception.
** WaitForAckHSTx :
**    The state machine is set to wait for a response from the DP for ending
**    of faxTxRx. On reception of the response the DP is again configured
**    in default low speed reception mode.
** WaitForTimeOutHL :
**    State machine is set to wait for the turn around time 
** ProcEndHdlcDataTx
**    HDLC mode of data transmision is ended
** ProcEndHdlcDataTx1 :
**    State machine is set to wait for acknowledge from the DP, for the 
**    HDLC frame that    was sent to it by the relay.
**    On reception of which and based on whether the previous frame was
**    a final frame, the state machine is reset appropriately.
** WtForSndngFlgs :
**    The state machine is set to wait for the DP to transmit preamble flags 
** WaitForAckHdlcTx :
**    State machine is set to wait for acknowledge from the DP, on reception
**    of which the DP is configured for default mode of low speed reception
** SetRxModeHdlc :
**    Set up DP in HDLC receive mode 
** ChkTxdFrm :
**    The previous HDLC frame transmitted is checked for MCF (message 
**    conformation) or PPR (partial page request )
** SNPEcm :
**    If MPS (More Pages to Send) request was received by DP, the state
**    machine waits for MCF or PPR.
**    Other wise DP is configured to default mode of low speed reception.    
** WaitForSigEnd :
**    State machine is set to wait for SIG-END packet, on receiving which
**    it is reset to allow for a delay in turning aroung from low speed
**    to high speed mode or vise versa.
** WaitForTimeOutLH :
**    State machine is set to wait for the turn around time (from low
**    speed to high speed mode).
** TxHSDataToDp :
**    Data is sent to the DP for transmision in high speed mode
** SetDpModeHSTx :
**    Configure DP in high speed transmision mode
** HdlcData :
**    HDLC data is copied from the message buffers to the data buffer
**    after byte reversal.
** InitHdlcRx :
**    Initialization of the variables required for Hdlc demodulation along
**    with resetting of data buffer pointers
** InitHdlcRxVar :
**    Initialization of the variables required for Hdlc demodulation
** ProcHdlcHSDataRx :
**    Process high speed data in HDLC mode.
** ProcHdlcLSDataRx :
**    Process low speed data in HDLC mode.
**    The the information bytes in the HDLC frame received from the DP are read
**    and processed apporpriately.
** ProcDCSFrameRx :
**    Processing the DCS frame to identify the data mode. The data pump
**    is set for receiving the TCF in the indentified mode (rate)
** ProcDCSFrV27Rx :
**    Initializing the variables for demodulation with V27 info.
** ProcDCSFrV29Rx :
**    Initializing the variables for demodulation with V29 info.
** ProcDCSFrV17Rx :
**    Initializing the variables for demodulation with V17 info.
** ProcEndHdlcHSDataRx :
**    State machine is set to wait for no carrier message from DP,
**    after receiving which, the relay forms the SIG-END indicator packet
**    and sets the DP in default mode of low speed reception.
** ProcEndHdlcLSDataRx :
**    State machine is set to wait for no carrier message from DP,
**    after receiving which, the relay forms the SIG-END indicator packet.
**    If the previous frame received by DP is DCS, the relay configures
**    DP to receive training in high speed mode.
**    Other wise the DP is configured in default mode of low speed reception.
** WaitForSpoofPreamble :
**    State machine is set wait for the spoofing timer to elapse, after which
**    it is reset to spoof preamble
**    If any valid HDLC frames are received before the timer elapses, the DP mode
** SpoofPreamble :
**    DP is set to spoof premable flags to the fax terminal.
** ByteReverse :
**    This function is used to reverse the given byte.
** ProcNotLastLSFrameRx :
**    If the previous HDLC frame received by the DP is not a final frame,
**    state machine is set to pack the frames which follow.
** ClearStruct :
**    To clear structure a pointer to which and the length are given.
** WaitToSpoofHSImage
**    State machine is set to wait for image data (Trn Indicator pkt)
**    till the timer elapses. After which, high speed fill bits (zeroes) 
**    are spoofed.
**WaitForMCF_DCS
**    State machine is set to wait for MCF or DCS data till the timer    elapses
**    If frames are not received, no carrier message waited from DP.
**WaitForNoCarMsg
**    State machine is set to wait for no carrier message from DP
**    After getting which, the frame queued by T38 is sent to DP for
**    transmision
**WaitForCFR_FTT
**    State machine is set to wait for response for training (CFR or FTT)
**    If response is not obtained before the timer elapses, set the state
**    machine to wait for no carrier message from DP, then the queued
**    frame (CFR or FTT) is sent to DP.
**SpoofTCF
**    State machine is set to spoof TCF data, data pump is set to transmit
**    in high speed mode.
**SendFTTPkt
**    FTT ( Failure To Train) frame is packetized and sent to the network
**SendDCN
**    This function is used to spoof the HDLC frame "DCN"
** The bytes in DCN frame are sent as messages to DP which inturn applies the
** low speed modulation technique (V.21) and send to the fax terminal.
**
**
**
**  Modification History:
**    Date                    By                  Description    
**  -------------      -----------------  ------------------
** 3/November/2000    Dhruva Kumar N      Created
**
** 2/January/2001     Dhruva Kumar N      Implementation changed to
**                                        use single port memory
**
** 30/January/2001    Dhruva Kumar N      Implementation of high speed spoofing
**
** 7/February/2001    Dhruva Kumar N      Spoofing in low speed mode redesigned
**
** 12/February/2001   Dhruva Kumar N      Implementation of handling of packet
**                                        losses and out of order packets.
**
** 7/May/2001         Dhruva Kumar N      Declaration of global Channel
**                                        structure pointer removed.
**
** 8/January/2003     Dhruva Kumar N      Packet structure modified to
**										  inter-operate with CISCO
*/
/***************************************************************************/


/*-------------------------------------------------------------------------
 INCLUDE FILES
--------------------------------------------------------------------------*/

#include <oswrapper.h>  
#if FAX

#include <ipvoice/include/t38.h>
#include <ipvoice/cli/voiceldef.h>

//#define CFRTESTING

extern FaxOpt_t gFaxParams;

//UINT32 HSspoofCntr = 0;
//UINT32 HSspoofCntrFlg = 0;


extern INT32 gFaxSessCreated;

extern INT16 SMM_Sharp_FirstFlg[MAX_VOICE_CHANNELS];
extern INT16 SMM_CEDCntFlag[MAX_VOICE_CHANNELS];
INT16 PageCount = 0;
extern INT16 iTxSeqNo[MAX_VOICE_CHANNELS];
extern INT16 iTxSeqPrevNo[MAX_VOICE_CHANNELS];

extern INT16 SMM_CRCErrorFlag[MAX_VOICE_CHANNELS];

extern INT16 CFRRxCntr[MAX_VOICE_CHANNELS];
extern INT16 SendCRPtoDP[MAX_VOICE_CHANNELS];

/*-----------------------------------------------------------------------
PRIVATE FUNCTION PROTOTYPES
------------------------------------------------------------------------*/

void WaitToSendCRPPkt(tFAXRELAY_ESL_T38 *); //dhruva 310804
void SendCRPPkt(tFAXRELAY_ESL_T38 *); //dhruva 010904

/*Function  declaration: Scope limited to this file*/
void SendCNGPkt(tFAXRELAY_ESL_T38 *);
void SendCRP(tFAXRELAY_ESL_T38 *);
void WaitForSpoofPreamble(tFAXRELAY_ESL_T38 *);
void SpoofPreamble(tFAXRELAY_ESL_T38 *);
void CopyHdlcDataToSpoofingBuf(tFAXRELAY_ESL_T38 *);
void AnalyzeInputPacket(tFAXRELAY_ESL_T38 *); 
void GetMsgFromDp(tFAXRELAY_ESL_T38 *);
void ConfigDpInHdlcModeRx(tFAXRELAY_ESL_T38 *,INT16 dRate);
void WaitForSettingMode(tFAXRELAY_ESL_T38 *);
void WaitForAckHSTx(tFAXRELAY_ESL_T38 *);
void WaitForTimeOutHL(tFAXRELAY_ESL_T38 *);
void WtForSndngFlgs(tFAXRELAY_ESL_T38 *);
void WaitForAckHdlcTx(tFAXRELAY_ESL_T38 *);
void WaitForTimeOutLH(tFAXRELAY_ESL_T38 *);
void WaitForSigEnd(tFAXRELAY_ESL_T38 *);
void SetDpModeHdlcLSRx(tFAXRELAY_ESL_T38 *);
void SetDpModeHdlcTx(tFAXRELAY_ESL_T38 *,INT16 dRate);
void SetRxModeHdlc(tFAXRELAY_ESL_T38 *);
void SetDpModeHSTx(tFAXRELAY_ESL_T38 *);
void ConfigDpInHdlcModeTx(tFAXRELAY_ESL_T38 *,INT16 dRate);
void TxPreamble(tFAXRELAY_ESL_T38 *);
void ChkHdlcDataAvl(tFAXRELAY_ESL_T38 *);
void TxHdlcDataToDp(tFAXRELAY_ESL_T38 *);
void TxHSDataToDp(tFAXRELAY_ESL_T38 *);
void SendMsgToHost(tFAXRELAY_ESL_T38 *);
void ProcHdlcLSDataRx(tFAXRELAY_ESL_T38 *);
void ProcHdlcHSDataRx(tFAXRELAY_ESL_T38 *);
void ProcEndHdlcHSDataRx(tFAXRELAY_ESL_T38 *);
void ProcEndHdlcLSDataRx(tFAXRELAY_ESL_T38 *);
void ProcEndHdlcDataTx(tFAXRELAY_ESL_T38 *);
void ProcEndHdlcDataTx1(tFAXRELAY_ESL_T38 *);
void ProcDCSFrameTx(tFAXRELAY_ESL_T38 *,INT16);
void ProcDCSFrV17Tx(tFAXRELAY_ESL_T38 *,INT16 mode);
void ProcDCSFrV29Tx(tFAXRELAY_ESL_T38 *,INT16 mode);
void ProcDCSFrV27Tx(tFAXRELAY_ESL_T38 *,INT16 mode);
void ProcNSFFrameTx(tFAXRELAY_ESL_T38 *);
void ProcEndHSDataTx(tFAXRELAY_ESL_T38 *);
void SendTxLvlMsgToDp(tFAXRELAY_ESL_T38 *,INT16 Id,INT16 TxLvl);
void SendTxRxMsgToDp(tFAXRELAY_ESL_T38 *,INT16 Id,INT16 Param);
void SendEndMsgToDp(tFAXRELAY_ESL_T38 *);
void CopyDataToBeTx(tFAXRELAY_ESL_T38 *,UCHAR8 *);
void DePacketize(tFAXRELAY_ESL_T38 *, UCHAR8 *,INT16,INT16);
void ConfigDpInHSLTrnModeTx(tFAXRELAY_ESL_T38 *);
void ConfigDpInHSSTrnModeTx(tFAXRELAY_ESL_T38 *);
void ConfigDpInHSModeTx(tFAXRELAY_ESL_T38 *,INT16 dRate);
void ConfigDpInHSLTrnModeRx(tFAXRELAY_ESL_T38 *);
void ConfigDpInHSSTrnModeRx(tFAXRELAY_ESL_T38 *);
void InitHighSpeedRx(tFAXRELAY_ESL_T38 *);
void InitHdlcRx(tFAXRELAY_ESL_T38 *);
void InitHdlcVarRx(tFAXRELAY_ESL_T38 *);
void InitHdlcTx(tFAXRELAY_ESL_T38 *);
void InitHdlcTxVar(tFAXRELAY_ESL_T38 *);
void SNPEcm(tFAXRELAY_ESL_T38 *);
void ChkTxdFrm(tFAXRELAY_ESL_T38 *);
void RecvTrn(tFAXRELAY_ESL_T38 *);
void RecvLTCF(tFAXRELAY_ESL_T38 *);
void RecvTCF(tFAXRELAY_ESL_T38 *);
void PackHdlcDataPkt(tFAXRELAY_ESL_T38 *);
void T38PktT30Ind(tFAXRELAY_ESL_T38 *,UCHAR8);
void T38PktT30Data(tFAXRELAY_ESL_T38 *,UCHAR8,UCHAR8,UCHAR8,UCHAR8,INT16);
void ProcessMessages(tFAXRELAY_ESL_T38 *);
void GetMsgFromHost(tFAXRELAY_ESL_T38 *,tMSGQSTRUCT *);
void IpToPkt(tFAXRELAY_ESL_T38 *,INT16 TypeOfMsg);
void ProcErrInHdlcDataRx(tFAXRELAY_ESL_T38 *);
void PackHdlcSigEnd(tFAXRELAY_ESL_T38 *);
void ProcDCSFrameRx(tFAXRELAY_ESL_T38 *);
void ProcDCSFrV17Rx(tFAXRELAY_ESL_T38 *,INT16 mode);
void ProcDCSFrV29Rx(tFAXRELAY_ESL_T38 *,INT16 mode);
void ProcDCSFrV27Rx(tFAXRELAY_ESL_T38 *,INT16 mode);
void ProcNotLastLSFrameRx(tFAXRELAY_ESL_T38 *);
void PackHSDataPkt(tFAXRELAY_ESL_T38 *);
void PackHSData(tFAXRELAY_ESL_T38 *);
//void PackEndOfHdlcFrame(tFAXRELAY_ESL_T38 *,INT16 TypeOfMsg);
void CallPktForRemainingHSData(tFAXRELAY_ESL_T38 *);
void HdlcData(tFAXRELAY_ESL_T38 *);
void HighSpeedData(tFAXRELAY_ESL_T38 *);
void WaitForConnMsgToTxFlgs(tFAXRELAY_ESL_T38 *);
void WaitForEndOfHSData(tFAXRELAY_ESL_T38 *);
#ifdef TCP_MODE
void ProcCFRFrameTx(tFAXRELAY_ESL_T38 *);
INT16 GetTCPIFPPacket(tFAXRELAY_ESL_T38 *,UCHAR8 *,INT16 );
#endif
void ClearStruct(UCHAR8 *Ptr,UINT16 NumBytes);
UCHAR8 ByteReverse(UCHAR8);
void WaitToSpoofHSImage(tFAXRELAY_ESL_T38 *);
void WaitForMCF_DCS(tFAXRELAY_ESL_T38 *);
void WaitForNoCarMsg(tFAXRELAY_ESL_T38 *);
void WaitForCFR_FTT(tFAXRELAY_ESL_T38 *);
void SpoofTCF(tFAXRELAY_ESL_T38 *);
void SendFTTPkt(tFAXRELAY_ESL_T38 *);
void SendDCN(tFAXRELAY_ESL_T38 *);
void analyze_field_data(tFAXRELAY_ESL_T38 *, UCHAR8 *);

#ifdef UDP_MODE
INT16 ReorderUDPPkts(tFAXRELAY_ESL_T38 *,UCHAR8 *,INT16);
INT16 ReorderCiscoUDPPkts(tFAXRELAY_ESL_T38 *,UCHAR8 *,INT16);
INT16 FindPktInBUF(tFAXRELAY_ESL_T38 *,INT16);
void GetSecondaryPkt(tFAXRELAY_ESL_T38 *,UCHAR8 *,INT16);
void UpDatePrevPktBuf(tFAXRELAY_ESL_T38 *,UINT16,UINT16);
UINT16 AddRedundantPac(tFAXRELAY_ESL_T38 *, UCHAR8 *,UCHAR8 *);
#endif
void ResetTCFParams(tFAXRELAY_ESL_T38 *);
void CopyChars(UCHAR8 *,UCHAR8 *,UINT16);
void ClearCRCBuf(tFAXRELAY_ESL_T38 *);
void ClearSizeBuf(tFAXRELAY_ESL_T38 *);
void AnalyzeDSImage(tFAXRELAY_ESL_T38 *ptT38Params);

void WaitForEndOfTCF(tFAXRELAY_ESL_T38 *);
void WaitForPreamble(tFAXRELAY_ESL_T38 *);


void WaitForSigEndTx1(tFAXRELAY_ESL_T38 *ptT38Params);

INT16 GetEarliestSavedPkt(tFAXRELAY_ESL_T38 *,INT16 );
void SendNoSigForHS(tFAXRELAY_ESL_T38 *);

/*This at present is declared for testing the iprelay */
 
void WaitForCngPkt(tFAXRELAY_ESL_T38 *);
void WaitForCedPkt(tFAXRELAY_ESL_T38 *);
void WaitForConnMsg(tFAXRELAY_ESL_T38 *);
void WaitForConnMsgAns(tFAXRELAY_ESL_T38 *);

extern INT16 debug_fax;

#if 1 
extern UINT16  maxFrames[] ; 
#endif
/*-------------------------------------------------------------------------

Function Name:  FAXRELAY_ESL_T38_GetMemInfo

Description :
	This function is called by the host application to find out the 
	size of memory required for one channel of T.38

Inputs: None

Output: Size of per channel data memory

---------------------------------------------------------------------------*/

UINT16 FAXRELAY_ESL_T38_GetMemInfo(void)
{
	return(sizeof(tFAXRELAY_ESL_T38));
}



/*-------------------------------------------------------------------------

Function Name:  FAXRELAY_ESL_T38OpenChannel

Description :
This function is called by the API.
The memory area used by the channel is cleared, and then initialzed.

Inputs: 
1. ptT38Params  : per-channel T.38 static Pointer
2. ptAlgConfig  : pointer to T.38 config structure

Output: Error code

---------------------------------------------------------------------------*/

UINT16 FAXRELAY_ESL_T38OpenChannel(tFAXRELAY_ESL_T38 *ptT38Params,
                                 tFAXRELAY_ESL_config *ptAlgConfig)
{

    INT16 i;
	UINT16 status;


   PageCount = 0;

/* clear all the variabels in the relay data structure */
    //ClearStruct( (UCHAR8 *)ptT38Params,sizeof(tFAXRELAY_ESL_T38) );
    //bzero((void *)ptT38Params,sizeof(tFAXRELAY_ESL_T38)); //Raju on 7th Aug
    bzero((tFAXRELAY_ESL_T38 *)ptT38Params,sizeof(tFAXRELAY_ESL_T38));

    ptT38Params->JitterBufLen = ptAlgConfig->wSizeOfJitterBuf;

//Initialize pointer to JitterBuffer
	ptAlgConfig->JitterBufPtr = &ptT38Params->JitterBuffer[0];

//call mode 1 => ORIGINATE, 2 => ANSWER
    ptT38Params->iCallMode = ptAlgConfig->iCallMode;

#ifdef UDP_MODE
//In case of UDP
    ptT38Params->iLSRedundancy = ptAlgConfig->LSRedundancy;
	ptT38Params->iHSRedundancy = ptAlgConfig->HSRedundancy;
#ifdef DEF_DEBUG
	if(ptT38Params->EnableDebug)
		printf("\nLS_redundancy = %d, HS_redundancy = %d",
		ptT38Params->iLSRedundancy,ptT38Params->iHSRedundancy);
#endif
#endif

//Frame length
    ptT38Params->iFrameLength = ptAlgConfig->wFrameLength;

//To store depacketised data.
    ptT38Params->iSizeOfDataBuf = BUFLEN;

//Used for spoofing
    ptT38Params->iSizeOfSpoofingBuf = SPOOFBUFLEN;

// memory required by one channel od relay
    ptT38Params->iChannelMemory = sizeof(tFAXRELAY_ESL_T38);

//Initializing the pointers

    ptT38Params->piReOrderBufPtr = ptT38Params->aiReOrderBuffer;
    ptT38Params->iRxSeqNo = 0;
    ptT38Params->iPrevPktRead = -1;

    for(i=0; i < REORDERBUFLEN; i+= RX_IFP_MAXSIZE)
        ptT38Params->aiReOrderBuffer[i] = ptT38Params->iPrevPktRead;

 
    ptT38Params->piDataBufPtr = &ptT38Params->aiDataBuffer[0];

    ptT38Params->piIpPktPtr = &ptT38Params->aiIpPktBuf[0];

    ptT38Params->piSpoofingBufPtr = &ptT38Params->aiSpoofBuffer[0];
    ptT38Params->piWrSpoofingBufPtr = 
    ptT38Params->piRdSpoofingBufPtr = 
    ptT38Params->piSpoofingBufPtr;

    ptT38Params->piFrameDataPtr = 
    ptT38Params->piReadDataPtr = 
    ptT38Params->piImageDataPtr = ptT38Params->piDataBufPtr;

//Set the pointer to message queue structures for T.38
    ptT38Params->ptT38InMsgQStr = ptAlgConfig->ptT38InMsgQStr;
    ptT38Params->ptT38OutMsgQStr = ptAlgConfig->ptT38OutMsgQStr;

	
	ptT38Params->pTempImageW = &ptT38Params->TempImageBuf[0];
	ptT38Params->pTempImageR = &ptT38Params->TempImageBuf[0];
	ptT38Params->iSizeOfTempImageBuf = TEMPBUFSIZE;	//180*6


/* If it is ORIGINATE send CNG packet, if it is ANSWER mode wait for CED 
   packet */

    //SendTxLvlMsgToDp(ptT38Params,mDP_SET_TXLEVEL,ptT38Params->iTxLevel);

    if(ptT38Params->iCallMode==ORIGINATEMODE)
        ptT38Params->fpStateFnPtr = (fpPtrIntFn)(SendCNGPkt);
    else
        ptT38Params->fpStateFnPtr = (fpPtrIntFn)(WaitForCngPkt);


	status = 0;

	if(ptAlgConfig->iCallMode < 1 || ptAlgConfig->iCallMode >2)
		status = 1;
	if(ptAlgConfig->wFrameLength < FL_5MS ||
		ptAlgConfig->wFrameLength > FL_20MS)
		status |= 2;
	if(ptAlgConfig->wSizeOfJitterBuf < 200)
		status |= 4;
	if(ptAlgConfig->LSRedundancy < 0 || ptAlgConfig->LSRedundancy > 6)
		status |= 8;
	if(ptAlgConfig->HSRedundancy < 0 || ptAlgConfig->HSRedundancy > 2)
		status |= 16;
	
	return(status);
}




/*-------------------------------------------------------------------------

Function Name:  FAXRELAY_ESL_T38_AlgEnter


Description :
    This function is called for every frame (10ms) by the API and is 
 linked with the datapump.
 The following functionality is performed in this:
 1. It gets messages from datapump. 
 2. The packets in the receive packet queue are reordered and analyzed.
 3. The state function is executed.
 4. If there is a packet to be sent to the network, it is sent by calling 
    PutPacket() function.

Inputs: 
1. ptT38Params  : per-channel T.38 static Pointer
2. pOpPktBufPtr : pointer to the buffer to which the packets that are to be
                  sent to IP network will be written into.
3. pInPktReadPtr :Pointer to the buffer which contains packets received from
    IP network.
4. ValidIpPkt :   When non zero, indicates the presence of a valid IFP packet
                  from the IP network.
5. SizeOfRxPkt : Size of packet recived

Return Value:   None

---------------------------------------------------------------------------*/

void FAXRELAY_ESL_T38_AlgEnter(tFAXRELAY_ESL_T38 *ptT38Params,
                               UCHAR8 *pOpPktBufPtr,
                               UCHAR8 *pInPktReadPtr,
                               INT16 ValidIpPkt,
                               INT16 SizeOfRxPkt)
{

#if 1
UCHAR8 *Ptr1;
#endif
    INT16 OpPktPresent = 0;


    ptT38Params->paiOpPktPtr = pOpPktBufPtr;

/* Reads messages in the Out message queue given by DP and processes them */
    GetMsgFromHost(ptT38Params,ptT38Params->ptT38InMsgQStr);

/* Depacketization of the packtes received from the IP network is done */
    DePacketize(ptT38Params,pInPktReadPtr,ValidIpPkt,SizeOfRxPkt);

/* Execute the state machine */
    ptT38Params->fpStateFnPtr(ptT38Params);

/* The state machine sends messages to DP and generates packets that are to
sent to the network */
    if(ptT38Params->iOpPktValid)
    {

#if 0
        Ptr1 = pOpPktBufPtr; //IFPParams[portNum].USPktBuf[0];
        iTxSeqNo[ptT38Params->iPortNo] = *Ptr1++;
        iTxSeqNo[ptT38Params->iPortNo] <<=8;
        iTxSeqNo[ptT38Params->iPortNo] |= (*Ptr1++ & 0xff);

        if(iTxSeqNo[ptT38Params->iPortNo] != iTxSeqPrevNo[ptT38Params->iPortNo] + 1)
           printf("\r\n USBuffer TxSeqNo [%x],port[%d]", iTxSeqNo[ptT38Params->iPortNo],ptT38Params->iPortNo);

        iTxSeqPrevNo[ptT38Params->iPortNo] = iTxSeqNo[ptT38Params->iPortNo];
#endif
        ptT38Params->tMessage.iId = mT38_VALID_OPPKT;
        ptT38Params->tMessage.iLen=3;
	ptT38Params->tMessage.aiParams[0] = ptT38Params->iOpPktValid;
	ptT38Params->tMessage.aiParams[1] = ptT38Params->TxPayloadSize & 0xff;
	ptT38Params->tMessage.aiParams[2] = (ptT38Params->TxPayloadSize >> 8) & 0xff;
//function call to update the message queue
        SendMsgToHost(ptT38Params);
    }

    ptT38Params->iOpPktValid = 0;


    return;
}



/*-------------------------------------------------------------------------

Function Name: SendCNGPkt 

Description :
	Packetize CNG indicator packet and switch the state machine to wait
	for CED packet

Inputs: 
pt_FAXRELAY_ESL_info->pt_T38_ASP_static = relay channel pointer

Return Value: None

---------------------------------------------------------------------------*/

void SendCNGPkt(tFAXRELAY_ESL_T38 *ptT38Params)
{
	T38PktT30Ind(ptT38Params,CNG);
	if(++ptT38Params->iCurrentRedundancy >= ptT38Params->iLSRedundancy)
		ptT38Params->iCurrentRedundancy = ptT38Params->iLSRedundancy;

	ptT38Params->fpStateFnPtr = (fpPtrIntFn)(WaitForCedPkt);
	InitHdlcRx(ptT38Params);
}

/*-------------------------------------------------------------------------

Function Name: WaitForCedPkt 

Description :
At the calling end (ORIGINATE), after transmitting the CNG packet we wait for
the CED packet , which is a response sent by the receiving end(ANSWER).
After receiving CED packet, we set the datapump in ANSWER mode, and set the
state function pointer to WaitForConnMsgAns() function.

Inputs: 
pt_FAXRELAY_ESL_info->pt_T38_ASP_static = relay channel pointer

Return Value: None

---------------------------------------------------------------------------*/

void WaitForCedPkt(tFAXRELAY_ESL_T38 *ptT38Params){

    if(ptT38Params->iCed || ptT38Params->iPreambleTxFlg)
    {
        ptT38Params->iCed = 0;

     /* This is at the calling end. After receiving CED packet relay
      sets the DP in ANSWER mode */
#if 0
//Message ID mT38MSGTODP indicates host to forward this message to DP
        ptT38Params->tMessage.iId = mDP_ANSWER;
        ptT38Params->tMessage.iLen=0;         //length of the message
//function call to update the message queue
        SendMsgToHost(ptT38Params);
#endif

#if 1  // May 27
      if(SMM_CEDCntFlag[ptT38Params->iPortNo] == 0)
      {

#if 0
   if((gDownloadOpts[ptT38Params->iPortNo]!=0x11)&&(ptT38Params->iPortNo < NUM_VOICE_CHANNELS))
   {
      dspDownloadCodec(0x11,ptT38Params->iPortNo);
   }
#endif

        printf("\r\n dspipcSendmAnswerCmd  to DSP [%d]",ptT38Params->iPortNo);
        dspipcSendmAnswerCmd(ptT38Params->iPortNo);
      }
#endif
/* State machine is set to wait for response from DP for the above command */
        ptT38Params->fpStateFnPtr = (fpPtrIntFn)(WaitForConnMsgAns);
    }
    return;
}

/*-------------------------------------------------------------------------

Function Name: WaitForConnMsgAns 

Description :
At the calling end (ORIGINATE), after setting datapump in ANSWER mode, we wait
for a connect message from the datapump. After getting this message, we set the
datapump 's transmission level(TXLEVEL) to a user defined value, and state 
function pointer to SetDpModeHdlcLSRx() function.

Inputs: 
ptT38Params = relay channel pointer


Return Value: None


---------------------------------------------------------------------------*/

void WaitForConnMsgAns(tFAXRELAY_ESL_T38 *ptT38Params){

    if(ptT38Params->iConnected)
     {
        //printf("\r\nWaitForConnMsgAns : iConnected [1]");
        ptT38Params->iConnected = 0;
        //SendTxLvlMsgToDp(ptT38Params,mDP_SET_TXLEVEL,ptT38Params->iTxLevel);
        ptT38Params->fpStateFnPtr=(fpPtrIntFn)(SetDpModeHdlcLSRx);
     }
    return;
}

/*-------------------------------------------------------------------------

Function Name: WaitForCngPkt 

Description :
At the receiving end (ANSWER), we wait for a CNG packet from the ORIGINATOR.
After receiving this packet, we send the fax m/c number to be dialed (dial 
string) to the data pump and then set the state function pointer to 
WaitForToneEndResp() function.

Inputs: 
ptT38Params = relay channel pointer

Return Value: None

---------------------------------------------------------------------------*/

void WaitForCngPkt(tFAXRELAY_ESL_T38 *ptT38Params){

    INT16 i;

    if(ptT38Params->iCng)
    {
         ptT38Params->iCng = 0;

        ptT38Params->tMessage.iId = mT38_RX_CNGPKT;
        ptT38Params->tMessage.iLen = 0;
        SendMsgToHost(ptT38Params);
		
#if 0
/* Configure DP in originate mode */
//Message ID mT38MSGTODP indicates host to forward this message to DP
        ptT38Params->tMessage.iId = mDP_ORIGINATE;
        ptT38Params->tMessage.iLen = 0;
        SendMsgToHost(ptT38Params);
#endif
        if(ptT38Params->iCallMode == ANSWERMODE)
	{
            //dspipcSendmOriginateCmd(ptT38Params->iPortNo); 

/* State machine is set wait for response from DP indicating detection
   of answer tone from the fax terminal */

	    ptT38Params->fpStateFnPtr = (fpPtrIntFn)(WaitForConnMsg);
        }


    }

    if(ptT38Params->iCallMode == ORIGINATEMODE)
    {
       //printf("\r\nT38: Originate mode;wait for CED on portno:%d", ptT38Params->iPortNo);
       T38PktT30Ind(ptT38Params,CNG);        //To form CED packet
       ptT38Params->fpStateFnPtr = (fpPtrIntFn )(WaitForCedPkt);
    }
    return;
}


/*-------------------------------------------------------------------------

Function Name: WaitForConnMsg 

Description :
At the receiving end (ANSWER), after setting the datapump in ORIGINATE mode, 
we wait for connected message from the datapump (the datapump sends this 
message after it receives answer tone). After getting this message, we
transmit a CED packet and set the datapump transmission level to a user defined
value and the state function pointer to SetDpModeHdlcLSRx().

Inputs: 
ptT38Params = relay channel pointer

Return Value: None

---------------------------------------------------------------------------*/


void WaitForConnMsg(tFAXRELAY_ESL_T38 *ptT38Params){

    if(ptT38Params->iConnected)
     {
	ptT38Params->iConnected = 0;
	T38PktT30Ind(ptT38Params,CED);        //To form CED packet
	if(++ptT38Params->iCurrentRedundancy >= ptT38Params->iLSRedundancy)
		ptT38Params->iCurrentRedundancy = ptT38Params->iLSRedundancy;

	//SendTxLvlMsgToDp(ptT38Params,mDP_SET_TXLEVEL,ptT38Params->iTxLevel );

/* State machine is set to configure DP in low speed receive mode */
       ptT38Params->fpStateFnPtr= (fpPtrIntFn)( SetDpModeHdlcLSRx);

     }
    return;
}


/*-------------------------------------------------------------------------

Function Name: SetDpModeHdlcLSRx 

Description :
We configure the datapump in default low Speed (V21300) reception mode and wait
for the reception of a preamble packet/ training packet / preamble message from 
datapump for configuring the datapump mode.

Inputs: 
ptT38Params = relay channel pointer


Return Value: None

---------------------------------------------------------------------------*/

void SetDpModeHdlcLSRx(tFAXRELAY_ESL_T38 *ptT38Params){

/* Configure DP in low speed (V.21) mode to receive HDLC frmaes */
    ConfigDpInHdlcModeRx(ptT38Params,RATE_V21300);

/* To initialize relay parameters for DP in receiever mode */
//    InitHdlcRx();
    ptT38Params->iHdlcEnable = 1;
/* To wait for status flags to be set either by DP messages or by packets
   received from the IP network */
    WaitForSettingMode(ptT38Params);

//Dhruva 080604
    if(((ptT38Params->iPrevFrameTx & 0x7F) == DCNID) || ((ptT38Params->iPrevFrameRx & 0x7F) == DCNID ))
    {
        ptT38Params->tMessage.iId = mT38_ENDOFSESSION;
        ptT38Params->tMessage.iLen = 0;
        SendMsgToHost(ptT38Params);
    }

    return;
}


/*-------------------------------------------------------------------------

Function Name: WaitForSettingMode 

Description :
We call the functions to set the datapump in Hdlc LS Tx/ Hdlc HS Tx (ECM),
 HS Tx/Hdlc LS Rx mode, depending on whether the PreambleTxFlg/TrnTxFlg /
PreambleRxFlg is set. If these flags are not set, the state function ptr is 
set to WaitForSettingMode.   

Inputs: 
ptT38Params = relay channel pointer

Return Value: None


---------------------------------------------------------------------------*/


void WaitForSettingMode(tFAXRELAY_ESL_T38 *ptT38Params){
    

   
    ptT38Params->fpStateFnPtr = (fpPtrIntFn )(WaitForSettingMode) ;


/* The PreambleRxFlg is set on detection of 3 7e's at the begining of low
   speed data reception */ 
     if(ptT38Params->iPreambleRxFlg)
     {
//dhruva~
		ptT38Params->IgnorePkts = 1;
//~dhruva
		ptT38Params->iDisablePktData=1;
        ptT38Params->iPreambleRxFlg=0;
        T38PktT30Ind(ptT38Params,V21PREAMBLE);
		if(++ptT38Params->iCurrentRedundancy >= ptT38Params->iLSRedundancy)
			ptT38Params->iCurrentRedundancy = ptT38Params->iLSRedundancy;

//dhruva
        ptT38Params->iResendPktCnt = 2;

        ptT38Params->fpStateFnPtr =(fpPtrIntFn )(PackHdlcDataPkt);
     }

    /* The TrnTxFlg is set on depacking of Trn Ind Pkt which is 
        the begining of High Speed Data Frame */
    else if(ptT38Params->iTrnTxFlg)
    {
        if( ((ptT38Params->iPrevFrameTx & 0x7F) == DCSID) && 
                    ((ptT38Params->iPrevFrameRx & 0x7F) != CFRID) )
        {
            ptT38Params->iImageDataTx = 0;
            SetDpModeHSTx(ptT38Params);        /* It is TCF data */
            ptT38Params->iConnected = 0;
        }
        else
        {
            /* This is High speed data in ECM */
            if(ptT38Params->iEcm)
            {
                ptT38Params->iTrnTxFlg = 0;
                ptT38Params->iHdlcHSEnableTx = 1;
                SetDpModeHdlcTx(ptT38Params,ptT38Params->idRateTx); 
            }
            else
            {
                printf("\r\n WaitForSettingMode: Setting HS mode");
                //Reset the paramters related to TCF
                ResetTCFParams(ptT38Params);
                 ptT38Params->iImageDataTx = 1;
                SetDpModeHSTx(ptT38Params); //High speed image data packets will
                ptT38Params->iConnected = 0;// be received
            }
        }
     }

	 /* The PreambleTxFlg is set on reception of Preamble Ind Pkt which is 
        the begining of Low Speed Data Frame */
    else if(ptT38Params->iPreambleTxFlg)
    {
        ptT38Params->iPreambleTxFlg = 0;
        SetDpModeHdlcTx(ptT38Params,RATE_V21300);
    }


    return;
}

/*-------------------------------------------------------------------------

Function Name: SetDpModeHdlcTx 

Description :
This function sets the datapump in Hdlc Tx mode with 
data rate = dRate. If it is Hdlc HS Tx mode we wait for connect message from
the data pump before starting to transmit the preamble and if it is Hdlc LS Tx 
we start transmitting preamble.

Input: 
ptT38Params = relay channel pointer
dRate - rate at which the DP is configured for HDLC transmision


Return Value: None


---------------------------------------------------------------------------*/

void SetDpModeHdlcTx(tFAXRELAY_ESL_T38 *ptT38Params,INT16 dRate){


    ConfigDpInHdlcModeTx(ptT38Params,dRate);
 
/* If mode of operation is high speed HDLC, preamble flags are not sent
   before getting connect message from DP.
   Preamble flags sent if mode of operation is low speed HDLC.
*/
    if(ptT38Params->iHdlcHSEnableTx)
        WaitForConnMsgToTxFlgs(ptT38Params);
    else
        TxPreamble(ptT38Params);

    return;
}

/*-------------------------------------------------------------------------

Function Name: WaitForConnMsgToTxFlgs 

Description :
In Hdlc HS mode, we wait till we get connected message from data pump, after 
getting it we set Preamble Tim to 20*10 = 200 ms ( the state function is 
called every 10ms). Then the state fn ptr is set to TxPreamble().

Input: 
ptT38Params = relay channel pointer


Return Value: None

---------------------------------------------------------------------------*/

void WaitForConnMsgToTxFlgs(tFAXRELAY_ESL_T38 *ptT38Params){

/* In the HS Hdlc mode , the Tx of flags should continue for 20ms after the
    connect message is obtained from the DP */

    if(ptT38Params->iConnected)
     {
        ptT38Params->iPreambleTim = 20;               //Frame length = 20ms

        if(ptT38Params->iFrameLength == FL_5MS)      //Frame length = 5ms
        {
            ptT38Params->iPreambleTim = 40;
        }
        if(ptT38Params->iFrameLength == FL_20MS)      //Frame length = 20ms
        {
            ptT38Params->iPreambleTim = 10;
        }
#ifdef T30_T38
        ptT38Params->iPreambleTim = 1;
#endif
        ptT38Params->fpStateFnPtr = (fpPtrIntFn)(TxPreamble);
     }

    return;
}

/*-------------------------------------------------------------------------

Function Name: TxPreamble 

Description :
Relay waits for 1second for the DP to send preamble flags to the fax terminal.
In Hdlc HS mode, we wait till we get connected message from data pump, after 
getting it we set Preamble Tim to 20*10 = 200 ms ( the state function is 
called every 10ms). Then the state fn ptr is set to TxPreamble().

Input: 
ptT38Params = relay channel pointer

Return Value:   None


---------------------------------------------------------------------------*/

void TxPreamble(tFAXRELAY_ESL_T38 *ptT38Params){

/* State machine waits for the DP to send flags to the fax terminal till
   the timer elapses */
    if(ptT38Params->iPreambleTim)
    {
         if(ptT38Params->iHdlcHSEnableTx)
              ptT38Params->iTxHdlcLSTimeOut++;
        --ptT38Params->iPreambleTim;
        ptT38Params->fpStateFnPtr = (fpPtrIntFn)(TxPreamble);
    }
    else
/* To check for availabilty of HDLC data from the IP network */
        ChkHdlcDataAvl(ptT38Params);

    return;
}

/*-------------------------------------------------------------------------

Function Name: ChkHdlcDataAvl 

Description :
In Hdlc mode, we wait till the HdlcDataAvl flag is set, once it is set we
start transmitting the data from the buffer pointed to by DataBufPtr to the
datapump. 

Input: 
ptT38Params = relay channel pointer

Return Value:   None

---------------------------------------------------------------------------*/

void ChkHdlcDataAvl(tFAXRELAY_ESL_T38 *ptT38Params){

    ptT38Params->fpStateFnPtr = (fpPtrIntFn)(ChkHdlcDataAvl);
/* If HDLC data is available from the IP network, it is copied to the spoofing
   buffer and sent to DP after analysing it in the function 
   "CopyHdlcDataToSpoofingBuf" */

    if(ptT38Params->iHdlcDataAvl) 
        CopyHdlcDataToSpoofingBuf(ptT38Params);
    else if(!ptT38Params->iHdlcHSEnableTx) {
		ptT38Params->iTxHdlcLSTimeOut++;   //time out counter is incremented
                          //if data is not available

//#ifdef commented on 160804 for CRC Error Fixes
//#ifdef T30_T38	//dhruva14
		if(ptT38Params->iTxHdlcLSTimeOut >= TXHDLCLSTIME) {
        //End FaxTx and set the Dp in Hdlc LS Rx
			ptT38Params->iPreambleTxFlg=0;
			SendEndMsgToDp(ptT38Params);
			ptT38Params->fpStateFnPtr = (fpPtrIntFn)(SetDpModeHdlcLSRx);
			InitHdlcRx(ptT38Params);
		}
//#endif
//#endif commented on 160804 for CRC Error Fixes

	}
    return;
}

/*-------------------------------------------------------------------------

Function Name: CopyHdlcDataToSpoofingBuf 

Description :
    The HDLC data bytes in the received packets are copied to a buffer
    after byte reversal.
    In low speed HDLC mode, the bytes are analysed and the relay 
    operation is appropriately guided.
    The reversed data bytes are sent to the DP through messages only after
    an indicator packet indicating the end of HDLC frame is recieved from
    the IP network.
    Information of correctness of the HDLC frame is also sent to the DP
    through messages ( FCSOK and FCSBAD)

Input: 
ptT38Params = relay channel pointer

Return Value:   None

---------------------------------------------------------------------------*/

void CopyHdlcDataToSpoofingBuf(tFAXRELAY_ESL_T38 *ptT38Params){
    
    UINT16 datalen,i,DataIn;

//If WritePtr = ReadPtr, check CRC, if it is CRCOK send 0x100 to DP 
//to indicate OK status else send 0x200 to indicate CRC Error

    //printf("\r\n CopyHdlcDataToSpoofingBuf Port[%d]",ptT38Params->iPortNo);
    ptT38Params->iTxHdlcLSTimeOut++; 

/* Check for HDLC frame end */
    if(ptT38Params->iCRCWIndex != ptT38Params->iCRCRIndex)
    {
        //printf("\r\n CopyHdlcDataToSpoofingBuf iTxHdlcLSTimeOut[%x] Port[%d]",ptT38Params->iTxHdlcLSTimeOut,ptT38Params->iPortNo);

		datalen = ptT38Params->aiSizeBuf[ptT38Params->iSizeRIndex];
 
		//ptT38Params->iTxHdlcLSTimeOut++; 
        ptT38Params->iHdlcTxByteCnt = 0;

		for(i=0 ; i < datalen ; i++)
		{

			ptT38Params->iHdlcTxByteCnt++;

//Check the control field of the frame
			if(ptT38Params->iHdlcTxByteCnt==2) {
				ptT38Params->iFinalFrame = *ptT38Params->piReadDataPtr & 0x08;
			}
			
	/* The "Info field" (3rd byte) in the HDLC frame is checked and appropraite functions are called */
			if(ptT38Params->iHdlcTxByteCnt==3)
			{
				ptT38Params->iPrevFrameTx = *ptT38Params->piReadDataPtr;

	#ifdef TCP_MODE
				if( (ptT38Params->iPrevFrameTx&0x7f) == CFRID || 
					(ptT38Params->iPrevFrameTx&0x7f) == FTTID)
				ProcCFRFrameTx(ptT38Params);
	#endif

			
				if( (ptT38Params->iPrevFrameTx&0x7f) == NSFID)
				{
                                        printf("\r\n Ignoring NSF Frame");
					//IgnoreNSFFrame
					ptT38Params->piReadDataPtr += 
					(ptT38Params->aiSizeBuf[ptT38Params->iSizeRIndex] - 2);
					if(ptT38Params->piReadDataPtr >= 
						 ptT38Params->piDataBufPtr + ptT38Params->iSizeOfDataBuf)
						ptT38Params->piReadDataPtr -= ptT38Params->iSizeOfDataBuf;

					ptT38Params->aiSizeBuf[ptT38Params->iSizeRIndex]=0;
					if(++ptT38Params->iSizeRIndex == SIZEBUFSIZE)
						ptT38Params->iSizeRIndex = 0;

					ptT38Params->aiCRCBuf[ptT38Params->iCRCRIndex] = 0;
					if(++ptT38Params->iCRCRIndex == CRCBUFSIZE)
						ptT38Params->iCRCRIndex = 0;

					ptT38Params->piWrSpoofingBufPtr = 
					ptT38Params->piRdSpoofingBufPtr = 
					ptT38Params->piSpoofingBufPtr;

					ptT38Params->iHdlcTxByteCnt = 0;

					return;
				}
			}
			else if(ptT38Params->iHdlcTxByteCnt==5)
			{
				//process DCS bytes
				if((ptT38Params->iPrevFrameTx&0x7f) ==DCSID)
					ProcDCSFrameTx(ptT38Params,*ptT38Params->piReadDataPtr);
				else if((ptT38Params->iPrevFrameTx&0x7f) ==CTCID)
					ProcDCSFrameTx(ptT38Params,*ptT38Params->piReadDataPtr);
///////////////////// to force V.27 ter mode   
#if 1
            else if((ptT38Params->iPrevFrameTx&0x7f) ==DISID)
            {
#if 1 
                if(gFaxParams.bRate == V27)
                    *ptT38Params->piReadDataPtr = 0x50;  //9.6
                else
                    *ptT38Params->piReadDataPtr = 0x60;  //4.8

#else
                *ptT38Params->piReadDataPtr = 0x50;  //4.8
                *ptT38Params->piReadDataPtr = 0x60;  //9.6
#endif

            }
#endif
///////////////////
			}

	// non ECM
			else if(ptT38Params->iHdlcTxByteCnt==7)
			{
				if((ptT38Params->iPrevFrameTx&0x7f) ==DISID)
					*ptT38Params->piReadDataPtr &= 0xdf;
			}

			DataIn = *ptT38Params->piReadDataPtr;
			if(++ptT38Params->piReadDataPtr == ptT38Params->piDataBufPtr + ptT38Params->iSizeOfDataBuf)
				  ptT38Params->piReadDataPtr = ptT38Params->piDataBufPtr;

			*ptT38Params->piWrSpoofingBufPtr = 
						ByteReverse((CHAR8)(DataIn&0xFF));
			if(++ptT38Params->piWrSpoofingBufPtr == 
			   ptT38Params->piSpoofingBufPtr + ptT38Params->iSizeOfSpoofingBuf)
					ptT38Params->piWrSpoofingBufPtr = 
								ptT38Params->piSpoofingBufPtr;
		}

////
        ptT38Params->aiSizeBuf[ptT38Params->iSizeRIndex]=0;
        if(++ptT38Params->iSizeRIndex == SIZEBUFSIZE)
            ptT38Params->iSizeRIndex = 0;

        ptT38Params->iPrevCRCStatus = 
            ptT38Params->aiCRCBuf[ptT38Params->iCRCRIndex];

        ptT38Params->aiCRCBuf[ptT38Params->iCRCRIndex] = 0;
        if(++ptT38Params->iCRCRIndex == CRCBUFSIZE)
            ptT38Params->iCRCRIndex = 0;

        ptT38Params->iHdlcTxByteCnt = 0;
        ptT38Params->iProcLSFrameTx = 0;    
        
        if(ptT38Params->piFrameDataPtr == ptT38Params->piReadDataPtr)
        {
            ptT38Params->iHdlcDataAvl = 0;
        }


#ifdef IMPAIRMENT
		if(ptT38Params->Impairment & REJECT_LS_FRAME && ptT38Params->iFinalFrame)
		{
			if((ptT38Params->iPrevFrameTx&0x7f) == DISID &&
				(ptT38Params->Impairment & BAD_DIS) ) {
				ptT38Params->Impairment &= ~BAD_DIS;
				ptT38Params->iPrevCRCStatus = 4;
#ifdef DEF_DEBUG
				if(ptT38Params->EnableDebug)
					printf("\n Rejecting DIS");
#endif
			}
			if( ((ptT38Params->iPrevFrameTx&0x7f) == DCSID ) &&
				(ptT38Params->Impairment & BAD_DCS) ) {
				ptT38Params->Impairment &= ~BAD_DCS;
				ptT38Params->iPrevCRCStatus = 4;
#ifdef DEF_DEBUG
				if(ptT38Params->EnableDebug)
					printf("\n Rejecting DCS");
#endif
			}
			if(((ptT38Params->iPrevFrameTx&0x7f) == CFRID ) &&
				(ptT38Params->Impairment & BAD_CFR) ) {
				ptT38Params->Impairment &= ~BAD_CFR;
				ptT38Params->iPrevCRCStatus = 4;
#ifdef DEF_DEBUG
				if(ptT38Params->EnableDebug)
					printf("\n Rejecting CFR");
#endif
			}
			if(((ptT38Params->iPrevFrameTx&0x7f) == FTTID ) &&
				(ptT38Params->Impairment & BAD_FTT) ) {
				ptT38Params->Impairment &= ~BAD_FTT;
				ptT38Params->iPrevCRCStatus = 4;
#ifdef DEF_DEBUG
				if(ptT38Params->EnableDebug)
					printf("\n Rejecting FTT");
#endif
			}
			if(((ptT38Params->iPrevFrameTx&0x7f) == MCFID ) &&
				(ptT38Params->Impairment & BAD_MCF) ) {
				ptT38Params->Impairment &= ~BAD_MCF;
				ptT38Params->iPrevCRCStatus = 4;
#ifdef DEF_DEBUG
				if(ptT38Params->EnableDebug)
					printf("\n Rejecting MCF");
#endif
			}
			if( (((ptT38Params->iPrevFrameTx&0x7f) == EOPID) ||
				((ptT38Params->iPrevFrameTx&0x7f) == MPSID ) )&&
				(ptT38Params->Impairment & BAD_POSTMSGCMD) ) {
					ptT38Params->Impairment &= ~BAD_POSTMSGCMD;
					ptT38Params->iPrevCRCStatus = 4;
#ifdef DEF_DEBUG
				if(ptT38Params->EnableDebug)
					printf("\n Rejecting Post image Command");
#endif
			}
		}
#endif

// Transmit HDLC data to DP through a message
        TxHdlcDataToDp(ptT38Params);

/* If CRCStatus is 1 or 3, then the HDLC frame is not in error else it is error
This information is sent to DP through a message [ FCSOK indicates an error
free HDLC frame FCSBAD indicates an erroneous HDLC frame
*/

//dhruva added on 19/11/2002
	if(!(ptT38Params->iPrevCRCStatus & 1))
		ptT38Params->iPrevFrameTx = 0;

//dhruva, added on 10/10/2002 to wait for sig-end pkt for non final frames
//to be sent to DP/T.30. debugging with cisco g/w


 	if(!ptT38Params->iFinalFrame)
   	{
//Message ID mT38MSGTODP indicates host to forward this message to DP
       
  	   ptT38Params->tMessage.iId = 
                (ptT38Params->iPrevCRCStatus & 1)?mDP_FCSOK:mDP_FCSBAD;
           ptT38Params->tMessage.iLen = 0;
           ptT38Params->iPreambleTxFlg = 0;
           SendMsgToHost(ptT38Params);

       /*If CRCStatus = 1 or 3 and if it is Hdlc LS mode we analyze 
         the Rx frame for its type and contains */ 

           ProcEndHdlcDataTx(ptT38Params);
   	}

    else {
	   WaitForSigEndTx1(ptT38Params);
	}

    }
    else if(ptT38Params->iTxHdlcLSTimeOut == TXHDLCLSTIME) 
    {    
            /*End FaxTx and set the Dp in Hdlc LS Rx*/
        printf("\r\n iTxHdlcLSTimeOut Port[%d]",ptT38Params->iPortNo);
        ptT38Params->iPreambleTxFlg=0;
        SendEndMsgToDp(ptT38Params);
        //ptT38Params->fpStateFnPtr = (fpPtrIntFn)(SetDpModeHdlcLSRx); 

        /* State machine is set to wait for acknowledge from DP for the above message*/
        ptT38Params->fpStateFnPtr = (fpPtrIntFn)(WaitForAckHdlcTx); 
        //////////////

        InitHdlcRx(ptT38Params);
    }
     
    return;
}

/*-----------------------------------------------------------------------*/

void WaitForSigEndTx1(tFAXRELAY_ESL_T38 *ptT38Params)
{

        ptT38Params->fpStateFnPtr = (fpPtrIntFn)(WaitForSigEndTx1);

	if(ptT38Params->iSigEnd)
	{
           printf("\r\n T382Dp FCS[%x] port[%d]",ptT38Params->iPrevCRCStatus ,ptT38Params->iPortNo);
	   ptT38Params->tMessage.iId = 
                (ptT38Params->iPrevCRCStatus & 1)?mDP_FCSOK:mDP_FCSBAD;
           ptT38Params->tMessage.iLen = 0;
           ptT38Params->iPreambleTxFlg = 0;
           SendMsgToHost(ptT38Params);

       /*If CRCStatus = 1 or 3 and if it is Hdlc LS mode we analyze 
         the Rx frame for its type and contains */ 


           ProcEndHdlcDataTx(ptT38Params);

	}

}




/*-------------------------------------------------------------------------

Function Name: TxHdlcDataToDp 

Description :
    The data bytes in the HDLC frame are put to the message parameter 
 buffer and then updated to the queue by calling  the function "SendDataToDP".


Input: 
ptT38Params = relay channel pointer

Return Value:   None

---------------------------------------------------------------------------*/

void TxHdlcDataToDp(tFAXRELAY_ESL_T38 *ptT38Params){
    INT16 datalen,i;
    UCHAR8 byte;

/*For Hdlc High Speed Data */
    if(ptT38Params->iHdlcHSEnableTx)
    {
        datalen = ptT38Params->piWrSpoofingBufPtr- ptT38Params->piRdSpoofingBufPtr;
        datalen = (datalen >= 0)? datalen : datalen+ ptT38Params->iSizeOfSpoofingBuf;
        if(datalen>= ptT38Params->iBytesInHS)
            ptT38Params->tMessage.iLen = ptT38Params->iBytesInHS;
        else
            ptT38Params->tMessage.iLen = datalen;

        for(i=0 ; i < ptT38Params->tMessage.iLen ; i++)
        {
            byte = *ptT38Params->piReadDataPtr;
            if(++ptT38Params->piReadDataPtr == ptT38Params->piDataBufPtr + ptT38Params->iSizeOfDataBuf)
                ptT38Params->piReadDataPtr = ptT38Params->piDataBufPtr;
            ptT38Params->tMessage.aiParams[i] = (UCHAR8)ByteReverse(byte);
        }

//Message ID mT38MSGTODP indicates host to forward this message to DP
        ptT38Params->tMessage.iId = mDP_MOD_DATA;
        if(ptT38Params->tMessage.iLen > 0)
            SendMsgToHost(ptT38Params);
    }

/* For Hdlc LS Data */
    else
    {
        datalen = ptT38Params->piWrSpoofingBufPtr- ptT38Params->piRdSpoofingBufPtr;
        if(datalen < 0)
            datalen += ptT38Params->iSizeOfSpoofingBuf;

        ptT38Params->tMessage.iLen = datalen;

        printf("\r\nT382Dp port[%d]", ptT38Params->iPortNo);
        for(i=0 ; i < ptT38Params->tMessage.iLen ; i++)
        {
            byte = *(ptT38Params->piRdSpoofingBufPtr);
            if(++ptT38Params->piRdSpoofingBufPtr == ptT38Params->piSpoofingBufPtr + ptT38Params->iSizeOfSpoofingBuf)
            {
                ptT38Params->piRdSpoofingBufPtr = ptT38Params->piSpoofingBufPtr;
            }
            ptT38Params->tMessage.aiParams[i] = byte; 
            printf(" %x", byte);
        }

/////////////////////////////////////////////////////////////////////////
//added for Testing the US Robotics Failure Case, CFR reaching to originator after 3 seconds.
#ifdef CFRTESTING 
        //Check the CFR bytes, if available then configure
        //Data pump in V.21 Transmit Mode for sending
        //CFR bytes V.21 signal to the Fax machine
        //for(i=0;i< ptT38Params->tMessage.iLen ; i++)
        for(i=0;i < 3 ; i++) //as only 3 bytes for CFR
        {
           byte = ptT38Params->tMessage.aiParams[i];  
           if((byte == 0xff)||(byte == 0x13)||(byte == 0x84))
              CFRRxCntr[ptT38Params->iPortNo]++;
        }
       
        if(CFRRxCntr[ptT38Params->iPortNo] == 3)
        {
           CFRRxCntr[ptT38Params->iPortNo] = 0;
           ConfigDpInHdlcModeTx(ptT38Params,RATE_V21300);
        }
#endif
////////////////////////////////////////////////////////////////////////
        printf("\r\n");
//Message ID mT38MSGTODP indicates host to forward this message to DP
        ptT38Params->tMessage.iId = mDP_MOD_DATA;
        if(ptT38Params->tMessage.iLen > 0)
            SendMsgToHost(ptT38Params);
    }
    return;
}


/*-------------------------------------------------------------------------

Function Name: SendCRP

Description :
    This function is used to send the HDLC frame "CRP" which asks for
 repeatation of the previous HDLC frame.
 The bytes in CRP frame are sent as messages to DP which inturn applies the
 low speed modulation technique (V.21) and send to the fax terminal.

Input: 
ptT38Params = relay channel pointer

Return Value:   None

---------------------------------------------------------------------------*/

void SendCRP(tFAXRELAY_ESL_T38 *ptT38Params){


        ptT38Params->tMessage.aiParams[0] = (UINT16)(255);//Address field
        ptT38Params->tMessage.aiParams[1] = (UINT16)(19); //Control field
        ptT38Params->tMessage.aiParams[2] = (UINT16)(27); //information field
        ptT38Params->tMessage.iLen = 3;                //length of the message
        ptT38Params->tMessage.iId = mDP_MOD_DATA;
        SendMsgToHost(ptT38Params);

/*Also a message indicating that the above HDLC frame is not in error is sent*/
//Message ID mT38MSGTODP indicates host to forward this message to DP
        ptT38Params->tMessage.iId = mDP_FCSOK;
        ptT38Params->tMessage.iLen = 0;
        SendMsgToHost(ptT38Params);

/* Function call to set the state machine for further operation od the relay */
        ptT38Params->iPrevCRCStatus = 3;    //CRP is a final frame
        ProcEndHdlcDataTx(ptT38Params);
        ptT38Params->iPrevCRCStatus = 0;
        return;

}

/*-------------------------------------------------------------------------

Function Name: SendMsgToHost 

Description :
    Whenever a message is available, it is put in the InMessage Queue.

Input: 
ptT38Params = relay channel pointer

Return Value:   None

---------------------------------------------------------------------------*/

void SendMsgToHost(tFAXRELAY_ESL_T38 *ptT38Params){
    
    UINT16 i;
    UCHAR8 *End;

    /* Up date the length of the message to the queue */    
    End = (UCHAR8 *)(ptT38Params->ptT38OutMsgQStr->paiBufPtr + 
                            ptT38Params->ptT38OutMsgQStr->wSize);

    *ptT38Params->ptT38OutMsgQStr->paiWrite = ptT38Params->tMessage.iLen;

    /* Increment the wirte pointer to the queue circularly */
    if( ++ptT38Params->ptT38OutMsgQStr->paiWrite == End)
        ptT38Params->ptT38OutMsgQStr->paiWrite = 
                ptT38Params->ptT38OutMsgQStr->paiBufPtr;

    /* Up date the message ID to the queue */    
    *ptT38Params->ptT38OutMsgQStr->paiWrite = ptT38Params->tMessage.iId;

    /* Increment the wirte pointer to the queue circularly */
    if( ++ptT38Params->ptT38OutMsgQStr->paiWrite == End)
        ptT38Params->ptT38OutMsgQStr->paiWrite = 
                    ptT38Params->ptT38OutMsgQStr->paiBufPtr;

    /* Up date the message parameters to the queue */    
    for(i=0; i < ptT38Params->tMessage.iLen; i++)
    {
        *ptT38Params->ptT38OutMsgQStr->paiWrite = 
            ptT38Params->tMessage.aiParams[i];
        /* Increment the wirte pointer to the queue circularly */
        if( ++ptT38Params->ptT38OutMsgQStr->paiWrite == End)
            ptT38Params->ptT38OutMsgQStr->paiWrite = 
                   ptT38Params->ptT38OutMsgQStr->paiBufPtr;
    }
    return;
}


/*-------------------------------------------------------------------------

Function Name: SendTxLvlMsgToDp 

Description :
Send the Tx level message to the datapump. 

Input: 
(1)ptT38Params = relay channel pointer
(2)    ID : message ID
(3) TxLvl : Transmit level

Return Value:   None

---------------------------------------------------------------------------*/

void SendTxLvlMsgToDp(tFAXRELAY_ESL_T38 *ptT38Params,INT16 Id,INT16 TxLvl){
    //Message ID mT38MSGTODP indicates host to forward this message to DP
    ptT38Params->tMessage.iId = Id;
    ptT38Params->tMessage.iLen = 1;
    ptT38Params->tMessage.aiParams[0] = TxLvl;
/* Function call to update the In message queue with the above parameters*/
    SendMsgToHost(ptT38Params);
    return;
}

/*-------------------------------------------------------------------------

Function Name: SendTxRxMsgToDp 

Description :
    Send the message to set the Tx / Rx mode to the datapump. 

Input: 
(1)ptT38Params = relay channel pointer
(2)    ID : message ID
(3) Param : Parameter to be put to the message queue

Return Value:   None

---------------------------------------------------------------------------*/

void SendTxRxMsgToDp(tFAXRELAY_ESL_T38 *ptT38Params,INT16 Id,INT16 Param){
//Message ID mT38MSGTODP indicates host to forward this message to DP
    ptT38Params->tMessage.iId = Id;
    ptT38Params->tMessage.iLen = 1;
    ptT38Params->tMessage.aiParams[0] = Param ;
/* Function call to update the In message queue with the above parameters*/
    SendMsgToHost(ptT38Params);
    return;
}

/*-------------------------------------------------------------------------

Function Name: CopyDataToBeTx 

Description :
    Copy data from the input packet to the buffer pointed to by DataBufPtr
    Also the bit stream in the packet are checked for EOL in high speed image
    transfer mode

Input: 
ptT38Params = relay channel pointer
  Ptr - pointer to the buffer containing data to be copied

Return Value:   None

---------------------------------------------------------------------------*/

void CopyDataToBeTx(tFAXRELAY_ESL_T38 *ptT38Params,UCHAR8 *Ptr)
{
   UINT16 len,i;
   UCHAR8 ch;

   len = (*Ptr++ << 8 ) & 0xff00;
   len += (*Ptr++);
   len += 1;
    
// If image data packets are being received, check the bit stream for
// EOL sequence
    //printf(" ImgFlg[%d] HdlcFlg[%d]",ptT38Params->iImageDataTx,ptT38Params->iHdlcDataAvl); 
    if(ptT38Params->iImageDataTx && !ptT38Params->iHdlcDataAvl)
    {
	for(i=0;i<len;i++) {
            *ptT38Params->pTempImageW = *Ptr++;
            /* Increment the write pointer circularly over the BUFLEN*/
            if( ++ptT38Params->pTempImageW == 
                   ptT38Params->TempImageBuf + ptT38Params->iSizeOfTempImageBuf)
            {
                ptT38Params->pTempImageW = ptT38Params->TempImageBuf;
            }
	}

    }
    else {
		do
    	{    
        /* Copy the data to the data buffer */
            ch = *Ptr++;
            *ptT38Params->piFrameDataPtr = ch;
            /* Increment the write pointer circularly over the BUFLEN*/
            if( ++ptT38Params->piFrameDataPtr == 
                   ptT38Params->piDataBufPtr + ptT38Params->iSizeOfDataBuf)
            {
                ptT38Params->piFrameDataPtr = ptT38Params->piDataBufPtr ;
            }

//increment the data counter
            ptT38Params->aiSizeBuf[ptT38Params->iSizeWIndex]++;

            /* Validate the TCF data for zeroes*/
            if(ptT38Params->iTCFdata)
            {
                if(ch)
                    ptT38Params->iTCFZeroCnt = 0;
                else
                    ptT38Params->iTCFZeroCnt++;

            /* if zeroes are received for more than for 1.35seconds, then valid
            TCF data packets have been received */

                if(ptT38Params->iTCFZeroCnt >= 
                        ptT38Params->iReqdZeroesInTCFForTx)
                    ptT38Params->iGoodTCFPkts = 1;

                else if (ptT38Params->iGoodTCFPkts != 1)
                    ptT38Params->iGoodTCFPkts = 2;
            }
	    } while(--len > 0);
	}

   return;
}


void AnalyzeDSImage(tFAXRELAY_ESL_T38 *ptT38Params)
{
   INT16 k,datalen,i;
   UCHAR8 ch;

	datalen = ptT38Params->pTempImageW - ptT38Params->pTempImageR;
	if(datalen < 0 )
		datalen += ptT38Params->iSizeOfTempImageBuf;
	
	if(datalen > 30)
		datalen = 30;
	
	for (i=0;i<datalen;i++) {

        ch = *ptT38Params->pTempImageR++;
        if( ptT38Params->pTempImageR == 
            ptT38Params->TempImageBuf + ptT38Params->iSizeOfTempImageBuf)
        {
            ptT38Params->pTempImageR = ptT38Params->TempImageBuf;
        }
        
		k=7;
		do
        {
            if((ch >> k)& 1)
            {
      /* Check if a bit 1 is received after continuos 11 or more zeroes*/
                if (ptT38Params->iZeroBitCount >= 11)
                {
                  //printf("\r\n AnalyzeDSImage: ptT38Params->iZeroBitCount[%d]",ptT38Params->iZeroBitCount);
                    ptT38Params->iZeroBitCount = 0;
                    ptT38Params->iEOLCnt++;
                    if(++ptT38Params->iSizeWIndex == SIZEBUFSIZE)
                        ptT38Params->iSizeWIndex = 0;
                    ptT38Params->aiSizeBuf[ptT38Params->iSizeWIndex] = 0;
                    break;
                }
                ptT38Params->iZeroBitCount = 0;
            }
            else
            {
                //printf(" %d",ptT38Params->iZeroBitCount);
            	ptT38Params->iZeroBitCount++;
            } 
        } while(--k >= 0);

        *ptT38Params->piFrameDataPtr = ch;
        /* Increment the write pointer circularly over the BUFLEN*/
        if( ++ptT38Params->piFrameDataPtr == 
            ptT38Params->piDataBufPtr + ptT38Params->iSizeOfDataBuf)
        ptT38Params->piFrameDataPtr = ptT38Params->piDataBufPtr ;

//increment the data counter
        ptT38Params->aiSizeBuf[ptT38Params->iSizeWIndex]++;
	}
	
    if((ptT38Params->pTempImageW == ptT38Params->pTempImageR) && 
          ptT38Params->iSigEndHS && datalen)
    {
        if(++ptT38Params->iSizeWIndex == SIZEBUFSIZE)
            ptT38Params->iSizeWIndex = 0;
        ptT38Params->aiSizeBuf[ptT38Params->iSizeWIndex] = 0;
    }
}



/*-------------------------------------------------------------------------

Function Name: GetMsgFromHost

Description :
Get messages from the datapump and put them in the message parameter buffer.
Read/analyse the message to set the relevant flags.

Input:  ptT38Params = relay channel pointers

Return Value:
    None

---------------------------------------------------------------------------*/

void GetMsgFromHost(tFAXRELAY_ESL_T38 *ptT38Params,tMSGQSTRUCT *ptMsgQPtr)
{
    UCHAR8 *End;
    INT16 i;

/* Upper boundary of the Out message queue */
    End = ptMsgQPtr->paiBufPtr + ptMsgQPtr->wSize;

/* If read pointer and write pointers are not equal, then there are messages
   by the messages */
    while(ptMsgQPtr->paiRead != ptMsgQPtr->paiWrite)
    {
/* Read the length of the message */
        ptT38Params->tMessage.iLen = *ptMsgQPtr->paiRead;

/* Increment read message pointer circularly */
        if( ++ptMsgQPtr->paiRead == End )
          ptMsgQPtr->paiRead = ptMsgQPtr->paiBufPtr;

/* Read the message ID */
        ptT38Params->tMessage.iId = *ptMsgQPtr->paiRead ;

/* Increment read message pointer circularly */
        if( ++ptMsgQPtr->paiRead == End )
          ptMsgQPtr->paiRead = ptMsgQPtr->paiBufPtr;

/* Read the message parameters  */
        for(i=0 ; i < ptT38Params->tMessage.iLen ; i++)
        {
            ptT38Params->tMessage.aiParams[i] = *ptMsgQPtr->paiRead;
            if( ++ptMsgQPtr->paiRead == End )
                ptMsgQPtr->paiRead = ptMsgQPtr->paiBufPtr;
        }

/* Function call to analyse the messages */        
        ProcessMessages(ptT38Params);
    }

    return;
}

/*-------------------------------------------------------------------------

Function Name: ProcessMessages 

Description :
    Analyse the message and set the relevant flags. 

Input: 
ptT38Params = relay channel pointer
Message Id

Return Value:
    None

---------------------------------------------------------------------------*/

void ProcessMessages(tFAXRELAY_ESL_T38 *ptT38Params)
{
/* Switch accroding to the ID of the message */    
    switch(ptT38Params->tMessage.iId)
    {
        case mDP_DEMOD_DATA :        // DataToHost
            if(!ptT38Params->iDisableData)
            {
                if(!ptT38Params->iHdlcEnable)
                    HighSpeedData(ptT38Params);
                else
                    HdlcData(ptT38Params);
            }
            break;

        case mDP_CONNECT:
		    ptT38Params->iConnected = 1;    
			break;

        case mDP_NO_CARRIER:
		    ptT38Params->iNoCarStat = 1;    
			break;

        case mDP_DEMOD_CRCOK:
			ptT38Params->iCRCStatusRx = 1;        
			ptT38Params->iResendPktCnt = 1; //dhruva 170203 //BF
			break;

        case mDP_DEMOD_CRCERROR:
			ptT38Params->iCRCStatusRx = 2;        
            ptT38Params->iResendPktCnt = 1; //dhruva 170203 BF
#if 1 //160804 CRC Error Fixes
                        if(ptT38Params->iPreambleRxFlg ==0) 
                           ptT38Params->iCRCStatusRx = 0;
#endif
			break;

        case mDP_FRAME_ACK:
			ptT38Params->iFrameAckFrmDp = 1;
			break;

        case mDP_XON:
			ptT38Params->iXOFF = 0;
			break;

        case mDP_XOFF:
			ptT38Params->iXOFF = 1;
			break;

        case mDP_FAX_TXRX_END:
			ptT38Params->iCmdAckFrmDp = 1;
			if((ptT38Params->iPrevFrameTx & 0x7F) == DCNID)
				ptT38Params->iShutDown = 1;
			break;

        case mDP_DEMOD_DET_FLAG:
			if(++ptT38Params->iCnt7e == 3)    
				ptT38Params->iPreambleRxFlg = 1;
			break;

        case mDP_DEMOD_DET_TRN :        /* Received Training */
            ptT38Params->iTrnRxFlg = 1;

            if(!ptT38Params->iEcm)
                ptT38Params->iHdlcEnable = 0;
            else
                ptT38Params->iHdlcHSEnableRx = 1;

            if( ((ptT38Params->iPrevFrameTx & 0x7F) == CFRID) || 
                ((ptT38Params->iPrevFrameTx & 0x7F) == FTTID) )
            {
                //Reset the paramters related to TCF
                ResetTCFParams(ptT38Params);
            }
			break;

        default :
            break;

    }

    return;
}
/*-------------------------------------------------------------------------

Function Name: ReorderUDPPkts 

Description :
    If no packet is received, then a packet with expected sequence number
    is serached in the reorder buffer.
    If a valid is packet is obtained, then the sequence number of that packet 
    is analyzed and the the following tasks are performed 
    * If the sequence number is same as that expected, the packet is processed.
    * If it is out of sequence, the packet is saved in reorder buffer for
      further consideration.

Input: 
 1. ptT38Params = relay channel pointer
 2. InQPtr = pointer to source buffer
 3. NewIPPkt = Flag to indicate the presence of a new packet.

Return Value:
If a valid packet is available it returns 1 else 0.
---------------------------------------------------------------------------*/
#ifdef UDP_MODE

#ifndef CISCO_GW
INT16 ReorderUDPPkts(tFAXRELAY_ESL_T38 *ptT38Params, 
                     UCHAR8 *InPtr,
                     INT16 NewIPPkt)
{
    UCHAR8 *Ptr1,*Ptr2,*Ptr3;
    UINT16 GotPkt = 0,len,seqno,PosOfReqSecPkt,temp=0,i,j;

    Ptr1 = InPtr;

//  If no packet is received, then a packet with expected sequence number
//    is searched in the reorder buffer.
    if(!NewIPPkt)
    {
        GotPkt = FindPktInBUF(ptT38Params,(ptT38Params->iPrevPktRead+1));
        return(GotPkt);
    }

    ptT38Params->iRxSeqNo = *Ptr1++;
    if(Ptr1 >= ptT38Params->JitterBuffer + ptT38Params->JitterBufLen)
        Ptr1 -= ptT38Params->JitterBufLen;

    ptT38Params->iRxSeqNo <<=8;
    ptT38Params->iRxSeqNo |= (*Ptr1++ & 0xff);
    if(Ptr1 >= ptT38Params->JitterBuffer + ptT38Params->JitterBufLen)
        Ptr1 -= ptT38Params->JitterBufLen;

	InPtr = Ptr1;

//Sequence number of received packet is same as the expected number
    if(ptT38Params->iRxSeqNo == ptT38Params->iPrevPktRead + 1)
    {
        /* Save the contents of the packet for further processing 
        return "1" to indicate analysis of valid packet */
        if(*Ptr1 > 0x1e)
        {
            if(++Ptr1 >= ptT38Params->JitterBuffer + ptT38Params->JitterBufLen)
                Ptr1 -= ptT38Params->JitterBufLen;

            temp = *Ptr1;    //Number of data feild

            Ptr1 += 2;
            if(Ptr1 >= ptT38Params->JitterBuffer + ptT38Params->JitterBufLen)
                Ptr1 -= ptT38Params->JitterBufLen;

            len = *Ptr1;    //Ptr1[3]
            len <<= 8;

            if(++Ptr1 >= ptT38Params->JitterBuffer + ptT38Params->JitterBufLen)
                Ptr1 -= ptT38Params->JitterBufLen;

            len += *Ptr1;    //Ptr1[4];
            if(++Ptr1 >= ptT38Params->JitterBuffer + ptT38Params->JitterBufLen)
                Ptr1 -= ptT38Params->JitterBufLen;

//            CopyChars(ptT38Params->piIpPktPtr,Ptr1,5+ len + temp);
            Ptr1 = InPtr;
            if(Ptr1 >= ptT38Params->JitterBuffer + ptT38Params->JitterBufLen)
                Ptr1 -= ptT38Params->JitterBufLen;
            for(i=0;i< (5+len+temp); i++) {
                ptT38Params->piIpPktPtr[i] = *Ptr1++;
                if(Ptr1 >= ptT38Params->JitterBuffer + ptT38Params->JitterBufLen)
                    Ptr1 -= ptT38Params->JitterBufLen;
            }
        }
    //It is an indicator pkt (thus has only 1 byte)
        else
        {
            *ptT38Params->piIpPktPtr = *Ptr1;
        }
        return(1);
	}

/*	else
		printf("\nDidnot Rx Pkt:0x%x",(ptT38Params->iPrevPktRead+1));
*/
////////////////////


// If a pkt which was already processed has arrived, then discard it
    if(ptT38Params->iRxSeqNo <= ptT38Params->iPrevPktRead)
    {
        //Search for the expected packet in the reorder buffer
        GotPkt = FindPktInBUF(ptT38Params,(ptT38Params->iPrevPktRead+1));
        return(GotPkt);
    }

//To find the number of redundant packets in the received payload

// Check for Type of message
    if(*Ptr1 > 0x1e)
    {
        if(++Ptr1 >= ptT38Params->JitterBuffer + ptT38Params->JitterBufLen)
            Ptr1 -= ptT38Params->JitterBufLen;

    //len = number of Data octets + Length of second Data Field (1)
        temp = *Ptr1;
        Ptr1 += 2;
        if(Ptr1 >= ptT38Params->JitterBuffer + ptT38Params->JitterBufLen)
            Ptr1 -= ptT38Params->JitterBufLen;

        len = *Ptr1++;
        if(Ptr1 >= ptT38Params->JitterBuffer + ptT38Params->JitterBufLen)
            Ptr1 -= ptT38Params->JitterBufLen;
        len <<= 8;
        len += *Ptr1;  //len = number of Data octets - 1
        len += temp;

        Ptr1 = InPtr + len + 8 - 2;
    //Ptr1 now points to the field having number of Secondary pkts
        if(Ptr1 >= ptT38Params->JitterBuffer + ptT38Params->JitterBufLen)
            Ptr1 -= ptT38Params->JitterBufLen;
   }
    else
    {
    //Ptr1 now points to the field having number of Secondary pkts
        Ptr1 += 2;
        if(Ptr1 >= ptT38Params->JitterBuffer + ptT38Params->JitterBufLen)
            Ptr1 -= ptT38Params->JitterBufLen;
    }

//Get the number of secondary packets present in the received pay load
    ptT38Params->iNoOfSecInRxPkt = *Ptr1;
//Ptr1 now points to beginig of Sec Pkts

    if(ptT38Params->iRxSeqNo <= 
        ptT38Params->iPrevPktRead + ptT38Params->iNoOfSecInRxPkt + 1)
    {
// Expected packet is among the redundant packets
        PosOfReqSecPkt = ptT38Params->iRxSeqNo - ptT38Params->iPrevPktRead - 1;

        seqno = ptT38Params->iRxSeqNo;
        Ptr1 = InPtr;
        if(Ptr1 >= ptT38Params->JitterBuffer + ptT38Params->JitterBufLen)
            Ptr1 -= ptT38Params->JitterBufLen;

        while(PosOfReqSecPkt > 0)
        {
// Read the sec pkt starting from Ptr1, Save it to ReOrderBuf[]

            if(*Ptr1 > 0x1e)
            {
                Ptr3 = Ptr1;
                if(++Ptr3 >= ptT38Params->JitterBuffer + ptT38Params->JitterBufLen)
                    Ptr3 -= ptT38Params->JitterBufLen;

        //len = number of Data octets + Length of second Data Field (1)
                temp = *Ptr3;

                Ptr3 += 2;
                if(Ptr3 >= ptT38Params->JitterBuffer + ptT38Params->JitterBufLen)
                    Ptr3 -= ptT38Params->JitterBufLen;

                len = *Ptr3++;
                if(Ptr3 >= ptT38Params->JitterBuffer + ptT38Params->JitterBufLen)
                    Ptr3 -= ptT38Params->JitterBufLen;
                len <<= 8;
                len += *Ptr3++;        //len = number of Data octets - 1
                len += (5+temp);
            }
            else
            {
                len = 1;
            }

            Ptr2 = ptT38Params->piReOrderBufPtr;
            *Ptr2++ = (seqno>>8)&0xff;
            *Ptr2++ = seqno & 0xff;
            *Ptr2++ = len;
            for(i=0; i<len ; i++) {
                //*Ptr2++ = (UINT16)*Ptr1++;
                *Ptr2++ = *Ptr1++;
                if(Ptr1 >= ptT38Params->JitterBuffer + ptT38Params->JitterBufLen)
                    Ptr1 -= ptT38Params->JitterBufLen;
            }

            ptT38Params->piReOrderBufPtr += RX_IFP_MAXSIZE;
            if(ptT38Params->piReOrderBufPtr == 
                    ptT38Params->aiReOrderBuffer + REORDERBUFLEN )
            {
                ptT38Params->piReOrderBufPtr = ptT38Params->aiReOrderBuffer;
            }

            PosOfReqSecPkt--;

            if(--seqno == ptT38Params->iRxSeqNo-1)
            {
        //discard the preamble for secondary and number of seconadry packets
                Ptr1+=2;
                if(Ptr1 >= ptT38Params->JitterBuffer + ptT38Params->JitterBufLen)
                    Ptr1 -= ptT38Params->JitterBufLen;
            }
        }

//Ptr1 points to the required Sec Pkt, Read it.
// Dont save any more pkts to ReOrderBuf
        if(*Ptr1 > 0x1e)
        {
            Ptr3 = Ptr1;
            if(++Ptr3 >= ptT38Params->JitterBuffer + ptT38Params->JitterBufLen)
                Ptr3 -= ptT38Params->JitterBufLen;

    //len = number of Data octets + Length of second Data Field (1)
            temp = *Ptr3;

            Ptr3 += 2;
            if(Ptr3 >= ptT38Params->JitterBuffer + ptT38Params->JitterBufLen)
                Ptr3 -= ptT38Params->JitterBufLen;

            len = *Ptr3++;
            if(Ptr3 >= ptT38Params->JitterBuffer + ptT38Params->JitterBufLen)
                Ptr3 -= ptT38Params->JitterBufLen;
            len <<= 8;
            len += *Ptr3++;        //len = number of Data octets - 1
            len += (5+temp);
        }
        else
        {
            len = 1;
        }

        for(i=0;i<len;i++) {
            ptT38Params->piIpPktPtr[i] = *Ptr1++;
            if(Ptr1 >= ptT38Params->JitterBuffer + ptT38Params->JitterBufLen)
                Ptr1 -= ptT38Params->JitterBufLen;
        }
        return(1);
    }

// If a packet which was supposed to be received much later is received
    else if(ptT38Params->iRxSeqNo > 
        ptT38Params->iPrevPktRead + ptT38Params->iNoOfSecInRxPkt + 1)
    {

        //Search for the expected packet in the reorder buffer
        GotPkt = FindPktInBUF(ptT38Params,(ptT38Params->iPrevPktRead+1));

//Save the primary along with the redundant packets to the reorder buffer
        Ptr1 = InPtr;
        seqno = ptT38Params->iRxSeqNo;

        for(j=0; j < ptT38Params->iNoOfSecInRxPkt+1; j++)
        {
            if(*Ptr1 > 0x1e)
            {
                Ptr3 = Ptr1;
                if(++Ptr3 >= ptT38Params->JitterBuffer + ptT38Params->JitterBufLen)
                    Ptr3 -= ptT38Params->JitterBufLen;

        //len = number of Data octets + Length of second Data Field (1)
                temp = *Ptr3;

                Ptr3 += 2;
                if(Ptr3 >= ptT38Params->JitterBuffer + ptT38Params->JitterBufLen)
                    Ptr3 -= ptT38Params->JitterBufLen;

                len = *Ptr3++;
                if(Ptr3 >= ptT38Params->JitterBuffer + ptT38Params->JitterBufLen)
                    Ptr3 -= ptT38Params->JitterBufLen;
                len <<= 8;
                len += *Ptr3++;        //len = number of Data octets - 1
                len += (5+temp);
            }
            else
            {
                len = 1;
            }

            Ptr2 = ptT38Params->piReOrderBufPtr;
            *Ptr2++ = (seqno>>8)&0xff;
            *Ptr2++ = seqno & 0xff;
            *Ptr2++ = len;
            for(i=0; i<len ; i++) {
                //*Ptr2++ = (UINT16)*Ptr1++;
                *Ptr2++ = *Ptr1++;
                if(Ptr1 >= ptT38Params->JitterBuffer + ptT38Params->JitterBufLen)
                    Ptr1 -= ptT38Params->JitterBufLen;
            }

            seqno--;
            ptT38Params->piReOrderBufPtr += RX_IFP_MAXSIZE;
            if(ptT38Params->piReOrderBufPtr == 
                    ptT38Params->aiReOrderBuffer + REORDERBUFLEN )
            {
                ptT38Params->piReOrderBufPtr = ptT38Params->aiReOrderBuffer;
            }

//ignore the "Preamble for Scondary packet" field and the field for number 
//of secondary packets
            if(j == 0) {
                Ptr1+=2;
                if(Ptr1 >= ptT38Params->JitterBuffer + ptT38Params->JitterBufLen)
                    Ptr1 -= ptT38Params->JitterBufLen;
            }
        }

        return(GotPkt);
    }

   return(GotPkt);
}
#endif
#endif

/*-------------------------------------------------------------------------

Function Name: FindPktInBUF

Description :
    This function searches the required packet in the reorder buffer

Input: 
 1. ptT38Params = relay channel pointer
 2. SeqNo = Sequence number of the packet that is to be searched in the
        reorder buffer.

Return Value:
 FoundPkt = If a valid packet is found it returns 1 else 0.

---------------------------------------------------------------------------*/
#ifdef UDP_MODE

INT16 FindPktInBUF(tFAXRELAY_ESL_T38 *ptT38Params,INT16 SeqNo)
{
    UCHAR8 *Ptr1,*Ptr2;
    INT16 FoundPkt = 0,i,j,Temp;


    for(i=0; i < REORDERBUFLEN; i+= RX_IFP_MAXSIZE)
    {
        Temp = (ptT38Params->aiReOrderBuffer[i] << 8) &0xff00;
        Temp |= (ptT38Params->aiReOrderBuffer[i+1] & 0xff);

        if(SeqNo == Temp)
        {
            Ptr2 = ptT38Params->aiReOrderBuffer + i + 3;
            for(j=0; j < ptT38Params->aiReOrderBuffer[i+2]; j++)
                ptT38Params->piIpPktPtr[j] = *Ptr2++;

            FoundPkt = 1;

			ptT38Params->iNoOfPktInROBuf--;

            Ptr1 = ptT38Params->aiReOrderBuffer + i;

            ptT38Params->piReOrderBufPtr -= RX_IFP_MAXSIZE;
            if(ptT38Params->piReOrderBufPtr < ptT38Params->aiReOrderBuffer)
            {
               ptT38Params->piReOrderBufPtr += REORDERBUFLEN;
            }

            if( Ptr1 != ptT38Params->piReOrderBufPtr)
            {
                CopyChars((void *)Ptr1,(void *)ptT38Params->piReOrderBufPtr,RX_IFP_MAXSIZE);
            }

            break;
        }
    }

   return(FoundPkt);
}

INT16 GetEarliestSavedPkt(tFAXRELAY_ESL_T38 *ptT38Params,INT16 EarliestPkt)
{
    INT16 FoundPkt = 0,i,j,Temp=0;

    for(i=0; i < REORDERBUFLEN; i+= RX_IFP_MAXSIZE)
    {
        Temp = (ptT38Params->aiReOrderBuffer[i] << 8) &0xff00;
        Temp |= (ptT38Params->aiReOrderBuffer[i+1] & 0xff);

        if(Temp < EarliestPkt)
			EarliestPkt = Temp;
    }

   return(EarliestPkt - 1);
}
#endif


/*-------------------------------------------------------------------------

Function Name: ReorderCiscoUDPPkts 

Description :
    If no packet is received, then a packet with expected sequence number
    is serached in the reorder buffer.
    If a valid is packet is obtained, then the sequence number of that packet 
    is analyzed and the the following tasks are performed 
    * If the sequence number is same as that expected, the packet is processed.
    * If it is out of sequence, the packet is saved in reorder buffer for
      further consideration.

Input: 
 1. ptT38Params = relay channel pointer
 2. InQPtr = pointer to source buffer
 3. NewIPPkt = Flag to indicate the presence of a new packet.

Return Value:
If a valid packet is available it returns 1 else 0.
---------------------------------------------------------------------------*/
#ifdef UDP_MODE
#ifdef CISCO_GW
INT16 ReorderCiscoUDPPkts(tFAXRELAY_ESL_T38 *ptT38Params, 
                     UCHAR8 *InPtr,
                     INT16 NewIPPkt)
{
    UCHAR8 *Ptr1,*Ptr2,*TempPtr[10];
	UINT16 TempLen[10],TempPos = 0;
    UINT16 GotPkt = 0,len,PriLen,seqno,PosOfReqSecPkt,i,j;


    Ptr1 = InPtr;

//  If no packet is received, then a packet with expected sequence number
//    is searched in the reorder buffer.
    if(!NewIPPkt)
    {
        GotPkt = FindPktInBUF(ptT38Params,(ptT38Params->iPrevPktRead+1));
        return(GotPkt);
    }

    ptT38Params->iRxSeqNo = *Ptr1++;
    if(Ptr1 >= ptT38Params->JitterBuffer + ptT38Params->JitterBufLen)
        Ptr1 -= ptT38Params->JitterBufLen;

    ptT38Params->iRxSeqNo <<=8;
    ptT38Params->iRxSeqNo |= (*Ptr1++ & 0xff);
    if(Ptr1 >= ptT38Params->JitterBuffer + ptT38Params->JitterBufLen)
        Ptr1 -= ptT38Params->JitterBufLen;

	if(*Ptr1 & 0x80) {
		PriLen = (*Ptr1++ & 0x7f); 
		if(Ptr1 >= ptT38Params->JitterBuffer + ptT38Params->JitterBufLen)
			Ptr1 -= ptT38Params->JitterBufLen;

		PriLen <<= 7;
		PriLen |= *Ptr1++ & 0x7f;

		if(Ptr1 >= ptT38Params->JitterBuffer + ptT38Params->JitterBufLen)
			Ptr1 -= ptT38Params->JitterBufLen;
	}
	else
	{
		PriLen = *Ptr1++;
		if(Ptr1 >= ptT38Params->JitterBuffer + ptT38Params->JitterBufLen)
			Ptr1 -= ptT38Params->JitterBufLen;
	}

	InPtr = Ptr1;

    //BF 260804
	if(PriLen > RX_IFP_MAXSIZE) {
//#ifdef DEF_DEBUG
//		if(ptT38Params->EnableDebug)
			printf("\nRx Primary Pkt Size is out of range");
//#endif
		return(0);
	}

	
//Sequence number of received packet is same as the expected number
    if(ptT38Params->iRxSeqNo == ptT38Params->iPrevPktRead + 1)
    {

// Save the contents of the packet for further processing 
// return "1" to indicate analysis of valid packet
        for(i=0; i<PriLen ; i++) {
            ptT38Params->piIpPktPtr[i] = *Ptr1++;
            if(Ptr1 >= ptT38Params->JitterBuffer + ptT38Params->JitterBufLen)
                Ptr1 -= ptT38Params->JitterBufLen;
        }
        return(1);
    }
/*
	else
		printf("\nDidnot Rx Pkt:0x%x",(ptT38Params->iPrevPktRead+1));
*/


// If a pkt which was already processed has arrived, then discard it
    if(ptT38Params->iRxSeqNo <= ptT38Params->iPrevPktRead)
    {
        //Search for the expected packet in the reorder buffer
        GotPkt = FindPktInBUF(ptT38Params,(ptT38Params->iPrevPktRead+1));
        return(GotPkt);
    }


//To find the number of redundant packets in the received payload

//Ptr1 now points to the field having number of Secondary pkts
	Ptr1 += PriLen + 1;
	if(Ptr1 >= ptT38Params->JitterBuffer + ptT38Params->JitterBufLen)
		Ptr1 -= ptT38Params->JitterBufLen;

//Get the number of secondary packets present in the received pay load
    ptT38Params->iNoOfSecInRxPkt = *Ptr1++;
//Ptr1 now points to beginning of Sec Pkts
	if(Ptr1 >= ptT38Params->JitterBuffer + ptT38Params->JitterBufLen)
		Ptr1 -= ptT38Params->JitterBufLen;

    //BF 260804
		if(ptT38Params->iNoOfSecInRxPkt > 10) {
//#ifdef DEF_DEBUG
//			if(ptT38Params->EnableDebug)
				printf("\n Level of Redundancy is out of range");
//#endif
			return(0);
		}



//save the starting address and len of primary and all the sec pkts to arrays

	Ptr1 = InPtr;
	len = PriLen;

	for(i=0; i < ptT38Params->iNoOfSecInRxPkt+1; i++)
    {

        //BF 260804
		if(len > RX_IFP_MAXSIZE) {
//#ifdef DEF_DEBUG
//		if(ptT38Params->EnableDebug)
			printf("\n Pkt Size is out of range");
//#endif
			return(0);
		}

		TempPtr[i] = Ptr1;
		TempLen[i] = len;

        Ptr1+=len;
        if(Ptr1 >= ptT38Params->JitterBuffer + ptT38Params->JitterBufLen)
            Ptr1 -= ptT38Params->JitterBufLen;

        if(i == 0) {
            Ptr1+=2;
            if(Ptr1 >= ptT38Params->JitterBuffer + ptT38Params->JitterBufLen)
                Ptr1 -= ptT38Params->JitterBufLen;
        }

		if(*Ptr1 & 0x80) {
			len = (*Ptr1++ & 0x7f); 
			if(Ptr1 >= ptT38Params->JitterBuffer + ptT38Params->JitterBufLen)
				Ptr1 -= ptT38Params->JitterBufLen;

			len <<= 7;
			len |= *Ptr1++ & 0x7f;

			if(Ptr1 >= ptT38Params->JitterBuffer + ptT38Params->JitterBufLen)
				Ptr1 -= ptT38Params->JitterBufLen;
		}
		else {
			len = *Ptr1++;
			if(Ptr1 >= ptT38Params->JitterBuffer + ptT38Params->JitterBufLen)
				Ptr1 -= ptT38Params->JitterBufLen;
		}
	}


// Expected packet is among the redundant packets
    if(ptT38Params->iRxSeqNo <= 
        ptT38Params->iPrevPktRead + ptT38Params->iNoOfSecInRxPkt + 1)
    {
        PosOfReqSecPkt = ptT38Params->iRxSeqNo - ptT38Params->iPrevPktRead - 1;

        //BF 260804
		if(PosOfReqSecPkt > 10) {
//#ifdef DEF_DEBUG
//			if(ptT38Params->EnableDebug)
				printf("\n Level of Redundancy is out of range");
//#endif
			return(0);
		}



//Ptr1 points to the required Sec Pkt, of size 'len', Read it.
		len = TempLen[PosOfReqSecPkt];
		Ptr1 = TempPtr[PosOfReqSecPkt];

        for(i=0;i<len;i++) {
            ptT38Params->piIpPktPtr[i] = *Ptr1++;
            if(Ptr1 >= ptT38Params->JitterBuffer + ptT38Params->JitterBufLen)
                Ptr1 -= ptT38Params->JitterBufLen;
        }

//start filling the ReorderBuffer with sec pkt having lowest SeqNo (Num Read+1)
        seqno = ptT38Params->iPrevPktRead + 2;
//Stop saving if reorder buffer is full


//		while(PosOfReqSecPkt > 0 && 
//			  ptT38Params->iNoOfPktInROBuf < MAX_IFP_REORDER )
	while(PosOfReqSecPkt > 0) 
        {
			ptT38Params->iNoOfPktInROBuf++;

            Ptr2 = ptT38Params->piReOrderBufPtr;
            *Ptr2++ = (seqno>>8)&0xff;
            *Ptr2++ = seqno & 0xff;

			len = TempLen[PosOfReqSecPkt-1];
			Ptr1 = TempPtr[PosOfReqSecPkt-1];

            *Ptr2++ = len;
            for(i=0; i<len ; i++) {
                //*Ptr2++ = (UINT16)*Ptr1++;
                *Ptr2++ = *Ptr1++;
                if(Ptr1 >= ptT38Params->JitterBuffer + ptT38Params->JitterBufLen)
                    Ptr1 -= ptT38Params->JitterBufLen;
            }

            ptT38Params->piReOrderBufPtr += RX_IFP_MAXSIZE;
            if(ptT38Params->piReOrderBufPtr == 
                    ptT38Params->aiReOrderBuffer + REORDERBUFLEN )
            {
                ptT38Params->piReOrderBufPtr = ptT38Params->aiReOrderBuffer;
            }

            PosOfReqSecPkt--;
            seqno++;
        }

        return(1);
    }

// If a packet which was supposed to be received much later is received
    else if(ptT38Params->iRxSeqNo > 
            ptT38Params->iPrevPktRead + ptT38Params->iNoOfSecInRxPkt + 1)
    {
		//Search for the expected packet in the reorder buffer
        GotPkt = FindPktInBUF(ptT38Params,(ptT38Params->iPrevPktRead+1));

//Save the primary along with the redundant packets to the reorder buffer

        TempPos = ptT38Params->iNoOfSecInRxPkt;
		len = TempLen[TempPos];
		Ptr1 = TempPtr[TempPos];
		
//start from sec pkt with lowest SeqNo
        seqno = ptT38Params->iRxSeqNo - TempPos;

        for(j=0; j < ptT38Params->iNoOfSecInRxPkt+1; j++)
        {
			ptT38Params->iNoOfPktInROBuf++;
            Ptr2 = ptT38Params->piReOrderBufPtr;
            *Ptr2++ = (seqno>>8)&0xff;
            *Ptr2++ = seqno & 0xff;
            *Ptr2++ = len;
            for(i=0; i<len ; i++) {
                //*Ptr2++ = (UINT16)*Ptr1++;
                *Ptr2++ = *Ptr1++;
                if(Ptr1 >= ptT38Params->JitterBuffer + ptT38Params->JitterBufLen)
                    Ptr1 -= ptT38Params->JitterBufLen;
            }

            ptT38Params->piReOrderBufPtr += RX_IFP_MAXSIZE;
            if(ptT38Params->piReOrderBufPtr == 
                    ptT38Params->aiReOrderBuffer + REORDERBUFLEN )
            {
                ptT38Params->piReOrderBufPtr = ptT38Params->aiReOrderBuffer;
            }


#if 0 
			if(ptT38Params->iNoOfPktInROBuf >= MAX_IFP_REORDER )
			{
				if(!GotPkt) {
				//skip the expected packet
					ptT38Params->iDroppedPkt = 1;
                                        printf("\r\n PktDropped[%d] port[%d]",ptT38Params->iPrevPktRead,ptT38Params->iPortNo);
				//look for succeeding pkt
//					ptT38Params->iPrevPktRead++;

#ifdef DEF_DEBUG
				if(ptT38Params->EnableDebug)
					printf("\nDropping Pkt %d",ptT38Params->iPrevPktRead);
#endif

					ptT38Params->iPrevPktRead = 
						GetEarliestSavedPkt(ptT38Params,ptT38Params->iRxSeqNo);
				}
				break;
			}
#endif 

       //sequence number decrementing changed to incrementing 
            //seqno--;
            seqno++;
			if(TempPos > 0)
			{
				TempPos--;
				len = TempLen[TempPos];
				Ptr1 = TempPtr[TempPos];
			}
			else
				break;

        }

        return(GotPkt);
    }

  return(GotPkt);
}


#endif
#endif

/*-------------------------------------------------------------------------

Function Name: DePacketize 

Description :
If it is TCP, we pop the packet from the Input packet queue and copy it to
ptT38Params->ps_IpPktPtr, set the IpPktValid flag and update the read queue ptr.
If the Valid flag is set the packet is analyzed.

If it is UDP, we pop the packet from the Input packet queue and call the 
ReorderUDPPkt function, this function gives the primary packet which is
then analyzed. 

Input: 
 1. ptT38Params = relay channel pointer

Return Value:   None
---------------------------------------------------------------------------*/

void DePacketize(tFAXRELAY_ESL_T38 *ptT38Params, 
                 UCHAR8 *pReadPktPtr,
                 INT16 iValidIpPkt,
                 INT16 SizeOfRxPkt)
{
    INT16 i,num,len=0,num1=0;
    UCHAR8 *IpPtr;

#ifdef TCP_MODE
/* TCP mode of packet transfer */
    if(iValidIpPkt)
    {
        ptT38Params->iIpPktValid = 0;
        ptT38Params->iTCPBytesRead = 0;
        ptT38Params->iTCPBytesRead = 
            GetTCPIFPPacket(ptT38Params,pReadPktPtr,SizeOfRxPkt);

        ptT38Params->iBytesRemaining = (ptT38Params->iTCPBytesRead < SizeOfRxPkt)?1:0;
    }
#endif

#ifdef UDP_MODE

/* UDP mode of packet transfer */
/* Packets are reodered any the primary packet is extracted */
    ptT38Params->iIpPktValid = 0;
    ptT38Params->iIpPktValid = 
#ifdef CISCO_GW
		ReorderCiscoUDPPkts(ptT38Params,pReadPktPtr,iValidIpPkt);
#else
		ReorderUDPPkts(ptT38Params,pReadPktPtr,iValidIpPkt);
#endif

    if(ptT38Params->iIpPktValid)
        ptT38Params->iPrevPktRead += 1;
#endif

	IpPtr = ptT38Params->piIpPktPtr;

/* If a vaild packet if received from the IP network is sent for analysis */

//dhruva~
//    if(ptT38Params->iIpPktValid)
    if(ptT38Params->iIpPktValid && ptT38Params->IgnorePkts == 0)
//~dhruva
    {

#ifdef CISCO_GW

		AnalyzeInputPacket(ptT38Params);

        if(*(IpPtr) > 0x1e)
        {
			num1 = *(IpPtr+1);
			IpPtr += 2;

			for(i=0; i < num1 - 1 ; i++)
		    {
                num = *(IpPtr+1);
                num <<= 8;
                num += *(IpPtr+2);
                IpPtr += num + 1 + 2 + 1;
                analyze_field_data(ptT38Params,IpPtr);
			}
		}

        ptT38Params->iIpPktValid = 0;

#else
///////// OKI - TA /////////////////////
        if(*(IpPtr) > 0x1e)
        {
            for(i=0; i< *(IpPtr+1); i++)
            {
                ptT38Params->iIpPktValid = 0;
                if(i == 0)
                {
                    AnalyzeInputPacket(ptT38Params);
                }
                else
                {
                    num = *(IpPtr+3);
                    num <<= 8;
                    num += *(IpPtr+4);
                    num += 1;
                    num += 5;
                    analyze_field_data(ptT38Params,(IpPtr+num));
                }
            }
        }
        else
        {
            ptT38Params->iIpPktValid = 0;
            AnalyzeInputPacket(ptT38Params);
        }
///////// OKI - TA /////////////////////
#endif
    
	
	}

	if(ptT38Params->iImageDataTx && !ptT38Params->iHdlcDataAvl)
		AnalyzeDSImage(ptT38Params);
	else {
		ptT38Params->pTempImageW = &ptT38Params->TempImageBuf[0];
		ptT38Params->pTempImageR = &ptT38Params->TempImageBuf[0];
	}

   return;
}

#ifdef TCP_MODE
/*-------------------------------------------------------------------------

Function Name: GetTCPIFPPacket 

Description :
    To check for a valid TCPIP IFP packet in the DS Packet Queue.

Input: 
 1. ptT38Params = relay channel pointer
 2. Read Pointer to DS_Packet Queue
 3. Size of the packet received

Return Value: Number of bytes read from the DS_packet_Queue
---------------------------------------------------------------------------*/

INT16 GetTCPIFPPacket(tFAXRELAY_ESL_T38 *ptT38Params,
                      UCHAR8 *pReadPktPtr,
                      INT16 SizeOfRxPkt)
{
    UCHAR8 *InPtr;
    INT16 TCPPktSize = 0,i,TempRetVal = 0,temp;
    UCHAR8 *pTempPtr;

    InPtr = ptT38Params->piIpPktPtr;

    while(*pReadPktPtr == 0 && TempRetVal < SizeOfRxPkt)
    {
        if(++pReadPktPtr >= ptT38Params->JitterBuffer + ptT38Params->JitterBufLen)
            pReadPktPtr -= ptT38Params->JitterBufLen;
        TempRetVal++;
    }

    if(SizeOfRxPkt <= TempRetVal)
        return(TempRetVal);

    SizeOfRxPkt -= TempRetVal;
    TCPPktSize++;                           //Type Of Msg
//Check for FIELD_DATA
    temp = (pReadPktPtr[0] >> 1)&0xff;
    if( temp >= 0x60 && temp <=0x69)
    {

        pTempPtr = pReadPktPtr;

        if(SizeOfRxPkt >= TCPPktSize + 4)
        {
            pReadPktPtr++;
            if(pReadPktPtr >= ptT38Params->JitterBuffer + ptT38Params->JitterBufLen)
                pReadPktPtr -= ptT38Params->JitterBufLen;

            TCPPktSize += *pReadPktPtr++;       //Number of Data Sequences

            pReadPktPtr++;
            if(pReadPktPtr >= ptT38Params->JitterBuffer + ptT38Params->JitterBufLen)
                pReadPktPtr -= ptT38Params->JitterBufLen;

            TCPPktSize++;                       //Field Type

            TCPPktSize += (*pReadPktPtr++ << 8);
            if(pReadPktPtr >= ptT38Params->JitterBuffer + ptT38Params->JitterBufLen)
                pReadPktPtr -= ptT38Params->JitterBufLen;
            TCPPktSize += *pReadPktPtr++;
            if(pReadPktPtr >= ptT38Params->JitterBuffer + ptT38Params->JitterBufLen)
                pReadPktPtr -= ptT38Params->JitterBufLen;

            TCPPktSize+= 3;

            if( SizeOfRxPkt < TCPPktSize )
                return(TempRetVal);

            ptT38Params->iIpPktValid = 1;
            for(i=0;i<TCPPktSize;i++)
            {
                *InPtr++ = *pTempPtr++;
                if(pTempPtr >= ptT38Params->JitterBuffer + ptT38Params->JitterBufLen)
                    pTempPtr -= ptT38Params->JitterBufLen;
            }

        }
        else
        {
            return(TempRetVal);
        }
    }
//Check for INDICATOR
    else if(pReadPktPtr[0] <= 0x1e)
    {
        *InPtr++ = *pReadPktPtr++;
        if(pReadPktPtr >= ptT38Params->JitterBuffer + ptT38Params->JitterBufLen)
            pReadPktPtr -= ptT38Params->JitterBufLen;
        ptT38Params->iIpPktValid = 1;
        while(*pReadPktPtr == 0 && TCPPktSize < SizeOfRxPkt)
        {
            if(++pReadPktPtr >= ptT38Params->JitterBuffer + ptT38Params->JitterBufLen)
                pReadPktPtr -= ptT38Params->JitterBufLen;
            TCPPktSize++;                   //Discard zero stuffed bytes
        }
    }
    return(TCPPktSize+TempRetVal);
}


#endif

/*-------------------------------------------------------------------------

Function Name: AnalyzeInputPacket 

Description :
The input packet is analyzed. If it is a control packet some flags are set, if
it is a data packet the data is copied.

Input: 
 1. ptT38Params = relay channel pointer

Return Value:   None
---------------------------------------------------------------------------*/

void AnalyzeInputPacket(tFAXRELAY_ESL_T38 *ptT38Params)
{
  UCHAR8 *IpPtr1;
  UCHAR8 temp;

  IpPtr1 = ptT38Params->piIpPktPtr;
  temp = *IpPtr1;
  //printf("\r\nAnalyzeInputPacket : Header = [%x]", *IpPtr1);
  switch(*IpPtr1++)
  {
    case NOSIG:    //0 :
      break;            /*NO SIG*/

    case CNG:        //1 : 
        ptT38Params->iCng = 1 ; /*CNG*/
        ptT38Params->iCallMode = ANSWERMODE ; /*CNG*/
        printf("\r\nCNG pkt rcvd");
        break;    

    case CED:        //2 : 
        ptT38Params->iCed = 1 ; /*cED*/
        ptT38Params->iCallMode = ORIGINATEMODE;
        printf("\r\nAnalyzeInputPacket : CED pkt rcvd");
        break;

    case V21PREAMBLE:    //3 :     
        //printf("\r\nV21 Preamble pkt rcvd");
        ptT38Params->iPreambleTim = 100;  /*V21PREAMBLE*/
        if(ptT38Params->iFrameLength == FL_5MS)      //Frame length = 5ms
        {
            ptT38Params->iPreambleTim = 200;
        }
        if(ptT38Params->iFrameLength == FL_20MS)      //Frame length = 20ms
        {
            ptT38Params->iPreambleTim = 50;
        }
        ptT38Params->iPreambleTxFlg = 1;
#ifdef T30_T38
        ptT38Params->iPreambleTim = 1;
#endif

        break;

    case V272400TRN:    //4 :             /*V272400TRN*/
    case V274800TRN:    //5 :             /*V274800TRN*/
    case V297200TRN:    //6 :             /*V297200TRN*/
    case V299600TRN:    //7 :             /*V299600TRN*/
    case V177200STTRN:    //8 :             /*V177200STTRN*/
    case V177200LTTRN:    //9 :             /*V177200LTTRN*/
    case V179600STTRN:    //10 :            /*V179600STTRN*/
    case V179600LTTRN:    //11 :            /*V179600LTTRN*/
    case V1712000STTRN:    //12 :            /*V1712000STTRN*/
    case V1712000LTTRN:    //13 :            /*V1712000LTTRN*/
    case V1714400STTRN:    //14 :            /*V1714400STTRN*/
    case V1714400LTTRN:    //15 :            /*V1714400LTTRN*/
          printf("\r\nTRN:%d",temp);
          ptT38Params->iTrnTxFlg = 1;
          //ptT38Params->s_HdlcHSEnableTx = 1;
          ptT38Params->pTempImageW = &ptT38Params->TempImageBuf[0];
          ptT38Params->pTempImageR = &ptT38Params->TempImageBuf[0];
/*dhruva, changed on 19/dec/02
#ifdef T30_T38
		ptT38Params->piFrameDataPtr = ptT38Params->piDataBufPtr;
		ptT38Params->piReadDataPtr = ptT38Params->piDataBufPtr;
#endif
*/
#ifdef DEF_DEBUG
		if(ptT38Params->EnableDebug)
			printf("\n T.38: Rx TRN_IND packet");
#endif
		break;

    case V21C2:        //16 :            /*V2100*/ 
    case V27_2400:    //17 :            /*V27_2400*/    
    case V27_4800:    //18 :            /*V27_4800*/    
    case V29_7200:    //19 :            /*V29_7200*/    
    case V29_9600:    //20 :            /*V29_9600*/    
    case V17_7200:    //21 :            /*V17_7200*/    
    case V17_9600:    //22 :            /*V17_9600*/    
    case V17_12000:    //23 :            /*V17_12000*/    
    case V17_14400:    //24 :             /*V17_14400*/    
        IpPtr1++;        //to discard the number of pkts field
        analyze_field_data(ptT38Params,IpPtr1);
        break;
  }
  return;
}

/*-------------------------------------------------------------------------

Function Name: analyze_field_data 

Description :
    The data feild of the input packet is analyzed.

Input: 
 1. ptT38Params = relay channel pointer
 2. IpPtr2 = pointer to input packet buffer

Return Value:   None
---------------------------------------------------------------------------*/

void analyze_field_data(tFAXRELAY_ESL_T38 *ptT38Params, UCHAR8 *IpPtr2)
{
    switch(*IpPtr2++)
    {
       case HDLCDATA :  //1
            if(!ptT38Params->iDisablePktData) {
                ptT38Params->iHdlcDataAvl = 1;
                //printf("\r\n iHdlcDataAvl on port[%d]",ptT38Params->iPortNo);
                CopyDataToBeTx(ptT38Params,IpPtr2);     /*HDLCDATA */
            }
            break;

        case HDLCSIGEND :      //Indicates signal end    //2
            ptT38Params->iSigEnd = 1;
            break;

        case HDLCFCSOK :        //3
            if(++ptT38Params->iSizeWIndex == SIZEBUFSIZE)
                ptT38Params->iSizeWIndex = 0;
            ptT38Params->aiSizeBuf[ptT38Params->iSizeWIndex] = 0;

//Good CRC
            ptT38Params->aiCRCBuf[ptT38Params->iCRCWIndex] = 1;

			if(ptT38Params->iDroppedPkt) {
				ptT38Params->aiCRCBuf[ptT38Params->iCRCWIndex] = 2;
				ptT38Params->iDroppedPkt = 0;
			}

            if(++ptT38Params->iCRCWIndex == CRCBUFSIZE)
                ptT38Params->iCRCWIndex = 0;

            //printf("\r\n HDLCFCSOK : iCRCWIndex[%x]port[%d]",ptT38Params->iCRCWIndex,ptT38Params->iPortNo);

            break;

        case HDLCFCSBAD :    //4
            if(++ptT38Params->iSizeWIndex == SIZEBUFSIZE)
                ptT38Params->iSizeWIndex = 0;
            ptT38Params->aiSizeBuf[ptT38Params->iSizeWIndex] = 0;

//Bad CRC
            ptT38Params->aiCRCBuf[ptT38Params->iCRCWIndex] = 2;
            if(++ptT38Params->iCRCWIndex == CRCBUFSIZE)
                ptT38Params->iCRCWIndex = 0;

            //printf("\r\n HDLCFCSBAD : iCRCWIndex[%x]port[%d]",ptT38Params->iCRCWIndex,ptT38Params->iPortNo);

            break;

        case HDLCFCSOKSIGEND :        //5
            ptT38Params->iSigEnd = 1;

	    if(++ptT38Params->iSizeWIndex == SIZEBUFSIZE)
                ptT38Params->iSizeWIndex = 0;
            ptT38Params->aiSizeBuf[ptT38Params->iSizeWIndex] = 0;

//Good CRC SIG END
            ptT38Params->aiCRCBuf[ptT38Params->iCRCWIndex] = 3;
			if(ptT38Params->iDroppedPkt) {
				ptT38Params->aiCRCBuf[ptT38Params->iCRCWIndex] = 4;
				ptT38Params->iDroppedPkt = 0;
			}

            if(++ptT38Params->iCRCWIndex == CRCBUFSIZE)
                ptT38Params->iCRCWIndex = 0;

            //printf("\r\n HDLCFCSOKSIGEND : iCRCWIndex[%x]port[%d]",ptT38Params->iCRCWIndex,ptT38Params->iPortNo);
            break;

        case HDLCFCSBADSIGEND :        //6
            ptT38Params->iSigEnd = 1;

            if(++ptT38Params->iSizeWIndex == SIZEBUFSIZE)
                ptT38Params->iSizeWIndex = 0;
            ptT38Params->aiSizeBuf[ptT38Params->iSizeWIndex] = 0;

//Bad CRC SIG END
            ptT38Params->aiCRCBuf[ptT38Params->iCRCWIndex] = 4;
            if(++ptT38Params->iCRCWIndex == CRCBUFSIZE)
                ptT38Params->iCRCWIndex = 0;
           
            //printf("\r\n HDLCFCSBADSIGEND : iCRCWIndex[%x]port[%d]",ptT38Params->iCRCWIndex,ptT38Params->iPortNo);

	     break;

        case T4NONECMDATA :            //7
			CopyDataToBeTx(ptT38Params,IpPtr2);     /*T4NONECMDATA*/
            break;

        case T4NONECMSIGEND :            // T4NONECMSIGEND
        case T4NONECMDATASIGEND :        // T4NONECMSIGEND 
            ptT38Params->iSigEndHS = 1;
            printf("\r\n analyze_field_data : T4 SIG END Rcvd");
//If DS HS pkts were skiped, reset the Pkt dropped flag 
			ptT38Params->iDroppedPkt = 0;

/*            if(ptT38Params->iImageDataTx)
            {
                if(++ptT38Params->iSizeWIndex == SIZEBUFSIZE)
                    ptT38Params->iSizeWIndex = 0;
                ptT38Params->aiSizeBuf[ptT38Params->iSizeWIndex] = 0;
            }
*/
#ifdef DEF_DEBUG
			if(ptT38Params->EnableDebug)
				printf("\n T.38: Rx T4_NONECM_SIGEND_IND packet");
#endif
            break;

    }

  return;
}


/*-------------------------------------------------------------------------

Function Name: ConfigDpInHSLTrnModeTx 

Description :
The data pump is configured in high speed Tx mode with long train, in case of
V.17

Input: 
 1. ptT38Params = relay channel pointer

Return Value:
---------------------------------------------------------------------------*/
void ConfigDpInHSLTrnModeTx(tFAXRELAY_ESL_T38 *ptT38Params){
    ConfigDpInHSModeTx(ptT38Params,ptT38Params->idRateTx+1);
    return;
}

/*-------------------------------------------------------------------------

Function Name: ConfigDpInHSSTrnModeTx 

Description :
The data pump is configured in high speed Tx mode with short train, in case of
V.17

Input: 
 1. ptT38Params = relay channel pointer

Return Value:   None
---------------------------------------------------------------------------*/
void ConfigDpInHSSTrnModeTx(tFAXRELAY_ESL_T38 *ptT38Params){
    printf("\r\nConfigDpInHSModeTx called");
    ConfigDpInHSModeTx(ptT38Params,ptT38Params->idRateTx);
    return;
}

/*-------------------------------------------------------------------------

Function Name: ConfigDpInHSModeTx 

Description :
We first select the datapump's mode as MSB to LSB. This is because the data pump
works in little endian mode and the T.38 is big endian. By selecting this mode
we can expect the datapump to convert the data from big endian to little endian.
Then we set the datapump in HS Tx with the given dRate. The state fn ptr is set
to TxHSDataToDp. We also initialize the variables used by T.38 to transmit high
speed data. 

Input: 
ptT38Params = relay channel pointer
dRate = rate at which the DP will be configured

Return Value:   None
---------------------------------------------------------------------------*/

void ConfigDpInHSModeTx(tFAXRELAY_ESL_T38 *ptT38Params,INT16 dRate){

/* Message to configure the DP in MSB to LSB mode of data transfer */
    SendTxRxMsgToDp(ptT38Params,mDP_DATAMODE_SELECT,DATA_M2L);

/* Message to configure the DP in fax transmision mode at the given rate  */
    printf("\r\n mDP_START_FAXTX port[%d]",ptT38Params->iPortNo);
    SendTxRxMsgToDp(ptT38Params,mDP_START_FAXTX,dRate);
    
/* The state machine is set for transmitting high speed data to DP */
    ptT38Params->fpStateFnPtr = (fpPtrIntFn)(TxHSDataToDp);
//    ptT38Params->s_SigEnd = 0;
    InitHdlcTx(ptT38Params);
//	printf("\n T.38: TxHSDataToDp");

    return;
}

/*-------------------------------------------------------------------------

Function Name: ConfigDpInHdlcModeTx 

Description :
We first select the datapump's mode as HDLC to enable it to transmit Hdlc data.
Then we set the datapump in Hdlc Tx with the given dRate. We also initialize 
the variables used by T.38 to transmit Hdlc data. 

Input: 
ptT38Params = relay channel pointer
dRate = rate at which the DP will be configured

Return Value:   None
---------------------------------------------------------------------------*/

void ConfigDpInHdlcModeTx(tFAXRELAY_ESL_T38 *ptT38Params,INT16 dRate){
/* Message to configure the DP in HDLC mode of data transfer */
    SendTxRxMsgToDp(ptT38Params,mDP_DATAMODE_SELECT,DATA_HDLC);
/* Message to configure the DP in fax transmision mode at the given rate  */
    printf("\r\n mDP_START_FAXTX V21port[%d]",ptT38Params->iPortNo);
    SendTxRxMsgToDp(ptT38Params,mDP_START_FAXTX,dRate);
//    ptT38Params->s_SigEnd =0;
    InitHdlcTx(ptT38Params);
    ptT38Params->iConnected = 0;

    return;
}

/*-------------------------------------------------------------------------

Function Name: ConfigDpInHdlcModeRx 

Description :
We first select the datapump's mode as HDLC to enable it to receive Hdlc data.
Then we set the datapump in Hdlc Rx with the given dRate. We also initialize 
the variables used by T.38 to receive Hdlc data. 

Input: 
ptT38Params = relay channel pointer
dRate = rate at which the DP will be configured

Return Value:
None
---------------------------------------------------------------------------*/


void ConfigDpInHdlcModeRx(tFAXRELAY_ESL_T38 *ptT38Params,INT16 dRate){
    printf("\r\n ConfigDpInHdlcModeRx: mDP_START_FAXRX V21 port[%d]",ptT38Params->iPortNo);

    //for logging V.21 packets received from the Fax machine
    //for debugging CRC ERROR case
    debug_fax = 300;

/* Message to configure the DP in HDLC mode of data transfer */
    SendTxRxMsgToDp(ptT38Params,mDP_DATAMODE_SELECT,DATA_HDLC);
/* Message to configure the DP in fax reception mode at the given rate  */
    SendTxRxMsgToDp(ptT38Params,mDP_START_FAXRX,dRate);
//    InitHdlcRx();
////dhruva

    ptT38Params->iTrnRxFlg = 0;
//    ptT38Params->iNoCarStat = 0;
    ptT38Params->iNumBytes = 0;
    ptT38Params->iPackedBytes = 0;
    ptT38Params->iByteCnt = 0;
    ptT38Params->iZeroesInTCF = 0;

///////


    return;
}

/*-------------------------------------------------------------------------

Function Name: ConfigDpInHSLTrnModeRx 

Description :
We first select the datapump's mode as MSB to LSB. This is because the data pump
works in little endian mode and the T.38 is big endian. By selecting this mode
we can expect the datapump to convert the data from big endian to little endian.
Then we set the datapump mode to HS Long Trn Rx with the given dRate. 
We also initialize the variables used by T.38 to receive high speed data. 

Input: 
ptT38Params = relay channel pointer

Return Value:
None
---------------------------------------------------------------------------*/

void ConfigDpInHSLTrnModeRx(tFAXRELAY_ESL_T38 *ptT38Params){

/* Message to configure the DP in MSB to LSB mode of data transfer */
    SendTxRxMsgToDp(ptT38Params,mDP_DATAMODE_SELECT,DATA_M2L);
    InitHighSpeedRx(ptT38Params);
/* Message to configure the DP in fax reception mode at the given rate  */
    SendTxRxMsgToDp(ptT38Params,mDP_START_FAXRX,ptT38Params->idRateRx+1);

//////////dhruva
    ptT38Params->iTrnRxFlg = 0;
//    ptT38Params->iNoCarStat = 0;
    ptT38Params->iNumBytes = 0;
    ptT38Params->iPackedBytes = 0;
    ptT38Params->iByteCnt = 0;
    ptT38Params->iZeroesInTCF = 0;

////
    return;
}

/*-------------------------------------------------------------------------

Function Name: ConfigDpInHSSTrnModeRx 

Description :
We first select the datapump's mode as MSB to LSB. This is because the data pump
works in little endian mode and the T.38 is in big endian. By selecting this 
mode we can expect the datapump to convert the data from big endian to little 
endian.  Then we set the datapump mode to HS short Trn Rx with the given dRate. 
We also initialize the variables used by T.38 to receive high speed data. 

Input: 
ptT38Params = relay channel pointer

Return Value:
None
---------------------------------------------------------------------------*/


void ConfigDpInHSSTrnModeRx(tFAXRELAY_ESL_T38 *ptT38Params){

/* Message to configure the DP in MSB to LSB mode of data transfer */
    SendTxRxMsgToDp(ptT38Params,mDP_DATAMODE_SELECT,DATA_M2L);
    InitHighSpeedRx(ptT38Params);
/* Message to configure the DP in fax reception mode at the given rate  */
    SendTxRxMsgToDp(ptT38Params,mDP_START_FAXRX,ptT38Params->idRateRx);
//////////dhruva
    ptT38Params->iTrnRxFlg = 0;
    //ptT38Params->iNoCarStat = 0;
    ptT38Params->iNumBytes = 0;
    ptT38Params->iPackedBytes = 0;
    ptT38Params->iByteCnt = 0;
    ptT38Params->iZeroesInTCF = 0;

////
    return;
}


/*-------------------------------------------------------------------------

Function Name: InitHighSpeedRx 

Description :
We initialize the variables used by T.38 to receive high speed data. 

Input: 
ptT38Params = relay channel pointer

Return Value:
None
---------------------------------------------------------------------------*/

void InitHighSpeedRx(tFAXRELAY_ESL_T38 *ptT38Params){
/* Initialize the data buffer read and write pointers to the lower
   boundary of the data buffer */
    ptT38Params->piImageDataPtr =ptT38Params->piDataBufPtr;
    ptT38Params->piReadDataPtr =ptT38Params->piDataBufPtr;
/* Clear the flags and ciunters */
    ptT38Params->iTrnRxFlg = 0;
    ptT38Params->iNoCarStat = 0;
    ptT38Params->iNumBytes = 0;
    ptT38Params->iPackedBytes = 0;
    ptT38Params->iByteCnt = 0;
    ptT38Params->iZeroesInTCF = 0;
            
    return;
}

/*-------------------------------------------------------------------------

Function Name: RecvTrn 

Description :
We wait till we receive Trn message from datapump, and then pack the Trn 
Indicator packet. The state fn ptr is set to PackHSDataPkt, 
since the high speed data is what will follow. This is in non ecm mode. 
In the Ecm mode we set the state fn ptr to PackHdlcDataPkt.

Input: 
ptT38Params = relay channel pointer

Return Value:
None
---------------------------------------------------------------------------*/

void RecvTrn(tFAXRELAY_ESL_T38 *ptT38Params){

/*State machine waits till it receives Trn message from datapump */
    if(ptT38Params->iTrnRxFlg)
     {
        ptT38Params->iTrnRxFlg = 0;

        if(ptT38Params->iEcm)
        {
            ptT38Params->fpStateFnPtr = (fpPtrIntFn)(PackHdlcDataPkt);
        }

        else
        {
            ptT38Params->fpStateFnPtr = (fpPtrIntFn)(PackHSDataPkt);
        }
/* Packetize Traning indicator packet */
        T38PktT30Ind(ptT38Params,(UCHAR8)ptT38Params->iIndPktdRateRx);
		ptT38Params->iCurrentRedundancy = ptT38Params->iHSRedundancy;
		ptT38Params->iOpPktValid = 3;//dhruva
     }

    return;
}

/*-------------------------------------------------------------------------

Function Name: RecvLTCF 

Description :
We wait till we receive Trn message from datapump , then pack the Trn 
Indicator packet. The state fn ptr is set to PackHSDataPkt, since the TCF 
data is what will follow.

Input: 
ptT38Params = relay channel pointer

Return Value:
None
---------------------------------------------------------------------------*/

void RecvLTCF(tFAXRELAY_ESL_T38 *ptT38Params){
/*State machine waits till it receives Trn message from datapump */
    if(ptT38Params->iTrnRxFlg)
     {
        ptT38Params->iTrnRxFlg = 0;
#ifdef UDP_MODE
            ptT38Params->fpStateFnPtr = (fpPtrIntFn)(PackHSDataPkt);
#endif
#ifdef TCP_MODE
            ptT38Params->fpStateFnPtr = (fpPtrIntFn)(WaitForEndOfHSData);
            ptT38Params->piFrameDataPtr = ptT38Params->piReadDataPtr = ptT38Params->piDataBufPtr;
#endif
/* Packetize Traning indicator packet */
        T38PktT30Ind(ptT38Params,(UCHAR8)ptT38Params->iIndPktdRateRx+2);
		ptT38Params->iCurrentRedundancy = ptT38Params->iHSRedundancy;
		ptT38Params->iOpPktValid = 3;//dhruva
     }

    return;
}

/*-------------------------------------------------------------------------

Function Name: RecvTCF 

Description :
We wait till we receive Trn message from datapump , then pack the Trn 
Indicator packet. The state fn ptr is set to PackHSDataPkt, since the TCF 
data is what will follow.

Input: 
ptT38Params = relay channel pointer

Return Value:
None
---------------------------------------------------------------------------*/

void RecvTCF(tFAXRELAY_ESL_T38 *ptT38Params){
/*State machine waits till it receives Trn message from datapump */
    if(ptT38Params->iTrnRxFlg)
    {
        ptT38Params->iTrnRxFlg = 0;
#ifdef UDP_MODE
        ptT38Params->fpStateFnPtr = (fpPtrIntFn)(PackHSDataPkt);
#endif
#ifdef TCP_MODE
        ptT38Params->fpStateFnPtr = (fpPtrIntFn)(WaitForEndOfHSData);
        ptT38Params->piFrameDataPtr = ptT38Params->piReadDataPtr = ptT38Params->piDataBufPtr;
#endif
/* Packetize Traning indicator packet */
        T38PktT30Ind(ptT38Params,(UCHAR8)ptT38Params->iIndPktdRateRx);
		ptT38Params->iCurrentRedundancy = ptT38Params->iHSRedundancy;
		ptT38Params->iOpPktValid = 3;//dhruva
    }
    return;
}


/*---------------------------------------------------------------------

Function Name: WaitForEndOfHSData 

Description :
    The state machine waits till no carrier message is received from the DP
 Then the number of zeros in TCF is evaluated and the local response is set.
 The DP and the relay are set to the default mode of Low speed reception.
Input: 
ptT38Params = relay channel pointer

Return Value:
None
---------------------------------------------------------------------------*/

void WaitForEndOfHSData(tFAXRELAY_ESL_T38 *ptT38Params){


/*The state machine waits till no carrier message is received from the DP*/
    if(ptT38Params->iNoCarStat)
    {
        ptT38Params->iNoCarStat = 0;

/* The DP and the relay are set to the default mode of Low speed reception */
//        SetDpModeHdlcLSRx(ptT38Params);
//        InitHdlcRx(ptT38Params);

//////////////
            /* A timer is set and state machine is set wait for spoofing */
        ptT38Params->iSpoofStartTimer = 270;

        if(ptT38Params->iFrameLength == FL_5MS)      //Frame length = 5ms
        {
            ptT38Params->iSpoofStartTimer = 540;
        }
        if(ptT38Params->iFrameLength == FL_20MS)      //Frame length = 20ms
        {
            ptT38Params->iSpoofStartTimer = 135;
        }

        WaitForCFR_FTT(ptT38Params);
///////////////
    }
    return;
}

/*-------------------------------------------------------------------------

Function Name: PackHdlcDataPkt 

Description : In this function the pktlen of Hdlc pkt is decided.
Before packing the data, Byte Count is checked if it is Pktlen + 2 , 
the packetizing function is called, else we check if it is Hdlc frame end, if it
is, the bytes which are remaining are packed , else we check if the nocar flag 
is set, if yes then we make CRCStauts = 4, i.e. CRCBAD and pack the remaining 
bytes. The Hdlc frame is processed only if CRCStatus is OK. 

Input: 
ptT38Params = relay channel pointer

Return Value:
None
---------------------------------------------------------------------------*/

void PackHdlcDataPkt(tFAXRELAY_ESL_T38 *ptT38Params){

    UCHAR8 TypeOfMsg,status;

//dhruva
    if(ptT38Params->iResendPktCnt > 0) {
		ptT38Params->iResendPktCnt--;
		T38PktT30Ind(ptT38Params,V21PREAMBLE);
		if(++ptT38Params->iCurrentRedundancy >= ptT38Params->iLSRedundancy)
			ptT38Params->iCurrentRedundancy = ptT38Params->iLSRedundancy;
		return;
	}

	TypeOfMsg = 
        (ptT38Params->iHdlcHSEnableRx)? ptT38Params->iPktdRateRx : V21C2;

        //BF 260804
    	//Dhruva 040804
    	ptT38Params->iByteCnt = ptT38Params->piFrameDataPtr - ptT38Params->piReadDataPtr;
    	if(ptT38Params->iByteCnt < 0)
    		ptT38Params->iByteCnt += ptT38Params->iSizeOfDataBuf;
        
/* On receiving indication of end of HDLC frame from the DP, packetize the
   remaining elemnts (if any) along with the CRCStatus */
    if(ptT38Params->iCRCStatusRx)
    {
/* HDLC data Bytes remaining  ? */
        if(ptT38Params->iByteCnt > 2)
         {
            ptT38Params->iNumBytes = ptT38Params->iByteCnt-2;
#ifdef T30_T38
            ptT38Params->iNumBytes = ptT38Params->iByteCnt;
#endif
            
			ptT38Params->iByteCnt = 0;

            ptT38Params->iHdlcFrameEnd = 
                    (ptT38Params->piDataBufPtr[1] & 8 ) >> 3;
            if(ptT38Params->iHdlcFrameEnd)
            {
                if(ptT38Params->iCRCStatusRx == 1)
                    status = HDLCFCSOKSIGEND;
                else
                    status = HDLCFCSBADSIGEND;
            }
            else
            {
                if(ptT38Params->iCRCStatusRx == 1)
                    status = HDLCFCSOK;
                else
                    status = HDLCFCSBAD;
            }
/* Packetize remaining bytes */
            T38PktT30Data(ptT38Params,TypeOfMsg, HDLCDATA, 2, status,ptT38Params->iCurrentRedundancy);

#if 1 //added on 160804 for CRC Error Fixes
            ptT38Params->iCRCStatusRx = 0;
#endif

			if(++ptT38Params->iCurrentRedundancy >= ptT38Params->iLSRedundancy)
				ptT38Params->iCurrentRedundancy = ptT38Params->iLSRedundancy;

#ifndef T30_T38
// To discard the 2 CRC bytes sent by the DP
            if(++ptT38Params->piReadDataPtr == 
                       ptT38Params->piDataBufPtr + ptT38Params->iSizeOfDataBuf)
                ptT38Params->piReadDataPtr = ptT38Params->piDataBufPtr;
            if(++ptT38Params->piReadDataPtr == 
                       ptT38Params->piDataBufPtr + ptT38Params->iSizeOfDataBuf)
                ptT38Params->piReadDataPtr = ptT38Params->piDataBufPtr;

#endif
            if(ptT38Params->piFrameDataPtr == ptT38Params->piReadDataPtr)
            {
               ptT38Params->piFrameDataPtr = 
               ptT38Params->piReadDataPtr = ptT38Params->piDataBufPtr;
            }
         }
        if(!ptT38Params->iHdlcHSEnableRx) 
            ptT38Params->fpStateFnPtr = (fpPtrIntFn)(ProcHdlcLSDataRx);
        else
            ptT38Params->fpStateFnPtr = (fpPtrIntFn)(ProcHdlcHSDataRx);

    }

/* When "No carrier state" is indicated by the DP, an indicator packet
   "SigEnd" is generated */

#if 0	//dhruva
    else if(ptT38Params->iNoCarStat)
     {
        ptT38Params->iNoCarStat = 0;
        ptT38Params->iCRCStatusRx = 4;
        ptT38Params->fpStateFnPtr= (fpPtrIntFn)(ProcErrInHdlcDataRx);
        if(ptT38Params->iByteCnt)
          ptT38Params->iNumBytes = ptT38Params->iByteCnt;
        ptT38Params->iByteCnt = 0;
        T38PktT30Data(ptT38Params,TypeOfMsg, HDLCDATA,1,0,ptT38Params->iCurrentRedundancy);
		if(++ptT38Params->iCurrentRedundancy >= ptT38Params->iLSRedundancy)
			ptT38Params->iCurrentRedundancy = ptT38Params->iLSRedundancy;
     }
#endif
    return;
}


/*-------------------------------------------------------------------------

Function Name: T38PktT30Ind 

Description :
We pack Indicator packets. If it is TCP, we pack the current packet, if it
is UDP, we pack the previous packets along with the current packet. How many
previous packets to pack depends on the user dependent parameter 
NumOfRedundantPackets. 

TCP PACKET :
 ________________________________
|TypeOfMessage    |    |    |    |
| (T30_INDICATOR) |  0 | 0  | 0  |
|_________________|____|____|____|

UDP packet :
______________________________________________________~_____________
|SequenceNumber | TypeOfMessage | FieldType | Length | RedundantPkt |
|               |(T30_INDICATOR)|    0      |    0   |              |
|_______________|_______________|___________|________|~_____________|


Input: 
ptT38Params = relay channel pointer
TypeOfMsg = T30_INDICATOR according to T.38 recommendation

Return Value:   None
---------------------------------------------------------------------------*/


void T38PktT30Ind(tFAXRELAY_ESL_T38 *ptT38Params,UCHAR8 TypeOfMsg){
    UCHAR8 *Ptr;
    INT16 i;
#ifdef UDP_MODE
    INT16 x1;
	UINT16 Offset = 0,NumOfSecPkt=0,PktSize=0;

	NumOfSecPkt = ptT38Params->iCurrentRedundancy;
#endif

	ptT38Params->TxPayloadSize = 0;

#ifdef TCP_MODE
/* TCP mode of packet transfer */
    ptT38Params->iOpPktValid = 1;    //flag to indicate that a 
                          //packet is ready for transmision
    Ptr = ptT38Params->paiOpPktPtr;    
    *Ptr++ = TypeOfMsg ;        //The indicator field is placed in
                    //the packet
    for(i=0; i<TX_IFP_MAXSIZE-1; i++)
        *Ptr++ = 0;
#endif

#ifdef UDP_MODE
/* UDP mode of packet transfer */
    
    ptT38Params->iOpPktValid = 1;    //flag to indicate that a 
                          //packet is ready for transmision

    /*Copying of previous packets. These are redundant packets to be sent 
      in accrodance with T.38 */

    if(NumOfSecPkt > 0)
    {
#ifdef CISCO_GW
        GetSecondaryPkt(ptT38Params,(ptT38Params->paiOpPktPtr + 4),NumOfSecPkt);
#else
        GetSecondaryPkt(ptT38Params,(ptT38Params->paiOpPktPtr + 3),NumOfSecPkt);
#endif

    }

    /* Formation of current packet */

    Ptr = ptT38Params->paiOpPktPtr;    

    x1 = ptT38Params->iSequenceNumber;
    *Ptr++ = (UCHAR8)((x1 >> 8) & 0xff);
    *Ptr++ = (UCHAR8)((ptT38Params->iSequenceNumber) & 0xff);

    ptT38Params->iSequenceNumber++;

	Offset = 2;
	PktSize = 1;

#ifdef CISCO_GW
    *Ptr++ = 1;
//	Offset = 3;
	PktSize = 2;
#endif
	
	*Ptr++ = (UCHAR8)TypeOfMsg ;        //The indicator field is placed in
                                        //the packet
//    if(TypeOfMsg == CNG || TypeOfMsg == CED)

#if 1 
      if(TypeOfMsg == CED)
      {
       printf("\r\n T38PktT30Ind : packetize CED pkt");
      }
#endif
	if(!NumOfSecPkt)
    {
        *Ptr++ = 0;
        *Ptr++ = 0;
		ptT38Params->TxPayloadSize += 2;
    }

	ptT38Params->TxPayloadSize += (PktSize + Offset);


    UpDatePrevPktBuf(ptT38Params,PktSize,Offset);


#endif

 return;
}


/*-------------------------------------------------------------------------

Function Name: T38PktT30Data 

Description : In this function depending on mode of pkt transport,i.e, whether
UDP or TCP the packet is formed. If it is:
 1. TCP : The current information like Type Of Msg, Field Type, Field Len and
          Field Data is filled.
 2. UDP : The current information is filled plus the redundant packet 
          information is filled.

TCP PACKET :
 ___________________________________________
|TypeOfMessage  | FieldType | Length |FIELD |
|(T30_DATA TYPE)|           |        |DATA  |
|_______________|___________|________|______|

UDP packet :
_____________________________________________________________~______________
|SequenceNumber | TypeOfMessage | FieldType | Length |FIELD | RedundantPkt |
|               |(T30_DATA TYPE)|           |        |DATA  |              |
|_______________|_______________|___________|________|______|~_____________|



Input: 
 1. ptT38Params = relay channel pointer
 2. TypeOfMsg
 3. FieldType

Return Value:
None
---------------------------------------------------------------------------*/

void T38PktT30Data(tFAXRELAY_ESL_T38 *ptT38Params,
                   UCHAR8 TypeOfMsg,
                   UCHAR8 FieldType, 
                   UCHAR8 numofpkts, 
                   UCHAR8 istatus,
				   INT16 NumOfSecPkt)
{
     UCHAR8 *Ptr;
     UCHAR8 *v21PktPtr;
     INT16 i;
#ifdef UDP_MODE
    INT16 x1;
	UINT16 tempsize=0,Offset = 0;
#endif

	ptT38Params->TxPayloadSize = 0;

#ifdef TCP_MODE
/* TCP mode of packet transfer */

//flag to indicate that a packet is ready for transmision
    ptT38Params->iOpPktValid = 1;
    Ptr = ptT38Params->paiOpPktPtr;
//The indicator field is placed in the packet
    *Ptr++ = TypeOfMsg ;

//In cases of HDLC data, T4 NonECM data, T4 Non ECM SigEnd, the packet contains
//the data elements also 
    if(FieldType==HDLCDATA || FieldType==T4NONECMDATA )    
    {
        if(ptT38Params->iNumBytes == 0)
        {
            *Ptr++ = 1;                 //num of primary pkts
            *Ptr++ = istatus;
            ptT38Params->iCRCStatusRx = 0;

            for(i=0; i<TX_IFP_MAXSIZE-3; i++)
                *Ptr++ = 0;
        }
        else
        {
            *Ptr++ = numofpkts;                 //num of primary pkts
            *Ptr++ = FieldType;
            *Ptr++ = 0;                         //Higher Word Of Length
            *Ptr++ = ptT38Params->iNumBytes - 1;//lower Word Of Length
            for(i=0 ; i < ptT38Params->iNumBytes ; i++)
            {
                *Ptr++ = *ptT38Params->piReadDataPtr;
                if(++ptT38Params->piReadDataPtr == ptT38Params->piDataBufPtr + ptT38Params->iSizeOfDataBuf)
                    ptT38Params->piReadDataPtr = ptT38Params->piDataBufPtr;
            }    
            if(numofpkts == 2)
            {
                *Ptr++ = istatus;
                ptT38Params->iCRCStatusRx = 0;
            }
            ptT38Params->iNumBytes+=numofpkts;
            ptT38Params->iNumBytes+=4;
            while(ptT38Params->iNumBytes < TX_IFP_MAXSIZE)
            {
                *Ptr++ = 0;
                ptT38Params->iNumBytes++;
            }
        }
    }
    else
        *Ptr++ = 0;

#endif

#ifdef UDP_MODE
/* UDP mode of packet transfer */
    ptT38Params->iOpPktValid = 1;    //flag to indicate that a 
                                     //packet is ready for transmision

    /*Copying of previous packets. These are redundant packets to be sent 
      in accordance with T.38 */

    Ptr = ptT38Params->paiOpPktPtr + (numofpkts - 1);
    if(ptT38Params->iNumBytes)
        Ptr += ptT38Params->iNumBytes+7;
    else
        Ptr = ptT38Params->paiOpPktPtr + 5; //dhr was 4

#ifdef CISCO_GW
	if((ptT38Params->iNumBytes + numofpkts + 4) <= 127) 
		Ptr += 1;
    else
		Ptr += 2;
#endif

	ptT38Params->TxPayloadSize = Ptr - ptT38Params->paiOpPktPtr;

    if(NumOfSecPkt > 0)
    {
        GetSecondaryPkt(ptT38Params,Ptr,NumOfSecPkt);
    }

//////////

    /* Formation of current packet ( Primary packet )*/

    Ptr = ptT38Params->paiOpPktPtr;
    v21PktPtr = Ptr;

    x1 = ptT38Params->iSequenceNumber;
    *Ptr++ = (UCHAR8)((x1 >> 8) & 0xff);
    *Ptr++ = (UCHAR8)(ptT38Params->iSequenceNumber & 0xff);
    ptT38Params->iSequenceNumber++;

	Offset = 2;

#ifdef CISCO_GW
    if(ptT38Params->iNumBytes == 0) {
	    *Ptr++ = 3;
//		Offset = 3;
	}
	else {
		tempsize = ptT38Params->iNumBytes + numofpkts + 4;
		if(tempsize <= 127) {
			*Ptr++ = tempsize & 0x7f;
//			Offset = 3;
		}
		else if (tempsize > 127) {
			*Ptr++ = ((tempsize >> 7) & 0x7f) | 0x80;
			*Ptr++ = tempsize & 0x7f;
//			Offset = 4;
		}
	}
#endif

    
	*Ptr++ = TypeOfMsg ;
//    *Ptr++ = FieldType;


//In cases of HDLC data, T4 NonECM data, T4 Non ECM SigEnd, the packet contains
//the data elements also 
    if(FieldType==HDLCDATA || FieldType==T4NONECMDATA )    
    {
        if(ptT38Params->iNumBytes == 0)
        {
            *Ptr++ = 1;                 //num of primary pkts
            *Ptr++ = istatus;
            ptT38Params->iCRCStatusRx = 0;
        }
        else
        {
            *Ptr++ = numofpkts;                 //num of primary pkts
            *Ptr++ = FieldType;
            *Ptr++ = 0;                         //Higher Word Of Length
            *Ptr++ = ptT38Params->iNumBytes - 1;//Lower Word Of Length
            for(i=0 ; i < ptT38Params->iNumBytes ; i++)
            {
                *Ptr++ = *ptT38Params->piReadDataPtr;
                if(++ptT38Params->piReadDataPtr == ptT38Params->piDataBufPtr + ptT38Params->iSizeOfDataBuf)
                    ptT38Params->piReadDataPtr = ptT38Params->piDataBufPtr;
            }    
            if(numofpkts == 2)
            {
                *Ptr++ = istatus;
                ptT38Params->iCRCStatusRx = 0;
            }
        }
    }
    else
        *Ptr++ = 0;

#if 0
     if(FieldType==HDLCDATA)
     {
        printf("\r\n V21Pkt Port[%d]:", ptT38Params->iPortNo);
        for(i=0; i<ptT38Params->TxPayloadSize; i++)
        {
          printf(" %x",*v21PktPtr++); 
        }
        printf("\r\n");
     }
#endif

    if(!NumOfSecPkt)
    {
        *Ptr++ = 0;
        *Ptr++ = 0;
        ptT38Params->TxPayloadSize += 2;
    }


//	if(NumOfSecPkt > 0)
	UpDatePrevPktBuf(ptT38Params,TX_IFP_MAXSIZE,Offset);
#endif

    return;
}


#ifdef UDP_MODE
/*-------------------------------------------------------------------------
Function Name: GetSecondaryPkt 

Description : Read the secondary packets from the buffer having previous
    packets and appends them to the present packet.

Input: 
    ptT38Params = relay channel pointer
    destaddr = pointer to packet buffer to which the secondary packet will
               be written into

Return Value:
    None
---------------------------------------------------------------------------*/
void GetSecondaryPkt(tFAXRELAY_ESL_T38 *ptT38Params,UCHAR8 *destaddr,INT16 NumOfSecPkt)
{
    INT16 len, i, j, k, NumBytes;
    UCHAR8 *SrcAddr;

    *destaddr++ = 0;                        //preamble
    *destaddr++ = NumOfSecPkt;

    j = ptT38Params->iPrevBufIndex;

    for(i=0; i < NumOfSecPkt; i++)
    {
        SrcAddr = ptT38Params->aiPrevPktBuf;
        if(--j < 0)
			j = MAXREDUNDANCY - 1;

        for(k = 0; k < j; k++)
            SrcAddr += TX_IFP_MAXSIZE;

        NumBytes = AddRedundantPac(ptT38Params,destaddr,SrcAddr);
        destaddr += NumBytes;
		ptT38Params->TxPayloadSize += NumBytes;
    }
	ptT38Params->TxPayloadSize += 2;
}


/*-------------------------------------------------------------------------
Function Name: AddRedundantPac 

Description : Read the secondary packets from the buffer having previous
    packets and appends them to the present packet.

Input: 
    ptT38Params = relay channel pointer
    destaddr = pointer to packet buffer to which the secondary packet will
               be written into

Return Value:
    Number of bytes in the secondary pay load

---------------------------------------------------------------------------*/

UINT16 AddRedundantPac(tFAXRELAY_ESL_T38 *ptT38Params, UCHAR8 *DestPtr,UCHAR8 *SourceAddr)
{
    INT16 len, i, RetVal = 0;


#ifdef CISCO_GW
	DestPtr[0] = SourceAddr[0];
	RetVal = 1;
	if(SourceAddr[0] &0x80) {
		DestPtr[1] = SourceAddr[1];
		RetVal = 2;
	}
	DestPtr += RetVal;
	SourceAddr += RetVal;

#endif    
	
	if(SourceAddr[0] >= NOSIG && SourceAddr[0] <= V1714400LTTRN)
    {
            DestPtr[0] = SourceAddr[0];
            return(RetVal+1);
    }
    if(SourceAddr[2] == HDLCDATA || SourceAddr[2] == T4NONECMDATA)
    {
        len = SourceAddr[3];
        len <<= 8;
        len += SourceAddr[4];
        len += 8;
        if(SourceAddr[1] > 1)	//~dhruva
        {
            DestPtr[len-2] = SourceAddr[len - 2];
            RetVal++;
        }

        for(i=len-1; i>=2; i--)
            DestPtr[i-2] = SourceAddr[i-2];

        return(len+RetVal-2);
    }

    for(i=2;i<5;i++)
        DestPtr[i-2] = SourceAddr[i-2];
    return(RetVal+3);
   
}

/*-------------------------------------------------------------------------
Function Name: UpDatePrevPktBuf 

Description : Updates the previous packet buffer with the present UDP IFP
    primary packet. 

Input: 
    ptT38Params = relay channel pointer
    iSizeOfPkt = Size of current primary packet

Return Value:
    None
---------------------------------------------------------------------------*/

void UpDatePrevPktBuf(tFAXRELAY_ESL_T38 *ptT38Params,UINT16 iSizeOfPkt, UINT16 Offset)
{
    INT16 i;
    UCHAR8 *Ptr,*SrcPtr;

    Ptr = ptT38Params->aiPrevPktBuf;
    for(i=0;i<ptT38Params->iPrevBufIndex;i++)
        Ptr += TX_IFP_MAXSIZE;

	for(i=0;i<TX_IFP_MAXSIZE;i++)
		Ptr[i]=0;

	SrcPtr = ptT38Params->paiOpPktPtr + Offset;
    
	CopyChars(Ptr,SrcPtr,iSizeOfPkt);
    if(++ptT38Params->iPrevBufIndex == MAXREDUNDANCY)
            ptT38Params->iPrevBufIndex = 0;
}

#endif

/*-------------------------------------------------------------------------

Function Name: ProcErrInHdlcDataRx 

Description : In this function we call a function to form the CRC status 
reporting packet.

Input: 
ptT38Params = relay channel pointer

Return Value:
None
---------------------------------------------------------------------------*/

void ProcErrInHdlcDataRx(tFAXRELAY_ESL_T38 *ptT38Params){
    INT16 TypeOfMsg;

    TypeOfMsg = (ptT38Params->iHdlcHSEnableRx)? ptT38Params->iPktdRateRx : V21C2;
    ptT38Params->iHdlcFrameEnd = 0;
/* State machine is set to packetize signal end status */
    ptT38Params->fpStateFnPtr = (fpPtrIntFn)(PackHdlcSigEnd);
    return;
}

/*-------------------------------------------------------------------------

Function Name: PackHdlcSigEnd 

Description :
    A packet having information of end of signal is formed .

Input: 
ptT38Params = relay channel pointer

Return Value:
None
---------------------------------------------------------------------------*/

void PackHdlcSigEnd(tFAXRELAY_ESL_T38 *ptT38Params){
    UCHAR8 TypeOfMsg;

    TypeOfMsg = 
        (ptT38Params->iHdlcHSEnableRx)? ptT38Params->iPktdRateRx : V21C2;
    ptT38Params->iNumBytes = 0;
    T38PktT30Data(ptT38Params,TypeOfMsg, HDLCSIGEND,1,0,ptT38Params->iCurrentRedundancy);
	if(++ptT38Params->iCurrentRedundancy >= ptT38Params->iLSRedundancy)
		ptT38Params->iCurrentRedundancy = ptT38Params->iLSRedundancy;

/*State machine is reset to configure DP in default mode of low speed recption*/
    ptT38Params->fpStateFnPtr = (fpPtrIntFn)(SetDpModeHdlcLSRx);
    InitHdlcRx(ptT38Params);


    return;
}

/*-------------------------------------------------------------------------

Function Name: PackHSDataPkt 

Input: 
ptT38Params = relay channel pointer

Description : In Non Ecm mode if ByteCnt >= Min Pkt Len we want, we call
a function to pack this high speed data else we check for No Carrier status, if
the flag is set, we pack the remaining bytes. 

Return Value:
None
---------------------------------------------------------------------------*/


void PackHSDataPkt(tFAXRELAY_ESL_T38 *ptT38Params){
/*
    if(ptT38Params->iResendPktCnt > 0) {
		ptT38Params->iResendPktCnt--;
		T38PktT30Ind(ptT38Params,(UCHAR8)ptT38Params->iIndPktdRateRx);
		return;
	}
*/
/* State machine waits for the minimum number of elements (data bytes) to be
   received from the DP for packetization */
    if(ptT38Params->iByteCnt >= ptT38Params->iMinImageLen)
    {
/* Function call to packetize high speed data */
        PackHSData(ptT38Params);
    }

/* If "No Carrier Status " is indicated by the DP, the remaining bytes are 
   packed along with the T4 Non ECM SIGEND information */
    else if(ptT38Params->iNoCarStat)
    {
        ptT38Params->iNoCarStat = 0;
        CallPktForRemainingHSData(ptT38Params);
    }

    return;
}

/*-------------------------------------------------------------------------

Function Name: PackHSData

Description :Packs the high speed data packet for non ecm. 

Input: 
ptT38Params = relay channel pointer

Return Value:
None
---------------------------------------------------------------------------*/

void PackHSData(tFAXRELAY_ESL_T38 *ptT38Params){

    ptT38Params->iNumBytes = ptT38Params->iMinImageLen;
    ptT38Params->iPackedBytes += ptT38Params->iMinImageLen;
    ptT38Params->iByteCnt -= ptT38Params->iMinImageLen;

/* Function call to packetize high speed data (T4 NON ECM DATA) */
    T38PktT30Data(ptT38Params,(UCHAR8)ptT38Params->iPktdRateRx, T4NONECMDATA,1,0,ptT38Params->iCurrentRedundancy);

    return;
}

    
/*-------------------------------------------------------------------------

Function Name: CallPktForRemainingHSData 

Description :Packs the high speed data bytes when a no carrier occurs.

Input: 
ptT38Params = relay channel pointer

Return Value:
None
---------------------------------------------------------------------------*/

void CallPktForRemainingHSData(tFAXRELAY_ESL_T38 *ptT38Params){

    ptT38Params->iNumBytes = ptT38Params->iByteCnt;
    ptT38Params->iPackedBytes += ptT38Params->iNumBytes ;
    ptT38Params->iByteCnt = 0;

/* Function call to packetize high speed data (T4 NON ECM DATA) */
    T38PktT30Data(ptT38Params,(UCHAR8)ptT38Params->iPktdRateRx, T4NONECMDATA, 2, T4NONECMSIGEND,ptT38Params->iCurrentRedundancy);

	ptT38Params->iCurrentRedundancy = 
		(ptT38Params->iLSRedundancy < ptT38Params->iHSRedundancy) ?
		ptT38Params->iLSRedundancy : ptT38Params->iHSRedundancy;

    ptT38Params->fpStateFnPtr = (fpPtrIntFn)(SendNoSigForHS);

#if 0
    /* check if the high speed data packed was TCF */
    if( ((ptT38Params->iPrevFrameRx&0x7f) == DCSID)  && 
            ((ptT38Params->iPrevFrameTx&0x7f) != CFRID) )
    {
        /* A timer is set and state machine is set wait for spoofing */
        ptT38Params->iSpoofStartTimer = 270;

        if(ptT38Params->iFrameLength == FL_5MS)      //Frame length = 5ms
        {
            ptT38Params->iSpoofStartTimer = 540;
        }
        if(ptT38Params->iFrameLength == FL_20MS)      //Frame length = 20ms
        {
            ptT38Params->iSpoofStartTimer = 135;
        }

//dhruva
        ptT38Params->piFrameDataPtr = 
        ptT38Params->piReadDataPtr = ptT38Params->piDataBufPtr;
        InitHdlcRx(ptT38Params);
        WaitForCFR_FTT(ptT38Params);
    }

    /* If it was image data */
    else
    {
        /*State machine is set to configure DP in default mode of
          low speed reception to get MPS/EOP */
        ptT38Params->fpStateFnPtr = (fpPtrIntFn)(SetDpModeHdlcLSRx);
        InitHdlcRx(ptT38Params);
           ptT38Params->piFrameDataPtr = 
                ptT38Params->piReadDataPtr = ptT38Params->piDataBufPtr;

    }
#endif
    return;
}


void SendNoSigForHS(tFAXRELAY_ESL_T38 *ptT38Params)
{

	T38PktT30Ind(ptT38Params,NOSIG);
	//BF 260804
	ptT38Params->iOpPktValid = 2; //1; //2;//dhruva

    /* check if the high speed data packed was TCF */
    if( ((ptT38Params->iPrevFrameRx&0x7f) == DCSID)  && 
            ((ptT38Params->iPrevFrameTx&0x7f) != CFRID) )
    {
        /* A timer is set and state machine is set wait for spoofing */
        ptT38Params->iSpoofStartTimer = 270;

        if(ptT38Params->iFrameLength == FL_5MS)      //Frame length = 5ms
        {
            ptT38Params->iSpoofStartTimer = 540;
        }
        if(ptT38Params->iFrameLength == FL_20MS)      //Frame length = 20ms
        {
            ptT38Params->iSpoofStartTimer = 135;
        }

//dhruva
        ptT38Params->piFrameDataPtr = 
        ptT38Params->piReadDataPtr = ptT38Params->piDataBufPtr;
        InitHdlcRx(ptT38Params);
        WaitForCFR_FTT(ptT38Params);
    }

    /* If it was image data */
    else
    {
        /*State machine is set to configure DP in default mode of
          low speed reception to get MPS/EOP */
        ptT38Params->fpStateFnPtr = (fpPtrIntFn)(SetDpModeHdlcLSRx);
        InitHdlcRx(ptT38Params);
           ptT38Params->piFrameDataPtr = 
                ptT38Params->piReadDataPtr = ptT38Params->piDataBufPtr;

    }


}



/*-------------------------------------------------------------------------

Function Name: InitHdlcTx and  InitHdlcTxVar

Description : Initialization of the variables required for Hdlc remodulation

Input: 
ptT38Params = relay channel pointer

Return Value:
None
---------------------------------------------------------------------------*/

void InitHdlcTx(tFAXRELAY_ESL_T38 *ptT38Params){

    if(ptT38Params->piFrameDataPtr == ptT38Params->piReadDataPtr)
    {
        ptT38Params->piFrameDataPtr = ptT38Params->piDataBufPtr;
        ptT38Params->piReadDataPtr = ptT38Params->piDataBufPtr;
//        ptT38Params->iCRCStatus = 0;
//        ptT38Params->iHdlcDataAvl = 0;
    }

    ptT38Params->piWrSpoofingBufPtr = ptT38Params->piRdSpoofingBufPtr = ptT38Params->piSpoofingBufPtr;

    InitHdlcTxVar(ptT38Params);
    return;
}

void InitHdlcTxVar(tFAXRELAY_ESL_T38 *ptT38Params){
    ptT38Params->iNoCarStat = 0;
    ptT38Params->iFrameAckFrmDp = 0;
    ptT38Params->iCmdAckFrmDp = 0;
    ptT38Params->iFlgTimer = 0;
    //ptT38Params->iTrnTxFlg = 0;
    ptT38Params->iHdlcHSEnableTx = 0;
    ptT38Params->iPrevCRCStatus = 0;
    ptT38Params->iLTrnTx = 0;
//    ptT38Params->iConnected = 0;
    //ptT38Params->s_SigEnd = 0;    
    ptT38Params->iHdlcTxByteCnt = 0;
    ptT38Params->iProcLSFrameTx = 0;
    //ptT38Params->iPreambleTxFlg = 0;
    ptT38Params->iTxHdlcLSTimeOut = 0;    
    ptT38Params->iSpoofPreambleTim = 0;
    return;
}

/*-------------------------------------------------------------------------

Function Name: ProcCFRFrameTx 

Description : 
    This is done in TCP mode presently not used.

Input: 
ptT38Params = relay channel pointer

Return Value:
None
---------------------------------------------------------------------------*/

#ifdef TCP_MODE
void ProcCFRFrameTx(tFAXRELAY_ESL_T38 *ptT38Params){

    if(ptT38Params->iPrevFrameTx == FTTID)
        ptT38Params->iLocalRespOfTCF = FTTID;

    ptT38Params->iPrevFrameTx = ptT38Params->iLocalRespOfTCF;
    ptT38Params->piDataBufPtr[2] = ptT38Params->iLocalRespOfTCF;
    return;
}
#endif



/*-------------------------------------------------------------------------

Function Name: ProcNSFFrameTx 

Description : Editing the FIF fields. 

Input: 
ptT38Params = relay channel pointer

Return Value:
None
---------------------------------------------------------------------------*/

void ProcNSFFrameTx(tFAXRELAY_ESL_T38 *ptT38Params){

    /* Editing the FIF fields, atleast 2 octets */

    ptT38Params->piDataBufPtr[3] = 0;
    ptT38Params->piDataBufPtr[4] = 0;
    return;
}

/*-------------------------------------------------------------------------

Function Name: ProcDCSFrameTx 

Description : Processing the DCS frame to identify the data mode 
    The bits in the DCS information field are extracted and appropriate
    functions are called.
Input: 
ptT38Params = relay channel pointer

Return Value:
None
---------------------------------------------------------------------------*/

void ProcDCSFrameTx(tFAXRELAY_ESL_T38 *ptT38Params,INT16 mode){

    ptT38Params->iV17ModeTx = 0;

    ResetTCFParams(ptT38Params);

/* Exatract the DCS infromation field from the DCS frame received from the 
   IP network */
    mode = (mode & 0x3F) >> 2;
    
/*Control is switched according to the mode bits in the DCS information field*/
    if( (mode & 1) > 0)
        ProcDCSFrV17Tx(ptT38Params,mode);
    else if( (mode &  8) > 0)
        ProcDCSFrV29Tx(ptT38Params,mode);
    else
        ProcDCSFrV27Tx(ptT38Params,mode);    


/* Calculate the number of zeros requried in TCF for the selected mode */
#ifdef UDP_MODE
    ptT38Params->iReqdZeroesInTCFForTx = ptT38Params->iBytesInHS * ZEROESFOR1_35SECS;
#endif
#ifdef TCP_MODE
//    ptT38Params->iReqdZeroesInTCFForTx = ptT38Params->iBytesInHS * ZEROESFOR1_5SECS;
    ptT38Params->iReqdZeroesInTCFForTx = ptT38Params->iBytesInHS * 140;
#endif

    ptT38Params->iTCFdata = 1;
    return;
}

/*-------------------------------------------------------------------------

Function Name: ProcDCSFrV27Tx 

Description : Initializing the variables for remodulation with V27 info. 

Input: 
ptT38Params = relay channel pointer
mode : information field in DCS frame

Return Value:
None
---------------------------------------------------------------------------*/

void ProcDCSFrV27Tx(tFAXRELAY_ESL_T38 *ptT38Params,INT16 mode){

/* Configuration for V.27 2400bps mode */
    if(!mode)
    {
        printf("\r\nV.27 2.4 KBPS Tx Mode\n");
        ptT38Params->idRateTx = RATE_V272400;
        ptT38Params->iPktdRateTx = V27_2400;
        ptT38Params->iIndPktdRateTx = V272400TRN ;
        ptT38Params->iBytesInHS = V2724MaxBytes;

    }    
/* Configuration for V.27 4800bps mode */
    else
    {
        printf("\r\nV.27 4.8 KBPS Tx Mode\n");
        ptT38Params->idRateTx = RATE_V274800;
        ptT38Params->iPktdRateTx = V27_4800;
        ptT38Params->iIndPktdRateTx = V274800TRN ;
        ptT38Params->iBytesInHS = V2748MaxBytes;
    }

    return;
}

/*-------------------------------------------------------------------------

Function Name: ProcDCSFrV17Tx 

Description : Initializing the variables for remodulation with V17 info. 

Input: 
ptT38Params = relay channel pointer
mode : information field in DCS frame


Return Value:
None
---------------------------------------------------------------------------*/

void ProcDCSFrV17Tx(tFAXRELAY_ESL_T38 *ptT38Params,INT16 mode){

    ptT38Params->iV17ModeTx = 1;

/* Configuration for V.17 14400bps mode */
    if( !(mode>>1) )
    {
	printf("\r\n V17 14.4\n");
        ptT38Params->idRateTx = RATE_V1714400S;
        ptT38Params->iPktdRateTx = V17_14400;
        ptT38Params->iIndPktdRateTx = V1714400STTRN ;
        ptT38Params->iBytesInHS = V17144MaxBytes;
    }

/* Configuration for V.17 12000bps mode */
    else if( !(mode & 8) )
    {
	printf("\r\n V17 12\n");
        ptT38Params->idRateTx = RATE_V1712000S;
        ptT38Params->iPktdRateTx = V17_12000;
        ptT38Params->iIndPktdRateTx = V1712000STTRN ;
        ptT38Params->iBytesInHS = V17120MaxBytes;

    }
/* Configuration for V.17 9600bps mode */
    else if( !(mode & 4) )
    {
	printf("\r\n V17 9.6\n");
        ptT38Params->idRateTx = RATE_V179600S;
        ptT38Params->iPktdRateTx = V17_9600;
        ptT38Params->iIndPktdRateTx = V179600STTRN ;
        ptT38Params->iBytesInHS = V1796MaxBytes;

    }
/* Configuration for V.17 7200bps mode */
    else 
    {
	printf("\r\n V17 7.2\n");
        ptT38Params->idRateTx = RATE_V177200S;
        ptT38Params->iPktdRateTx = V17_7200;
        ptT38Params->iIndPktdRateTx = V177200STTRN ;
        ptT38Params->iBytesInHS = V1772MaxBytes;
        
    }
    return;
}

/*-------------------------------------------------------------------------

Function Name: ProcDCSFrV29Tx 

Description : Initializing the variables for remodulation with V29 info. 

Input: 
ptT38Params = relay channel pointer
mode : information field in DCS frame


Return Value:
None
---------------------------------------------------------------------------*/

void ProcDCSFrV29Tx(tFAXRELAY_ESL_T38 *ptT38Params,INT16 mode){


/* Configuration for V.29 9600bps mode */
    if( !(mode & 4) )
    {
        printf("\r\nV.29 9.6 KBPS Tx Mode\n");
        ptT38Params->idRateTx = RATE_V299600;
        ptT38Params->iPktdRateTx = V29_9600;
        ptT38Params->iIndPktdRateTx = V299600TRN ;
        ptT38Params->iBytesInHS = V2996MaxBytes;

    }
/* Configuration for V.29 7200bps mode */
    else
    {
        printf("\r\nV.29 7.2 KBPS Tx Mode\n");
        ptT38Params->idRateTx = RATE_V297200;
        ptT38Params->iPktdRateTx = V29_7200;
        ptT38Params->iIndPktdRateTx = V297200TRN ;
        ptT38Params->iBytesInHS = V2972MaxBytes;
    }

    return;
}

/*-------------------------------------------------------------------------

Function Name: ProcEndHSDataTx

Description :
    DP is configured to end fax transmision/reception.
    State machine is set to wait for the response from DP stating that
    it has ended fax TxRx.
Input: 
ptT38Params = relay channel pointer

Return Value:
None
---------------------------------------------------------------------------*/

void ProcEndHSDataTx(tFAXRELAY_ESL_T38 *ptT38Params)
{
    SendEndMsgToDp(ptT38Params);
    ptT38Params->fpStateFnPtr = (fpPtrIntFn)(WaitForAckHSTx);
	if(!ptT38Params->iHdlcDataAvl) {
		//BF 260804
        ptT38Params->piFrameDataPtr = ptT38Params->piReadDataPtr = ptT38Params->piDataBufPtr; //dhruva 140203
		ClearSizeBuf(ptT38Params);
		ClearCRCBuf(ptT38Params);
	}
    return;
}

/*-------------------------------------------------------------------------

Function Name: SendEndMsgToDp

Description :

    A message is sent to DP indicating it to end fax transmision/reception.


Input: 
ptT38Params = relay channel pointer

Return Value:
None
---------------------------------------------------------------------------*/
void SendEndMsgToDp(tFAXRELAY_ESL_T38 *ptT38Params)
{
//Message ID mT38MSGTODP indicates host to forward this message to DP
    //printf("\r\n mDP_ENDFAX_TXRX port[%d]",ptT38Params->iPortNo);
    ptT38Params->tMessage.iId = mDP_ENDFAX_TXRX;
    ptT38Params->tMessage.iLen = 0;
    SendMsgToHost(ptT38Params);
    return;
}

/*-------------------------------------------------------------------------

Function Name: WaitForAckHSTx

Description : 
    The state machine is set to wait for a response from the DP for ending
    of faxTxRx. On reception of the response the DP is again configured
    in default low speed reception mode.
 
Input: 
ptT38Params = relay channel pointer

Return Value:
None
---------------------------------------------------------------------------*/

void WaitForAckHSTx(tFAXRELAY_ESL_T38 *ptT38Params){

/* wait for acknowledgement from DP */
    if(ptT38Params->iCmdAckFrmDp)
     {
        ptT38Params->iCmdAckFrmDp = 0;    
/* If the previous frame received by the DP was CFR or MCF (ie. succesfull 
   reception of high speed data) a turn around from high speed to low speed
   delay is given and the state machine is set to wait for the turn around
   time */
        if( ((ptT38Params->iPrevFrameRx & 0x7F) == CFRID)
                || ((ptT38Params->iPrevFrameRx&0x7f)==MCFID))
         {
//            ConfigDpInHdlcModeRx(ptT38Params,V21300);
//            InitHdlcRx(ptT38Params);
//dhruva!

#if 0  
            if(ptT38Params->piFrameDataPtr == ptT38Params->piReadDataPtr)
            {
                ptT38Params->piFrameDataPtr = ptT38Params->piReadDataPtr = ptT38Params->piDataBufPtr;
                ClearSizeBuf(ptT38Params);
                ClearCRCBuf(ptT38Params);   //dhruva
            }
#endif

            ptT38Params->iTimerHL = TURNAROUNDTIME;
            if(ptT38Params->iFrameLength == FL_5MS)      //Frame length = 5ms
            {
                ptT38Params->iTimerHL = TURNAROUNDTIME*2;
            }
            if(ptT38Params->iFrameLength == FL_20MS)      //Frame length = 20ms
            {
                ptT38Params->iTimerHL = 3;    //TURNAROUNDTIME/2;
            }

            ptT38Params->fpStateFnPtr = (fpPtrIntFn)(WaitForTimeOutHL);
#ifdef T30_T38
		    ptT38Params->fpStateFnPtr = (fpPtrIntFn)(WaitForSettingMode);
#endif
		}
        else
         {
            SetDpModeHdlcLSRx(ptT38Params);
            InitHdlcRx(ptT38Params);
         }
        
        if(ptT38Params->iTCFdata)
        {
            ptT38Params->iTCFdata = 0;
#if 0 //dhruva added on 19/dec/02
            if(ptT38Params->iGoodTCFPkts == 2)
            {
                T38PktT30Ind(ptT38Params,V21PREAMBLE);
				if(++ptT38Params->iCurrentRedundancy >= ptT38Params->iLSRedundancy)
					ptT38Params->iCurrentRedundancy = ptT38Params->iLSRedundancy;
                ptT38Params->fpStateFnPtr = (fpPtrIntFn)(SendFTTPkt);
            }
#endif
            ptT38Params->iGoodTCFPkts = 0;
        }
     }
    return;
}

/*-------------------------------------------------------------------------

Function Name: WaitForTimeOutHL

Description : 
    State machine is set to wait for the turn around time (from high
    speed to low speed mode).
Input: 
ptT38Params = relay channel pointer

Return Value:
None
---------------------------------------------------------------------------*/
void WaitForTimeOutHL(tFAXRELAY_ESL_T38 *ptT38Params){

/* After the timer elapses the state machine is set to wait for status
   flags for further settings */
    ptT38Params->fpStateFnPtr = (fpPtrIntFn)(WaitForSettingMode);
    //if(ptT38Params->iTimerHL--)
    if(ptT38Params->iTimerHL > 0)
    {
     ptT38Params->fpStateFnPtr = (fpPtrIntFn)(WaitForTimeOutHL);
     ptT38Params->iTimerHL--;
    } 
     
    return;
}

/*-------------------------------------------------------------------------

Function Name: ProcEndHdlcDataTx

Description :
    HDLC mode of data transmision is ended
Input: 
ptT38Params = relay channel pointer

Return Value:
None
---------------------------------------------------------------------------*/
void ProcEndHdlcDataTx(tFAXRELAY_ESL_T38 *ptT38Params){


/* Reset the read and write data pointers if there is no data to be read */
    if(ptT38Params->piFrameDataPtr == ptT38Params->piReadDataPtr)
    {
        ptT38Params->piFrameDataPtr = ptT38Params->piDataBufPtr;
        ptT38Params->piReadDataPtr = ptT38Params->piDataBufPtr;
    }

    ptT38Params->piWrSpoofingBufPtr = ptT38Params->piRdSpoofingBufPtr = ptT38Params->piSpoofingBufPtr;
    ProcEndHdlcDataTx1(ptT38Params);
    return;
}

/*-------------------------------------------------------------------------

Function Name: ProcEndHdlcDataTx1

Description :
    State machine is set to wait for acknowledge from the DP, for the 
    HDLC frame that    was sent to it by the relay.
    On reception of which and based on whether the previous frame was
    a final frame, the state machine is reset appropriately.

Input: 
ptT38Params = relay channel pointer

Return Value:
None
---------------------------------------------------------------------------*/
void ProcEndHdlcDataTx1(tFAXRELAY_ESL_T38 *ptT38Params){

    ptT38Params->fpStateFnPtr = (fpPtrIntFn)(ProcEndHdlcDataTx1);

/* wait for acknowledge from the DP */
    if(ptT38Params->iFrameAckFrmDp)
     {
        ptT38Params->iFrameAckFrmDp = 0;

/* If the previous HDLC frame transmitted was non a final frame configure
   the state machine to wait for more HDLC frames */
//        if((ptT38Params->iPrevCRCStatus == 1) || (ptT38Params->iPrevCRCStatus == 2) )
//dhruva
//		if( ptT38Params->iSigEnd == 0)
		if(!ptT38Params->iFinalFrame )
		{
            ptT38Params->iPrevCRCStatus = 0 ;
            InitHdlcTxVar(ptT38Params);
            ptT38Params->iConnected = 0;
            ptT38Params->fpStateFnPtr = (fpPtrIntFn)(ChkHdlcDataAvl);    

        }

/* If it was a final frame, the state machine is set to wait for the DP to 
   transmit preamble flags for 10ms in ECM mode, 120ms in low speed mode */
        else
         {
			ptT38Params->iFinalFrame = 0;

            if(ptT38Params->iHdlcHSEnableTx)
                ptT38Params->iFlgTimer = 1;
            else
            {
                ptT38Params->iFlgTimer = 12;
        
                if(ptT38Params->iFrameLength == FL_5MS)      //Frame length = 5ms
                {
                    ptT38Params->iFlgTimer = 24;
                }
                if(ptT38Params->iFrameLength == FL_20MS)      //Frame length = 20ms
                {
                    ptT38Params->iFlgTimer = 6;
                }
            }
            ptT38Params->fpStateFnPtr =  (fpPtrIntFn)(WtForSndngFlgs);

#ifdef T30_T38
/* DP is configured to end fax transmision/reception after the timer elapses*/
         SendEndMsgToDp(ptT38Params); 
/* State machine is set to wait for acknowledge from DP for the above message*/
        ptT38Params->fpStateFnPtr = (fpPtrIntFn)(WaitForAckHdlcTx);    
        ptT38Params->iConnected = 0;
#endif
			
		}
     }
        
    return;
}

/*-------------------------------------------------------------------------

Function Name: WtForSndngFlgs

Description :
    The state machine is set to wait for the DP to transmit preamble flags 
Input: 
ptT38Params = relay channel pointer

Return Value:
None
---------------------------------------------------------------------------*/
void WtForSndngFlgs(tFAXRELAY_ESL_T38 *ptT38Params){

    if(ptT38Params->iFlgTimer--)
      ptT38Params->fpStateFnPtr =  (fpPtrIntFn)(WtForSndngFlgs);
    else 
     {

/* DP is configured to end fax transmision/reception after the timer elapses*/
         SendEndMsgToDp(ptT38Params); 
/* State machine is set to wait for acknowledge from DP for the above message*/
        ptT38Params->fpStateFnPtr = (fpPtrIntFn)(WaitForAckHdlcTx);    
        ptT38Params->iConnected = 0;
     }

    return;
}

/*-------------------------------------------------------------------------

Function Name: WaitForAckHdlcTx

Description :
    State machine is set to wait for acknowledge from the DP, on reception
 of which the DP is configured for default mode of low speed reception

Input: 
ptT38Params = relay channel pointer

Return Value:
None
---------------------------------------------------------------------------*/

void WaitForAckHdlcTx(tFAXRELAY_ESL_T38 *ptT38Params){
       INT16 CRCStatus;
    
/* wait for acknowledge from the DP */
    if(ptT38Params->iCmdAckFrmDp)
    {
        ptT38Params->iCmdAckFrmDp = 0;
        CRCStatus = ptT38Params->iPrevCRCStatus ;
        ptT38Params->iPrevCRCStatus = 0;
        ptT38Params->fpStateFnPtr = (fpPtrIntFn)(SetDpModeHdlcLSRx);
        InitHdlcRx(ptT38Params);
//           ptT38Params->piFrameDataPtr = ptT38Params->piReadDataPtr = ptT38Params->ps_DataBufPtr;

/*//dhruva, added on 30/dec/02
		ptT38Params->piFrameDataPtr = ptT38Params->piDataBufPtr;
		ptT38Params->piReadDataPtr = ptT38Params->piDataBufPtr;
*/

        /* If it is Hdlc Low speed */
        if(!ptT38Params->iHdlcHSEnableTx)
        {

			if(CRCStatus == 0x4 || CRCStatus == 0x2) {
				ptT38Params->iSigEnd = 0;
//dhruva added on 23/dec/02
			    ptT38Params->piFrameDataPtr = ptT38Params->piDataBufPtr;
				ptT38Params->piReadDataPtr = ptT38Params->piDataBufPtr;
			}

        /* If CRC is 1 or 3 i.e CRCOK or CRCOKSIGEND, then depending on
        prev frame Tx and Rx set the data pump mode. */
            if(CRCStatus & 0x1 )
            {

                if((ptT38Params->iPrevFrameTx & 0x7F) !=DCSID) 
				{ 
                    SetRxModeHdlc(ptT38Params);
					ptT38Params->iSigEnd = 0;

#if 1 
/////////dhruva 310804, added to send CRP packet
		if ((ptT38Params->iPrevFrameTx & 0x7F) ==MPSID || (ptT38Params->iPrevFrameTx & 0x7F) ==EOPID) 
		{
			ptT38Params->iSpoofStartTimer = 280;
			if(ptT38Params->iFrameLength == FL_5MS)      //Frame length = 5ms
				ptT38Params->iSpoofStartTimer = 560;
			if(ptT38Params->iFrameLength == FL_20MS)      //Frame length = 20ms
				ptT38Params->iSpoofStartTimer = 140;

			ptT38Params->fpStateFnPtr = (fpPtrIntFn)(WaitToSendCRPPkt);

#if 1 //dhruva 010904
                        //dhruva 010904
		        ptT38Params->iPreambleRxFlg=0; //Dhruva 010804
			ptT38Params->iWaitForPreambleTxFlgTim = 0;
                        //dhruva 010904
#endif
		}
/////////////            
#endif
                }
				else
                {
                    ConfigDpInHdlcModeRx(ptT38Params,RATE_V21300);
                    InitHdlcVarRx(ptT38Params);
                    /* If the previous HDLC frame transmitted was DCS, which is a final frame,
                    the state machine is set to wait for SIGEND indicator packet */
                    ptT38Params->fpStateFnPtr = (fpPtrIntFn)(WaitForSigEnd);        
                }

            }    
        }    
        else /*if Hdlc HS */
        {
             if(CRCStatus & 0x1 )
             {
                 if((ptT38Params->iPrevFrameRx & 0x7F) == CFRID)
                 {
                    ConfigDpInHdlcModeRx(ptT38Params,ptT38Params->idRateRx);
                    InitHdlcRx(ptT38Params);
                    ptT38Params->iTimerHL = TURNAROUNDTIME;
                    if(ptT38Params->iFrameLength == FL_5MS)      //Frame length = 5ms
                    {
                        ptT38Params->iTimerHL = TURNAROUNDTIME*2;
                    }
                    if(ptT38Params->iFrameLength == FL_20MS)      //Frame length = 20ms
                    {
                        ptT38Params->iTimerHL = 3;    //TURNAROUNDTIME/2;
                    }

                    ptT38Params->fpStateFnPtr = (fpPtrIntFn)(WaitForSigEnd);
                 }
             }
        }
    }
    return;
}

/*-------------------------------------------------------------------------

Function Name: SetRxModeHdlc 

Description : 
    Set up DP in HDLC receive mode 

Input: 
ptT38Params = relay channel pointer

Return Value:
None
---------------------------------------------------------------------------*/

void SetRxModeHdlc(tFAXRELAY_ESL_T38 *ptT38Params){

    INT16  FrameTypeRx;

/* Check for the previous HDLC frame transmitted and switch accordingly */
    switch(ptT38Params->iPrevFrameTx & 0x7F )
    {
		//dhruva 220103
		case DISID:
		case FTTID:
			ptT38Params->piFrameDataPtr = ptT38Params->piDataBufPtr;
			ptT38Params->piReadDataPtr = ptT38Params->piDataBufPtr;
			break;
	
		case 33 :        
               if(ptT38Params->iEcm)
               {
                    ConfigDpInHdlcModeRx(ptT38Params,ptT38Params->idRateRx);
                   InitHdlcRx(ptT38Params);
                }
                else    
                    ConfigDpInHSSTrnModeRx(ptT38Params);    /*CFRID*/
                ptT38Params->fpStateFnPtr = (fpPtrIntFn)(RecvTrn);
                break;

        case 35 :
               if(ptT38Params->iEcm)
               {                
                    ConfigDpInHdlcModeRx(ptT38Params,ptT38Params->idRateRx);
                    InitHdlcRx(ptT38Params);
               }
               else    
                    ConfigDpInHSSTrnModeRx(ptT38Params); /*CTRID*/
                ptT38Params->fpStateFnPtr = (fpPtrIntFn)(RecvTrn);
                break;

        default :
                FrameTypeRx = ptT38Params->iPrevFrameRx & 0x7F;
                if(FrameTypeRx == PPSID)
                    SNPEcm(ptT38Params);
                else if(FrameTypeRx == MPSID)
                    ChkTxdFrm(ptT38Params);
                else
                {
//debug                    ptT38Params->fpStateFnPtr = (fpPtrIntFn)(SetDpModeHdlcLSRx);
                    SetDpModeHdlcLSRx(ptT38Params);
                    InitHdlcRx(ptT38Params);
                }
                break;

    }

    return;
}

/*-------------------------------------------------------------------------

Function Name: ChkTxdFrm

Description :
    The previous HDLC frame transmitted is checked for MCF (message 
  conformation) or PPR (partial page request ) and appropriate modes are
  set.

Input: 
ptT38Params = relay channel pointer

Return Value:
None
---------------------------------------------------------------------------*/

void ChkTxdFrm(tFAXRELAY_ESL_T38 *ptT38Params){

    INT16 FrameTx;

/* If the previous HDLC frame transmitted is  MCF or PPR, the DP is configured
   to receive trainig */

    FrameTx = ptT38Params->iPrevFrameTx & 0x7F;
    if(FrameTx == MCFID)
    {
        if(ptT38Params->iEcm)
        {
            //printf("\r\n ChkTxdFrm:ConfigDpInHdlcModeRx Port[%d]",ptT38Params->iPortNo);
            ConfigDpInHdlcModeRx(ptT38Params,ptT38Params->idRateRx);
            InitHdlcRx(ptT38Params);
        }
        else    
        {
            ConfigDpInHSSTrnModeRx(ptT38Params);
         
        }
        ptT38Params->fpStateFnPtr = (fpPtrIntFn)(RecvTrn);
    }
    else if(FrameTx == PPRID)
    {
        if(ptT38Params->iEcm)
        {
            ConfigDpInHdlcModeRx(ptT38Params,ptT38Params->idRateRx);
            InitHdlcRx(ptT38Params);
        }
        else
        {    
           ConfigDpInHSSTrnModeRx(ptT38Params);
        }
        ptT38Params->fpStateFnPtr = (fpPtrIntFn)(RecvTrn);
     }
    else
	{
//dhruva, 220103
//        SNPEcm(ptT38Params);
		SetDpModeHdlcLSRx(ptT38Params);
		InitHdlcRx(ptT38Params);
	}

    return;
}

/*-------------------------------------------------------------------------

Function Name: SNPEcm 

Description :
    If MPS (More Pages to Send) request was received by DP, the state
    machine waits for MCF or PPR.
    Other wise DP is configured to default mode of low speed reception.    
Input: 
ptT38Params = relay channel pointer

Return Value:
None
---------------------------------------------------------------------------*/

void SNPEcm(tFAXRELAY_ESL_T38 *ptT38Params){
    INT16 frame;
    
    frame = ptT38Params->iPrevFrameRxPPS & 0x7F;
    if( (frame == MPSID) || (frame == NULLID) )
      ChkTxdFrm(ptT38Params);
    else
    {
      ptT38Params->fpStateFnPtr = (fpPtrIntFn)(SetDpModeHdlcLSRx);
      InitHdlcRx(ptT38Params);
    }

    return;
}

/*-------------------------------------------------------------------------

Function Name : WaitForSigEnd

Description :
    State machine is set to wait for SIG-END packet, on receiving which
    it is reset to allow for a delay in turning aroung from low speed
    to high speed mode or vise versa.

Input: 
ptT38Params = relay channel pointer

Return Value:
None
---------------------------------------------------------------------------*/

void WaitForSigEnd(tFAXRELAY_ESL_T38 *ptT38Params){

    if(ptT38Params->iSigEnd)
     {
        ptT38Params->iSigEnd = 0;
/* If the current configuration is low speed Tx mode*/
        if(!ptT38Params->iHdlcHSEnableTx)
        {
            ptT38Params->fpStateFnPtr = (fpPtrIntFn)(WaitForTimeOutLH);
            ptT38Params->iTimerLH = TURNAROUNDTIME;
            if(ptT38Params->iFrameLength == FL_5MS)      //Frame length = 5ms
            {
                ptT38Params->iTimerLH = TURNAROUNDTIME*2;
            }
            if(ptT38Params->iFrameLength == FL_20MS)      //Frame length = 20ms
            {
                ptT38Params->iTimerLH = 3;    //TURNAROUNDTIME/2;
            }

            WaitForTimeOutLH(ptT38Params);
        }
/* If the current configuration is high speed Tx mode*/
        else
        {
            ptT38Params->fpStateFnPtr = (fpPtrIntFn)(WaitForTimeOutHL);
            ptT38Params->iTimerHL = TURNAROUNDTIME;
            if(ptT38Params->iFrameLength == FL_5MS)      //Frame length = 5ms
            {
                ptT38Params->iTimerHL = TURNAROUNDTIME*2;
            }
            if(ptT38Params->iFrameLength == FL_20MS)      //Frame length = 20ms
            {
                ptT38Params->iTimerHL = 3;    //TURNAROUNDTIME/2;
            }

            WaitForTimeOutHL(ptT38Params);
        }
      }
    return;
} 

/*-------------------------------------------------------------------------

Function Name: WaitForTimeOutLH

Description : 
    State machine is set to wait for the turn around time (from low
    speed to high speed mode).

Input: 
ptT38Params = relay channel pointer

Return Value:
None
---------------------------------------------------------------------------*/

void WaitForTimeOutLH(tFAXRELAY_ESL_T38 *ptT38Params){

/* After the timer elapses the state machine is set to wait for status
   flags for further settings */
    if(!ptT38Params->iTimerLH)
     {
        ptT38Params->fpStateFnPtr = (fpPtrIntFn)(SpoofTCF);
        if(ptT38Params->iV17ModeTx)
             ptT38Params->iLTrnTx = 1;
        else
             ptT38Params->iLTrnTx = 0;
        
        ptT38Params->iTimerLH = 0; 

#ifdef T30_T38	//dhruva14
		if(ptT38Params->iDroppedPkt) {
	        ptT38Params->fpStateFnPtr = (fpPtrIntFn)(WaitForSettingMode);
			ptT38Params->iDroppedPkt = 0;
		}
#endif

	}
    else {
        --ptT38Params->iTimerLH;
	}

    return;
}


/*-------------------------------------------------------------------------

Function Name: TxHSDataToDp

Description :
    Data is sent to the DP for transmision in high speed mode after checking
    the readiness of the DP (XOFF)
    If the Data is TCF and mode of packet transfer is TCP, zeroes are
    sent as data through messages to DP.
Input: 
ptT38Params = relay channel pointer

Return Value:
None
---------------------------------------------------------------------------*/
//UINT32 HSImageCntrFlg, HSImageCntr = 0;
void TxHSDataToDp(tFAXRELAY_ESL_T38 *ptT38Params){

    INT16 datalen,i,N,NumOfBytes,M;
    UINT16 datatodp=0;

/* While in high speed spoof mode, enable HS data as soon as trn indicator
   packet is recived */
    if(ptT38Params->iTrnTxFlg)
    {
        ptT38Params->iTrnTxFlg = 0;
        //enable high speed image transmision
        if(!ptT38Params->iTCFdata)
        {
#if 0
           printf("\r\n TxHSDataToDp :HSspoofCntr[%d]",HSspoofCntr);
           HSspoofCntrFlg = HSspoofCntr = 0;
           HSImageCntrFlg = 1;
#endif

            ptT38Params->iImageDataTx = 1;
           //Added for HS Image Spoof Timer elapase cases
           ptT38Params->iHdlcDataAvl = 0; 
        }
    }

    NumOfBytes = ptT38Params->iBytesInHS;


/* check for  the readiness of the DP */
    if(ptT38Params->iConnected && (!ptT38Params->iXOFF) )
    {

#ifdef TCP_MODE
/*    If the Data is TCF and mode of packet transfer is TCP, zeroes are
    sent as data through messages to DP. */
        if( ((ptT38Params->iPrevFrameTx & 0x7F) == DCSID) && 
                    ((ptT38Params->iPrevFrameRx & 0x7F) != CFRID) )
        {
            ptT38Params->tMessage.iLen = ptT38Params->iBytesInHS;

            if(ptT38Params->iFrameLength == FL_20MS)      //Frame length = 20ms
                ptT38Params->tMessage.iLen = 40;

            for(i=0 ; i<ptT38Params->tMessage.iLen ; i++)
                ptT38Params->tMessage.aiParams[i] = 0;    //zeroes as TCF bytes

            ptT38Params->iZeroesInTCFTx +=ptT38Params->tMessage.iLen;

//Message ID mT38MSGTODP indicates host to forward this message to DP
            ptT38Params->tMessage.iId = mDP_MOD_DATA;
            SendMsgToHost(ptT38Params);

            ptT38Params->fpStateFnPtr = (fpPtrIntFn)(TxHSDataToDp);

            /* After the requried number of zeroes in TCF are sent, state machine is 
            configured to end high speed data mode */
            if(ptT38Params->iZeroesInTCFTx >= ptT38Params->iReqdZeroesInTCFForTx)
            {
                ptT38Params->iConnected = 0;
                ProcEndHSDataTx(ptT38Params);
            }    
        }
/* Image Data is not sent to DP for transmision till EOL is received */
        else if(ptT38Params->iImageDataTx)
        {
            if(ptT38Params->iEOLCnt > 0)
            {
                if(ptT38Params->iFrameLength == FL_20MS)  //Frame length = 20ms
                    NumOfBytes += ptT38Params->iBytesInHS;

    // Find number of data elements (bytes) to be sent to DP
                while(ptT38Params->PrevLineData < NumOfBytes && ptT38Params->iEOLCnt > 0)
                {
                    ptT38Params->PrevLineData += 
                        ptT38Params->aiSizeBuf[ptT38Params->iSizeRIndex];
                    ptT38Params->aiSizeBuf[ptT38Params->iSizeRIndex]=0;
                    if(++ptT38Params->iSizeRIndex == SIZEBUFSIZE)
                        ptT38Params->iSizeRIndex = 0;
                    ptT38Params->iEOLCnt--;
                }
            }
            if(ptT38Params->PrevLineData > 0)
            {
                datalen = ptT38Params->PrevLineData;

                if(datalen > NumOfBytes)
                    N = NumOfBytes;
                else
                    N = datalen;

                ptT38Params->PrevLineData -= N;

                // The data is copied to message buffer
                for(i=0; i<N; i++)
                {
                    datatodp = *ptT38Params->piReadDataPtr;
                    if(++ptT38Params->piReadDataPtr == (ptT38Params->piDataBufPtr+ptT38Params->iSizeOfDataBuf))
                        ptT38Params->piReadDataPtr=ptT38Params->piDataBufPtr;
                    ptT38Params->tMessage.aiParams[i] = datatodp; 
                }

//Message ID mT38MSGTODP indicates host to forward this message to DP
                ptT38Params->tMessage.iId = mDP_MOD_DATA;
                ptT38Params->tMessage.iLen = i;

/* function call to update message queue */
                if(ptT38Params->tMessage.iLen > 0)
                {
                    SendMsgToHost(ptT38Params);
                }
            }

/* On reception of SIG-END state machine is configured to end high speed 
data mode */
            if(ptT38Params->iSigEndHS && ptT38Params->iEOLCnt == 0 && ptT38Params->PrevLineData == 0)
            {
                N = ptT38Params->aiSizeBuf[ptT38Params->iSizeRIndex];
                N += ptT38Params->PrevLineData;
                ptT38Params->PrevLineData = 0;
                ptT38Params->aiSizeBuf[ptT38Params->iSizeRIndex]=0;
                if(++ptT38Params->iSizeRIndex == SIZEBUFSIZE)
                    ptT38Params->iSizeRIndex = 0;

                if(N > 18) {
                    M = N-18;
                    N = 18;
                }
                else M = 0;

                for(i=0; i<N; i++) {
                    datatodp = *ptT38Params->piReadDataPtr;
                    if(++ptT38Params->piReadDataPtr == (ptT38Params->piDataBufPtr+ptT38Params->iSizeOfDataBuf))
                        ptT38Params->piReadDataPtr=ptT38Params->piDataBufPtr;
                    ptT38Params->tMessage.aiParams[i] = datatodp;
                    if(ptT38Params->piReadDataPtr == ptT38Params->piFrameDataPtr)
                        break;
                }

                while((ptT38Params->piReadDataPtr != ptT38Params->piFrameDataPtr)) {
                    if(++ptT38Params->piReadDataPtr == (ptT38Params->piDataBufPtr+ptT38Params->iSizeOfDataBuf))
                        ptT38Params->piReadDataPtr=ptT38Params->piDataBufPtr;
                }

//Message ID mT38MSGTODP indicates host to forward this message to DP
                ptT38Params->tMessage.iId = mDP_MOD_DATA;
                ptT38Params->tMessage.iLen = i;

/* function call to update message queue */
                if(ptT38Params->tMessage.iLen > 0) {
                    SendMsgToHost(ptT38Params);
                }

                ptT38Params->iImageDataTx = 0;
                ptT38Params->iSigEndHS = 0;
                ptT38Params->iConnected = 0;
                ProcEndHSDataTx(ptT38Params);
            }    
        }
#endif

#ifdef UDP_MODE
//TCF data
        if(!ptT38Params->iImageDataTx)
        {
            /* Find number of data elements (bytes) to be sent to DP*/
            datalen = ptT38Params->piFrameDataPtr - ptT38Params->piReadDataPtr;
            if( datalen != 0)
            {
                if(datalen < 0)
                    datalen += ptT38Params->iSizeOfDataBuf;
                if(ptT38Params->iFrameLength == FL_20MS)  //Frame length = 20ms
                {
                    NumOfBytes += ptT38Params->iBytesInHS;
                }

                if(datalen > NumOfBytes)
                    N = NumOfBytes;
                else
                    N = datalen;

				/* The data is copied to message buffer */
                for(i=0; i<N; i++)
                {
                    datatodp = *ptT38Params->piReadDataPtr;
                    if(++ptT38Params->piReadDataPtr == (ptT38Params->piDataBufPtr+ptT38Params->iSizeOfDataBuf))
                        ptT38Params->piReadDataPtr=ptT38Params->piDataBufPtr;
                    ptT38Params->tMessage.aiParams[i] = datatodp;
                }

//Message ID mT38MSGTODP indicates host to forward this message to DP
                ptT38Params->tMessage.iId = mDP_MOD_DATA;
                ptT38Params->tMessage.iLen = i; //Raju 0907

// function call to update message queue
                if(ptT38Params->tMessage.iLen > 0)
                    SendMsgToHost(ptT38Params);

// On reception of SIG-END state machine is configured to end high speed 
// data mode
                if(ptT38Params->iSigEndHS)
                {
                   if(ptT38Params->piFrameDataPtr == ptT38Params->piReadDataPtr)
                   {
                        ptT38Params->iSigEndHS = 0;
                        ptT38Params->iConnected = 0;
                        ProcEndHSDataTx(ptT38Params);
                    }                  
                }    
            }
            else
            {
                if(ptT38Params->iSigEndHS)
                {
                    printf("\r\n TxHSDataToDp: T4 SIG END TCF or SPOOF case");
                    ptT38Params->iSigEndHS = 0;
                    ptT38Params->iConnected = 0;
                    ProcEndHSDataTx(ptT38Params);
                }
				//dhruva14
				if(ptT38Params->iDroppedPkt) {
					ptT38Params->iDroppedPkt=0;
					ptT38Params->iSigEndHS = 1;
				}

            }

        }

// Image Data is not sent to DP for transmision till EOL is received 
        else
        {

            if(ptT38Params->iEOLCnt > 0)
            {
                if(ptT38Params->iFrameLength == FL_20MS)  //Frame length = 20ms
                    NumOfBytes += ptT38Params->iBytesInHS;

    // Find number of data elements (bytes) to be sent to DP
                while(ptT38Params->PrevLineData < NumOfBytes && ptT38Params->iEOLCnt > 0)
                {
                    ptT38Params->PrevLineData += 
                        ptT38Params->aiSizeBuf[ptT38Params->iSizeRIndex];
                    ptT38Params->aiSizeBuf[ptT38Params->iSizeRIndex]=0;
                    if(++ptT38Params->iSizeRIndex == SIZEBUFSIZE)
                        ptT38Params->iSizeRIndex = 0;
                    ptT38Params->iEOLCnt--;
                }
            }
            if(ptT38Params->PrevLineData > 0)
            {
                datalen = ptT38Params->PrevLineData;

                if(datalen > NumOfBytes)
                    N = NumOfBytes;
                else
                    N = datalen;

                ptT38Params->PrevLineData -= N;

                // The data is copied to message buffer
                for(i=0; i<N; i++)
                {
                    datatodp = *ptT38Params->piReadDataPtr;
                    if(++ptT38Params->piReadDataPtr == (ptT38Params->piDataBufPtr+ptT38Params->iSizeOfDataBuf))
                        ptT38Params->piReadDataPtr=ptT38Params->piDataBufPtr;
                    ptT38Params->tMessage.aiParams[i] = datatodp;
                }

//Message ID mT38MSGTODP indicates host to forward this message to DP
                ptT38Params->tMessage.iId = mDP_MOD_DATA;
                ptT38Params->tMessage.iLen = i; //Raju 0907

/* function call to update message queue */
                if(ptT38Params->tMessage.iLen > 0)
                {
                    SendMsgToHost(ptT38Params);
                }
            }

/* On reception of SIG-END state machine is configured to end high speed 
data mode */
            if(ptT38Params->iSigEndHS && ptT38Params->iEOLCnt == 0 && ptT38Params->PrevLineData == 0)
            {
                N = ptT38Params->aiSizeBuf[ptT38Params->iSizeRIndex];
                N += ptT38Params->PrevLineData;
                ptT38Params->PrevLineData = 0;
                ptT38Params->aiSizeBuf[ptT38Params->iSizeRIndex]=0;
                if(++ptT38Params->iSizeRIndex == SIZEBUFSIZE)
                    ptT38Params->iSizeRIndex = 0;
#if 1 //Raju 0807
                if(N > 18) {
                    M = N-18;
                    N = 18;
                }
                else M = 0;
#endif

                for(i=0; i<N; i++) {
                    datatodp = *ptT38Params->piReadDataPtr;
                    if(++ptT38Params->piReadDataPtr == (ptT38Params->piDataBufPtr+ptT38Params->iSizeOfDataBuf))
                        ptT38Params->piReadDataPtr=ptT38Params->piDataBufPtr;
                    ptT38Params->tMessage.aiParams[i] = datatodp;
                    if(ptT38Params->piReadDataPtr == ptT38Params->piFrameDataPtr)
                        break;
                }

                //while(M-- > 0 && (ptT38Params->piReadDataPtr != ptT38Params->piFrameDataPtr)) { //Raju 0807
                while((ptT38Params->piReadDataPtr != ptT38Params->piFrameDataPtr)) {
                    if(++ptT38Params->piReadDataPtr == (ptT38Params->piDataBufPtr+ptT38Params->iSizeOfDataBuf))
                        ptT38Params->piReadDataPtr=ptT38Params->piDataBufPtr;
                }

//Message ID mT38MSGTODP indicates host to forward this message to DP
                ptT38Params->tMessage.iId = mDP_MOD_DATA;
                ptT38Params->tMessage.iLen = i; //Raju 0907

/* function call to update message queue */
                if(ptT38Params->tMessage.iLen > 0) {
                    SendMsgToHost(ptT38Params);
                }

//BF 260804
///////////dhruva 140203
				N = ptT38Params->aiSizeBuf[ptT38Params->iSizeRIndex];
				M = ptT38Params->iSizeRIndex + 1;
				if(M == SIZEBUFSIZE)
					M = 0;
				M = ptT38Params->aiSizeBuf[M];

                while( N != 3 && M != 0)
				{
					ptT38Params->piReadDataPtr += N;
                    if(ptT38Params->piReadDataPtr >= (ptT38Params->piDataBufPtr+ptT38Params->iSizeOfDataBuf))
                        ptT38Params->piReadDataPtr -= ptT38Params->iSizeOfDataBuf;
					if(++ptT38Params->iSizeRIndex == SIZEBUFSIZE)
						ptT38Params->iSizeRIndex = 0;

					N = ptT38Params->aiSizeBuf[ptT38Params->iSizeRIndex];
					M = ptT38Params->iSizeRIndex + 1;
					if(M == SIZEBUFSIZE)
						M = 0;
					M = ptT38Params->aiSizeBuf[M];
				}
////////////

                ptT38Params->iImageDataTx = 0;
                ptT38Params->iSigEndHS = 0;
                ptT38Params->iConnected = 0;
                ProcEndHSDataTx(ptT38Params);
            }    
//dhruva14
            if(ptT38Params->iDroppedPkt && ptT38Params->iEOLCnt == 0 && ptT38Params->PrevLineData == 0
				&& (ptT38Params->iHdlcDataAvl || ptT38Params->iPreambleTxFlg))
            {
				ptT38Params->iDroppedPkt=0;
				ptT38Params->iSigEndHS = 1;
			}

        }
#endif
    }    /* End of If connected */
    return;
}

/*-------------------------------------------------------------------------

Function Name: SetDpModeHSTx

Description : 
    Configure DP in high speed transmision mode, either for Long train
    (if V.17 Long train mode is enabled) or to Short train mode.
Input: 
ptT38Params = relay channel pointer

Return Value:
None
---------------------------------------------------------------------------*/

void SetDpModeHSTx(tFAXRELAY_ESL_T38 *ptT38Params){
    
    ptT38Params->iTrnTxFlg = 0 ;
/* Check if long train mode is enabled */
    if(!ptT38Params->iLTrnTx)
    {
        ConfigDpInHSSTrnModeTx(ptT38Params);
        printf("\r\nDP set to HSSTrnTx"); 
    }
    else 
     {
        ptT38Params->iLTrnTx = 0;
        ConfigDpInHSLTrnModeTx(ptT38Params);
        //printf("\r\nDP set to HSLTrnTx"); 
     }
    return;
}

/*-------------------------------------------------------------------------

Function Name: HdlcData

Description : 
    HDLC data is copied from the message buffers to the data buffer
    after byte reversal.
Input: 
ptT38Params = relay channel pointer

Return Value:
None
---------------------------------------------------------------------------*/

void HdlcData(tFAXRELAY_ESL_T38 *ptT38Params){
    UINT16 i;
    UCHAR8 revByte;

    for(i = 0 ; i < ptT38Params->tMessage.iLen ; i++)
    {
/* Byets are reversed, because the DP works in little endian mode 
   and the T.38 in big endian mode */

        revByte = ByteReverse(ptT38Params->tMessage.aiParams[i] );
        *ptT38Params->piFrameDataPtr = revByte ;

        if(++ptT38Params->piFrameDataPtr == 
            ptT38Params->piDataBufPtr + ptT38Params->iSizeOfDataBuf )
            ptT38Params->piFrameDataPtr = ptT38Params->piDataBufPtr;
    }

    /*  ByteCnt keeps the count of the bytes to be packed */
    ptT38Params->iByteCnt += ptT38Params->tMessage.iLen ;    
    return;
}

/*-------------------------------------------------------------------------

Function Name: HighSpeedData

Description : 
    High speed data is copied from the message buffers to the data buffer.

Input: 
ptT38Params = relay channel pointer

Return Value:
None
---------------------------------------------------------------------------*/
void HighSpeedData(tFAXRELAY_ESL_T38 *ptT38Params){
    INT16 i;

/*    If the Data is TCF and mode of packet transfer is TCP, data in 
    message buffer is checked for zeroes and a counter is maintained
    to count continuous zeroes */
#ifdef TCP_MODE

    if( ( (ptT38Params->iPrevFrameRx & 0x7F) == DCSID ) && ( (ptT38Params->iPrevFrameTx & 0x7F) != CFRID) )
    {
        for(i=0 ; i <ptT38Params->tMessage.iLen ; i++)
        {
            if(ptT38Params->tMessage.aiParams[i] == 0)
            {
                ptT38Params->iZeroesInTCF++;
            }
            else
            {
                ptT38Params->iZeroesInTCF = 0;
            }
/* If number of zeros in TCF is greater than or equal to the minmum required
   zeros, the local response is set to CFR else it is FTT (Failed To Train) */
            if(ptT38Params->iLocalRespOfTCF != CFRID)
            {
               //printf("\r\n HSData: iZerosInTCF: %d\n",ptT38Params->iZeroesInTCF);
               //printf("\r\n HSData: iMinReqdZeroesInTCF: %d\n",ptT38Params->iMinReqdZeroesInTCF);
                if(ptT38Params->iZeroesInTCF >= ptT38Params->iMinReqdZeroesInTCF)
                {
		    printf("\r\nHSData:CFRID\n");
                    ptT38Params->iLocalRespOfTCF = CFRID;
                }
                else
                {
  		    printf("\r\nHSData:FTTID\n");
                    ptT38Params->iLocalRespOfTCF = FTTID;
                }
            }
        }    
        return;
    }

#endif
/*If mode of packet transfer is UDP or if image data is sent by DP */
    for(i = 0 ; i < ptT38Params->tMessage.iLen ; i++)
    {
/* copy data from the message buffers to the data buffer.*/        
        *ptT38Params->piImageDataPtr = ptT38Params->tMessage.aiParams[i] ;

        if(++ptT38Params->piImageDataPtr == (ptT38Params->piDataBufPtr + ptT38Params->iSizeOfDataBuf)) 
            ptT38Params->piImageDataPtr = ptT38Params->piDataBufPtr ;
	}

    /*  ByteCnt keeps the count of the bytes to be packed */
    ptT38Params->iByteCnt += ptT38Params->tMessage.iLen ;    
    return;
}

/*-------------------------------------------------------------------------

Function Name: InitHdlcRx and  InitHdlcRxVar

Description : Initialization of the variables required for Hdlc demodulation

Input: 
ptT38Params = relay channel pointer

Return Value:
None
---------------------------------------------------------------------------*/


void InitHdlcRx(tFAXRELAY_ESL_T38 *ptT38Params){
    if(ptT38Params->piFrameDataPtr == ptT38Params->piReadDataPtr)
    {
        ptT38Params->piFrameDataPtr = ptT38Params->piReadDataPtr = ptT38Params->piDataBufPtr;
        ClearSizeBuf(ptT38Params);
        ClearCRCBuf(ptT38Params);
    }

    InitHdlcVarRx(ptT38Params);
    return;
}

void InitHdlcVarRx(tFAXRELAY_ESL_T38 *ptT38Params){
    ptT38Params->iWaitForPreambleTxFlgTim = 0;
    ptT38Params->iHdlcEnable = 1;
    ptT38Params->iCnt7e = 0 ;
    ptT38Params->iPreambleRxFlg = 0;

//dhruva 220103
#ifdef T30_T38
    ptT38Params->iPreambleTxFlg = 0;
#endif
//    ptT38Params->iPreambleTxFlg = 0;

    ptT38Params->iNoCarStat = 0 ;
    ptT38Params->iNumBytes = 0;
    ptT38Params->iPackedBytes = 0;
    ptT38Params->iCRCStatusRx = 0 ;

    return; 
}

/*-------------------------------------------------------------------------

Function Name: ProcHdlcHSDataRx

Description : 
    Process high speed data in HDLC mode.

Input: 
ptT38Params = relay channel pointer

Return Value:
None
---------------------------------------------------------------------------*/

void ProcHdlcHSDataRx(tFAXRELAY_ESL_T38 *ptT38Params){
    ptT38Params->iHdlcFrameEnd = 1; 
    ptT38Params->iHdlcFrameEnd = 0;
//    ptT38Params->fpStateFnPtr = (fpPtrIntFn)(ProcEndHdlcHSDataRx);

    return;
}

/*-------------------------------------------------------------------------

Function Name: ProcHdlcLSDataRx

Description : 
    Process low speed data in HDLC mode.
 The the information bytes in the HDLC frame received from the DP are read
 and processed apporpriately.


Input: 
ptT38Params = relay channel pointer

Return Value:
None
---------------------------------------------------------------------------*/

void ProcHdlcLSDataRx(tFAXRELAY_ESL_T38 *ptT38Params)
{

/* Control field in HDLC frame is checked to find out whether the frame
   is a final frame or not */

    ptT38Params->iHdlcFrameEnd = (ptT38Params->piDataBufPtr[1] & 8 ) >> 3;
//    PackEndOfHdlcFrame(ptT38Params,V21C2);

    ptT38Params->iPrevFrameRx = ptT38Params->piDataBufPtr[2];
    
    /*To get the extended control field with PPS Id*/
    ptT38Params->iPrevFrameRxPPS =  ptT38Params->piDataBufPtr[3];

    //printf("\r\nCase rcvd:%x",(ptT38Params->iPrevFrameRx & 0x7F));
    switch(ptT38Params->iPrevFrameRx & 0x7F)
        {

            case 1 :        /*DISID*/
                    printf("\r\nDISID received port[%d]",ptT38Params->iPortNo);
                        //debug_fax = 100;
                        break;

            case 2 :        /*CSIID*/
                    printf("\r\nCSIID received port[%d]",ptT38Params->iPortNo);
                        break;

            case 65 : 
                    printf("\r\nDCSID received port[%d]",ptT38Params->iPortNo);
                        ProcDCSFrameRx(ptT38Params); /*DCSID*/
                        break;
            
            case 33 :        /*CFRID*/
                    printf("\r\nCFRID received port[%d]",ptT38Params->iPortNo);
                         break;

            case 34 :        /*FTTID*/
                    printf("\r\nFTTID received port[%d]",ptT38Params->iPortNo);
                         break;
            case 49 :        /*MCFID*/
                    printf("\r\nMCFID received port[%d]",ptT38Params->iPortNo);
                    break;

            case 4 :
                        printf("\r\nNSFID received port[%d]",ptT38Params->iPortNo);
                        break;

            case 95 :          /*DCNID*/
                    printf("\r\nDCNID received port[%d]",ptT38Params->iPortNo);
                        ptT38Params->iShutDown = 1;
                        ptT38Params->iErrTimer = 500 ;
                        if(ptT38Params->iFrameLength == FL_5MS)      //Frame length = 5ms
                        {
                            ptT38Params->iErrTimer = 1000;
                        }
                        if(ptT38Params->iFrameLength == FL_20MS)      //Frame length = 20ms
                        {
                            ptT38Params->iErrTimer = 250;
                        }
                        break;

        }

/* The state mecahine set depending on whether the frame is a final one or not*/
    if(!ptT38Params->iHdlcFrameEnd) 
        ProcNotLastLSFrameRx(ptT38Params);
    else
      {
        ptT38Params->iHdlcFrameEnd = 0;
        ptT38Params->iDisablePktData=0;
//dhruva~
		ptT38Params->IgnorePkts = 0;
//~dhruva

        printf("\r\nProcHdlcLSDataRx : Next Function ProcEndHdlcLSDataRx");
        ptT38Params->fpStateFnPtr = (fpPtrIntFn)(ProcEndHdlcLSDataRx);
      }

    return;
}

/*-------------------------------------------------------------------------

Function Name: ProcDCSFrameRx

Description : Processing the DCS frame to identify the data mode 
    The bits in the DCS information field are extracted and appropriate
    functions are called.

Input: 
ptT38Params = relay channel pointer

Return Value:
None
---------------------------------------------------------------------------*/

void ProcDCSFrameRx(tFAXRELAY_ESL_T38 *ptT38Params){

       INT16 mode;

/* Exatract the DCS infromation field from the DCS frame received from the 
   DP */
    ptT38Params->iV17ModeRx = 0;
    mode = (ptT38Params->piDataBufPtr[4] & 0x3F) >> 2;

/*Control is switched according to the mode bits in the DCS information field*/
    if( (mode & 1) > 0)
        ProcDCSFrV17Rx(ptT38Params,mode);
    else if( (mode &  8) > 0)
    {
        ProcDCSFrV29Rx(ptT38Params,mode);
    }
    else
    {
        ProcDCSFrV27Rx(ptT38Params,mode);
    }
    return;
}

/*-------------------------------------------------------------------------

Function Name: ProcDCSFrV27Rx 

Description : Initializing the variables for demodulation with V27 info. 

Input: 
ptT38Params = relay channel pointer
mode : information field in DCS frame

Return Value:
None
---------------------------------------------------------------------------*/

void ProcDCSFrV27Rx(tFAXRELAY_ESL_T38 *ptT38Params,INT16 mode){

    if(!mode)
     {
/* Configuration for V.27 2400bps mode */
        printf("\r\nV.27 2.4 KBPS Rx Mode\n");
        ptT38Params->idRateRx = RATE_V272400;
        ptT38Params->iPktdRateRx = V27_2400;
        ptT38Params->iIndPktdRateRx = V272400TRN ;
        ptT38Params->iMinReqdZeroesInTCF = V2724MaxBytes * ZEROESFOR1_35SECS ; 
        ptT38Params->iMinImageLen = 30;
     }    
    else
     {
/* Configuration for V.27 4800bps mode */
        printf("\r\nV.27 4.8 KBPS Rx Mode\n");
        ptT38Params->idRateRx = RATE_V274800;
        ptT38Params->iPktdRateRx = V27_4800;
        ptT38Params->iIndPktdRateRx = V274800TRN ;
        ptT38Params->iMinReqdZeroesInTCF = V2748MaxBytes * ZEROESFOR1_35SECS ; 
        ptT38Params->iMinImageLen = 60;
    }    

    return;
}


/*-------------------------------------------------------------------------

Function Name: ProcDCSFrV217Rx 

Description : Initializing the variables for demodulation with V17 info. 

Input: 
ptT38Params = relay channel pointer
mode : information field in DCS frame

Return Value:
None
---------------------------------------------------------------------------*/

void ProcDCSFrV17Rx(tFAXRELAY_ESL_T38 *ptT38Params,INT16 mode){

    ptT38Params->iV17ModeRx = 1;

/* Configuration for V.17 14400bps mode */
    if( !(mode>>1) )
     {
    
        ptT38Params->idRateRx = RATE_V1714400S;
        ptT38Params->iPktdRateRx = V17_14400;
        ptT38Params->iIndPktdRateRx = V1714400STTRN ;
        ptT38Params->iMinReqdZeroesInTCF = V17144MaxBytes * ZEROESFOR1_35SECS ; 
        ptT38Params->iMinImageLen = 180;
     }

/* Configuration for V.17 12000bps mode */
    else if( !(mode & 8) )
     {
        ptT38Params->idRateRx = RATE_V1712000S;
        ptT38Params->iPktdRateRx = V17_12000;
        ptT38Params->iIndPktdRateRx = V1712000STTRN ;
        ptT38Params->iMinReqdZeroesInTCF = V17120MaxBytes * ZEROESFOR1_35SECS ; 
        ptT38Params->iMinImageLen = 150;
     }
/* Configuration for V.17 9600bps mode */
    else if( !(mode & 4) )
     {
        ptT38Params->idRateRx = RATE_V179600S;
        ptT38Params->iPktdRateRx = V17_9600;
        ptT38Params->iIndPktdRateRx = V179600STTRN ;
        ptT38Params->iMinReqdZeroesInTCF = V1796MaxBytes * ZEROESFOR1_35SECS ; 
        ptT38Params->iMinImageLen = 120;
     }
/* Configuration for V.17 7200bps mode */
    else 
     {
        ptT38Params->idRateRx = RATE_V177200S;
        ptT38Params->iPktdRateRx = V17_7200;
        ptT38Params->iIndPktdRateRx = V177200STTRN ;
        ptT38Params->iMinReqdZeroesInTCF = V1772MaxBytes * ZEROESFOR1_35SECS ; 
        ptT38Params->iMinImageLen = 90;
     }
    return;
}


/*-------------------------------------------------------------------------

Function Name: ProcDCSFrV29Rx 

Description : Initializing the variables for demodulation with V29 info. 

Input: 
ptT38Params = relay channel pointer
mode : information field in DCS frame

Return Value:
None
---------------------------------------------------------------------------*/

void ProcDCSFrV29Rx(tFAXRELAY_ESL_T38 *ptT38Params,INT16 mode){

/* Configuration for V.29 9600bps mode */
    if( !(mode & 4) )
     {
        printf("\r\nV.29 9.6 KBPS Rx Mode\n");
        ptT38Params->idRateRx = RATE_V299600;
        ptT38Params->iPktdRateRx = V29_9600;
        ptT38Params->iIndPktdRateRx = V299600TRN ;
        ptT38Params->iMinReqdZeroesInTCF = V2996MaxBytes * ZEROESFOR1_35SECS ; 
        ptT38Params->iMinImageLen = 120;
     }

/* Configuration for V.29 7200bps mode */
    else
     {
        printf("\r\nV.29 7.2 KBPS Rx Mode\n");
        ptT38Params->idRateRx = RATE_V297200;
        ptT38Params->iPktdRateRx = V29_7200;
        ptT38Params->iIndPktdRateRx = V297200TRN ;
        ptT38Params->iMinReqdZeroesInTCF = V2972MaxBytes * ZEROESFOR1_35SECS ; 
        ptT38Params->iMinImageLen = 90;
     }

    return;
}

/*-------------------------------------------------------------------------

Function Name: ProcEndHdlcHSDataRx

Description : 
    State machine is set to wait for no carrier message from DP,
    after receiving which, the relay forms the SIG-END indicator packet
    and sets the DP in default mode of low speed reception.

Input: 
ptT38Params = relay channel pointer

Return Value:
None
---------------------------------------------------------------------------*/
#if 0	//dhruva
void ProcEndHdlcHSDataRx(tFAXRELAY_ESL_T38 *ptT38Params){

/* wait for no carrier message from DP */

    if(ptT38Params->iNoCarStat)
     {
        ptT38Params->iNumBytes = 0;

/* packetize SIG-END */
        T38PktT30Data(ptT38Params,(UCHAR8)ptT38Params->iPktdRateRx,HDLCSIGEND,1,0,ptT38Params->iCurrentRedundancy);
        ptT38Params->iNoCarStat = 0;
/*set DP in default mode of low speed reception.*/
        ptT38Params->fpStateFnPtr = (fpPtrIntFn)(SetDpModeHdlcLSRx);
        InitHdlcRx(ptT38Params);

     }

    return;
}
#endif


/*-------------------------------------------------------------------------

Function Name: ProcEndHdlcLSDataRx

Description : 
    State machine is set to wait for no carrier message from DP,
    after receiving which, the relay forms the SIG-END indicator packet.
    If the previous frame received by DP is DCS, the relay configures
    DP to receive training in high speed mode.
    Other wise the DP is configured in default mode of low speed reception.

Input: 
ptT38Params = relay channel pointer

Return Value:
None
---------------------------------------------------------------------------*/

void ProcEndHdlcLSDataRx(tFAXRELAY_ESL_T38 *ptT38Params){

/* wait for no carrier message from DP */
   if(ptT38Params->iNoCarStat)
   {
    printf("\r\n NO CARRIERRRRRRRRRRRRRRR port[%d]",ptT38Params->iPortNo);
    ptT38Params->iNumBytes = 0;

	T38PktT30Ind(ptT38Params,NOSIG);
	//BF 260804
	ptT38Params->iOpPktValid = 2;//dhruva
    
	ptT38Params->iNoCarStat = 0;

  /* Check for previous frame Received from the DP, and change
     the state machine accordingly */

    ptT38Params->iWaitForPreambleTxFlgTim = 0; //clear timer

    /* If DIS was received DCS is awaited */
    if( (ptT38Params->iPrevFrameRx & 0x7F) == DISID) 
    {
        printf("\r\n DISID,DCS awaited port[%d]",ptT38Params->iPortNo); 
        /*set DP in default mode of low speed reception.*/
        ConfigDpInHdlcModeRx(ptT38Params,RATE_V21300);
        InitHdlcRx(ptT38Params);
        /* state machine is set to wait for DCS frame */
        ptT38Params->iSpoofStartTimer = 270;
        if(ptT38Params->iFrameLength == FL_5MS)      //Frame length = 5ms
        {
            ptT38Params->iSpoofStartTimer = 540;
        }
        if(ptT38Params->iFrameLength == FL_20MS)      //Frame length = 20ms
        {
            ptT38Params->iSpoofStartTimer = 135;
        }

        WaitForMCF_DCS(ptT38Params);
    }

    /* If CFR was received page is awaited */
    else if( (ptT38Params->iPrevFrameRx & 0x7F) == CFRID) 
    {
        printf("\r\n CFRID, page awaited port[%d]",ptT38Params->iPortNo);
        /* state machine is set to wait for Page indicator pkt */
        //ptT38Params->iSpoofStartTimer = 280;
        ptT38Params->iSpoofStartTimer = 200; 
        if(ptT38Params->iFrameLength == FL_5MS)      //Frame length = 5ms
        {
            ptT38Params->iSpoofStartTimer = 560;
        }
        if(ptT38Params->iFrameLength == FL_20MS)      //Frame length = 20ms
        {
            ptT38Params->iSpoofStartTimer = 140;
        }
        //printf("\r\nWaitToSpoofHSImage is called");
        WaitToSpoofHSImage(ptT38Params);
    }

    /* If MCF was received */
    else if( (ptT38Params->iPrevFrameRx & 0x7F) == MCFID) 
    {
    /* Ckeck the previous command pkt received */
    
        /* If it was MPS, page is awaited */
        if( (ptT38Params->iPrevFrameTx & 0x7F) == MPSID)
        {
           /* state machine is set to wait for Page indicator pkt */
           //For Minolta,Nominal,Mita Faxlab profiles where Image starts after
           //2.2 seconds
           ptT38Params->iSpoofStartTimer = 200;
           //ptT38Params->iSpoofStartTimer = 280;
           if(ptT38Params->iFrameLength == FL_5MS)      //Frame length = 5ms
           {
               ptT38Params->iSpoofStartTimer = 560;
           }
           if(ptT38Params->iFrameLength == FL_20MS)      //Frame length = 20ms
           {
               ptT38Params->iSpoofStartTimer = 140;
           }
           WaitToSpoofHSImage(ptT38Params);
        }

        /* If it was EOP, DCN is awaited */
        else if((ptT38Params->iPrevFrameTx & 0x7F) == EOPID)
        {
           /*set DP in default mode of low speed reception.*/
           printf("\r\n EOPID, DCN awaited Port[%d]",ptT38Params->iPortNo); 
           ConfigDpInHdlcModeRx(ptT38Params,RATE_V21300);
           InitHdlcRx(ptT38Params);
           ptT38Params->piFrameDataPtr = ptT38Params->piReadDataPtr = ptT38Params->piDataBufPtr;
        /* state machine is set to wait for spoofing preamble */
           ptT38Params->iSpoofStartTimer = 250;
           if(ptT38Params->iFrameLength == FL_5MS)      //Frame length = 5ms
           {
               ptT38Params->iSpoofStartTimer = 500;
           }
           if(ptT38Params->iFrameLength == FL_20MS)      //Frame length = 20ms
           {
               ptT38Params->iSpoofStartTimer = 125;
           }
           WaitForSpoofPreamble(ptT38Params);
        }
    }

    /* If MPS or EOP was received, MCF is awaited */
    else if((ptT38Params->iPrevFrameRx & 0x7F) == MPSID || (ptT38Params->iPrevFrameRx & 0x7F) == EOPID)
    {
        printf("\r\n ProcEndHdlcLSDataRx : MPSID port[%d] PageCount[%d]",ptT38Params->iPortNo,++PageCount);
        /*set DP in default mode of low speed reception.*/
        ConfigDpInHdlcModeRx(ptT38Params,RATE_V21300);
        InitHdlcRx(ptT38Params);
        /* state machine is set to wait for MCF frame */
        ptT38Params->iSpoofStartTimer = 310; //270;
        //ptT38Params->iSpoofStartTimer = 540;
        if(ptT38Params->iFrameLength == FL_5MS)      //Frame length = 5ms
        {
            ptT38Params->iSpoofStartTimer = 540;
        }
        if(ptT38Params->iFrameLength == FL_20MS)      //Frame length = 20ms
        {
            ptT38Params->iSpoofStartTimer = 135;
        }
#ifdef T30_T38
		ptT38Params->piFrameDataPtr = ptT38Params->piDataBufPtr; 
		ptT38Params->piReadDataPtr = ptT38Params->piDataBufPtr;
#endif

        WaitForMCF_DCS(ptT38Params);
    }

    /* If it is DCS configure DP for High speed Rx with long training*/
    else if((ptT38Params->iPrevFrameRx & 0x7f) == DCSID)
    {

        /*Configure DP to receive training in high speed mode.*/
        if(ptT38Params->iV17ModeRx)
        {
           printf("\r\n RecvTCF set port[%d]",ptT38Params->iPortNo);
           ConfigDpInHSLTrnModeRx(ptT38Params);    
           ptT38Params->fpStateFnPtr = (fpPtrIntFn)(RecvLTCF);
        }
        else    
        {
           printf("\r\n RecvTCF set port[%d]",ptT38Params->iPortNo);
           ConfigDpInHSSTrnModeRx(ptT38Params);
           ptT38Params->fpStateFnPtr = (fpPtrIntFn)(RecvTCF);
        }
     }

    else
    {
        /*set DP in default mode of low speed reception.*/
        ConfigDpInHdlcModeRx(ptT38Params,RATE_V21300);
        InitHdlcRx(ptT38Params);
        ptT38Params->piFrameDataPtr = ptT38Params->piReadDataPtr = ptT38Params->piDataBufPtr;
        /* state machine is set to wait for spoofing preamble */
        ptT38Params->iSpoofStartTimer = 250;
        if(ptT38Params->iFrameLength == FL_5MS)      //Frame length = 5ms
        {
            ptT38Params->iSpoofStartTimer = 500;
        }
        if(ptT38Params->iFrameLength == FL_20MS)      //Frame length = 20ms
        {
            ptT38Params->iSpoofStartTimer = 125;
        }
        printf("\r\n ProcEndHdlcLSDataRx : WaitForSpoofPreamble");
        WaitForSpoofPreamble(ptT38Params);
    }

   }

   return;
}

/*-------------------------------------------------------------------------

Function Name: WaitForSpoofPreamble

Description : 
    State machine is set wait for the spoofing timer to elapse, after which
 it is reset to spoof preamble
 If any valid HDLC frames are received before the timer elapses, the DP mode
 and state machine are set appropriately.

Input: 
ptT38Params = relay channel pointer

Return Value:
None
---------------------------------------------------------------------------*/

void WaitForSpoofPreamble(tFAXRELAY_ESL_T38 *ptT38Params){

    ptT38Params->fpStateFnPtr = (fpPtrIntFn)(WaitForSpoofPreamble);

/* If training is successful or if message conformation is received for MPS 
then DP is configured for high speed transmision mode */
    if(   ((ptT38Params->iPrevFrameRx & 0x7f) == CFRID) || 
    (((ptT38Params->iPrevFrameTx & 0x7f) == MPSID) && 
    ((ptT38Params->iPrevFrameRx&0x7f) == MCFID)))
    {

        if(ptT38Params->iTrnTxFlg)
        { 
            //Reset the paramters related to TCF
            ResetTCFParams(ptT38Params);
            ptT38Params->iImageDataTx = 1;
            SetDpModeHSTx(ptT38Params);
            ptT38Params->iConnected = 0;
        }

    }

    else if( (ptT38Params->iPrevFrameRx &0x7f) !=FTTID)
    {

        if(ptT38Params->iPreambleTxFlg)
        {
            ptT38Params->iPreambleTxFlg = 0;
            ConfigDpInHdlcModeTx(ptT38Params,RATE_V21300);
            TxPreamble(ptT38Params);
        }
        /* wait for the spoofing timer to elapse */
        else if(ptT38Params->iWaitForPreambleTxFlgTim == 
                    ptT38Params->iSpoofStartTimer)
        {
        /* reset state machine to spoof preamble and configure DP to low speed
        transmision to send flags*/
            ptT38Params->iWaitForPreambleTxFlgTim = 0;
            ConfigDpInHdlcModeTx(ptT38Params,RATE_V21300);
            ptT38Params->iPreambleTim=100;
            if(ptT38Params->iFrameLength == FL_5MS)      //Frame length = 5ms
            {
                ptT38Params->iPreambleTim=200;
            }
            if(ptT38Params->iFrameLength == FL_20MS)      //Frame length = 20ms
            {
                ptT38Params->iPreambleTim=50;
            }
            ptT38Params->fpStateFnPtr = (fpPtrIntFn)(SpoofPreamble);
        }
        else
        {
            ptT38Params->iWaitForPreambleTxFlgTim++; 
        }
    }

/* If traning has failed DP is set to low speed transmision to send flags */
    else if( (ptT38Params->iPrevFrameRx & 0x7F) == FTTID)
    {

        if(ptT38Params->iPreambleTxFlg)
        {
            ptT38Params->iPreambleTxFlg = 0;
            ConfigDpInHdlcModeTx(ptT38Params,RATE_V21300);
            TxPreamble(ptT38Params);
        }
        else if(ptT38Params->iWaitForPreambleTxFlgTim == 
                        ptT38Params->iSpoofStartTimer)
        {
           printf("\r\n WaitForSpoofPreamble *******");
            ptT38Params->iWaitForPreambleTxFlgTim = 0;
            WaitForSettingMode(ptT38Params);
        }
        else
        {
            ptT38Params->iWaitForPreambleTxFlgTim++; 
        }
    }
    return;
}

/*-------------------------------------------------------------------------

Function Name: SpoofPreamble

Description : 
    DP is set to spoof premable flags to the fax terminal.
Input: 
ptT38Params = relay channel pointer

Return Value:
None
---------------------------------------------------------------------------*/


void SpoofPreamble(tFAXRELAY_ESL_T38 *ptT38Params)
{

    INT16 Timer = 250;                                //For 10ms frame length

    ptT38Params->fpStateFnPtr = (fpPtrIntFn)(SpoofPreamble);

    if(ptT38Params->iFrameLength == FL_5MS)      //Frame length = 5ms
    {
        Timer = 500;
    }
    if(ptT38Params->iFrameLength == FL_20MS)      //Frame length = 20ms
    {
        Timer = 125;
    }

 /* if the spoofing timer elapses */
    if(ptT38Params->iSpoofPreambleTim == Timer)
    {
        /* If EOP was recived, spoof DCN ( disconect)*/
        if((ptT38Params->iPrevFrameTx & 0x7F) == EOPID)
        {
#if 1 
           //Skip sending CRP to DP during MPS/MCF/EOP collision
           // or IGN CMD case
           if(SendCRPtoDP[ptT38Params->iPortNo] == TRUE)
           {
#endif
            SendDCN(ptT38Params);
#if 1 
           }
           else
           {
              SendCRPtoDP[ptT38Params->iPortNo] = TRUE;
              ptT38Params->iPrevCRCStatus = 0 ;
              InitHdlcTxVar(ptT38Params);
              ptT38Params->iConnected = 0;
              ptT38Params->fpStateFnPtr = (fpPtrIntFn)(ChkHdlcDataAvl);
           }
#endif

        }
//If previous received frame was RTN/RTP send flags for an additional 0.5s
 //Changed to avoid sending CRP to receiving fax machine after EoP
 // if((ptT38Params->iPrevFrameRx & 0x7F) == RTNID ||
        else if((ptT38Params->iPrevFrameRx & 0x7F) == RTNID ||
            (ptT38Params->iPrevFrameRx & 0x7F) == RTPID)
        {
            ptT38Params->iSpoofPreambleTim -= 50;
            if(ptT38Params->iFrameLength == FL_5MS)      //Frame length = 5ms
                ptT38Params->iSpoofPreambleTim -= 100;
            if(ptT38Params->iFrameLength == FL_20MS)      //Frame length = 20ms
                ptT38Params->iSpoofPreambleTim -= 25;
        }

        /* In any other case send command repeat frame */
        else
        {
#if 1
           //Skip sending CRP to DP during MPS/MCF/EOP collision
           // or IGN CMD case
           if(SendCRPtoDP[ptT38Params->iPortNo] == TRUE)
           {
#endif
            SendCRP(ptT38Params);
#if 1 
           }
           else
           {
              SendCRPtoDP[ptT38Params->iPortNo] = TRUE;
              ptT38Params->iPrevCRCStatus = 0 ;
              InitHdlcTxVar(ptT38Params);
              ptT38Params->iConnected = 0;
              ptT38Params->fpStateFnPtr = (fpPtrIntFn)(ChkHdlcDataAvl);
           }
#endif
        }
    }
    else
    {
        ptT38Params->iSpoofPreambleTim++;

/* If the preamble indicator packet is received before the timer is elapsed,
   state machine is configured to received HDLC data from IP network */
        if(ptT38Params->iPreambleTxFlg)
        {
            if(ptT38Params->iSpoofPreambleTim >=ptT38Params->iPreambleTim)
                ChkHdlcDataAvl(ptT38Params);
            else
            {
                ptT38Params->iPreambleTim -= ptT38Params->iSpoofPreambleTim;
                TxPreamble(ptT38Params);
            }
            ptT38Params->iSpoofPreambleTim = 0;
            ptT38Params->iPreambleTxFlg = 0;
        }
    }

    if( (ptT38Params->iPrevFrameRx & 0x7F) == DCNID) 
    {
        ptT38Params->iShutDown = 1;
    }

    return;
}

/*-------------------------------------------------------------------------

Function Name: ByteReverse

Description : 
    This function is used to reverse the given byte.


Input:     Source byte
    

Return Value:
    Reversed byte

---------------------------------------------------------------------------*/

UCHAR8 ByteReverse(UCHAR8 byte)
{
    INT16 i;
    UCHAR8 revByte = 0;

    for(i=0;i<8;i++)
     {
        revByte = (revByte << 1) | (byte & 1);    
        byte = byte >> 1 ;
     }    
    return(revByte);
}

/*-------------------------------------------------------------------------

Function Name: ProcNotLastLSFrameRx

Description : 
    If the previous HDLC frame received by the DP is not a final frame,
    state machine is set to pack the frames which follow.

Input: 
ptT38Params = relay channel pointer

Return Value:
None
---------------------------------------------------------------------------*/

void ProcNotLastLSFrameRx(tFAXRELAY_ESL_T38 *ptT38Params){
    InitHdlcRx(ptT38Params);
//D    ptT38Params->piFrameDataPtr = ptT38Params->piReadDataPtr = ptT38Params->ps_DataBufPtr;
//dhruva
    ptT38Params->iResendPktCnt = 0;
    ptT38Params->fpStateFnPtr = (fpPtrIntFn)(PackHdlcDataPkt);    
    return;
}


/*-------------------------------------------------------------------------

Function Name: WaitToSpoofHSImage

Description : 
    State machine is set to wait for image data (Trn Indicator pkt)
    till the timer elapses. After which, high speed fill bits (zeroes) 
    are spoofed.

Input:
    ptT38Params = relay channel pointer


Return Value: 
None

---------------------------------------------------------------------------*/

void WaitToSpoofHSImage(tFAXRELAY_ESL_T38 *ptT38Params)
{
   ptT38Params->fpStateFnPtr = (fpPtrIntFn)(WaitToSpoofHSImage);
#if 0
   if(HSspoofCntrFlg == 1)
       HSspoofCntr++;
#endif

/* Got high speed training indicator packet ? */
   if(ptT38Params->iTrnTxFlg)
   { 
        //printf("\r\n WaitToSpoofHSImage :HSspoofCntr[%d]",HSspoofCntr);
        //HSspoofCntrFlg = HSspoofCntr = 0;
        //Reset the paramters related to TCF
        ResetTCFParams(ptT38Params);
        ptT38Params->iImageDataTx = 1;  //enable high speed image transmision
        SetDpModeHSTx(ptT38Params);      //Configure DP to transmit in high
        ptT38Params->iConnected = 0;    //speed mode
	ptT38Params->iHdlcDataAvl = 0;//dhruva
   }

   else if(ptT38Params->iWaitForPreambleTxFlgTim == 
                ptT38Params->iSpoofStartTimer)
   {
/*
// reset state machine to spoof HS image data. Set DP to high speed 
//transmision to send fill bits zeroes 

        ptT38Params->iWaitForPreambleTxFlgTim = 0;
//will be set when trn indicator pkt is received
                                           //is received
        ptT38Params->iImageDataTx = 0;
        SetDpModeHSTx(ptT38Params);
*/

		if(!ptT38Params->iDroppedPkt) {
                printf("\r\n WaitToSpoofHSImage : Spoofing HS Image");
#if 0
                HSspoofCntrFlg = 1;
#endif
// reset state machine to spoof HS image data. Set DP to high speed 
//transmision to send fill bits zeroes 
	        ptT38Params->iWaitForPreambleTxFlgTim = 0;

//BF 260804
/////////dhruva 140203
			//Reset the paramters related to TCF
			ResetTCFParams(ptT38Params);
			ptT38Params->iConnected = 0;    //speed mode
			ptT38Params->iHdlcDataAvl = 0;//dhruva
			ptT38Params->iPreambleTxFlg = 0;
////////////

//will be set when trn indicator pkt is received
		    ptT38Params->iImageDataTx = 0;
			SetDpModeHSTx(ptT38Params);

                    //for debugging
                    debug_fax = 400;
		}
		else {
	        ptT38Params->fpStateFnPtr = (fpPtrIntFn)(WaitForSettingMode);
			ptT38Params->iDroppedPkt = 0;
		}
   
   }
   else
   {
           ptT38Params->iWaitForPreambleTxFlgTim++; 
           //printf("\r\n Wait for preamble tx flg");
   }
   return;
}



/*-------------------------------------------------------------------------

Function Name: WaitForMCF_DCS

Description : 
    State machine is set to wait for MCF or DCS data till the timer    elapses
    If frames are not received, no carrier message waited from DP.

Input:
    ptT38Params = relay channel pointer


Return Value: 
None

---------------------------------------------------------------------------*/

void WaitForMCF_DCS(tFAXRELAY_ESL_T38 *ptT38Params)
{
   ptT38Params->fpStateFnPtr = (fpPtrIntFn)(WaitForMCF_DCS);

/* Set the DP to transmit HDLC frmaes in V21 mode, on reception of preamble
   indiator packet */
   if(ptT38Params->iPreambleTxFlg)
   {
        ptT38Params->iPreambleTxFlg = 0;
        ConfigDpInHdlcModeTx(ptT38Params,RATE_V21300);

        //BF 260804
        if(!ptT38Params->iHdlcDataAvl)	//dhurva 280704 renesas
        {
        ptT38Params->piFrameDataPtr = ptT38Params->piDataBufPtr;
        ptT38Params->piReadDataPtr = ptT38Params->piDataBufPtr;
        }


        TxPreamble(ptT38Params);
   }
/* wait for the spoofing timer to elapse */
   else if(ptT38Params->iWaitForPreambleTxFlgTim == ptT38Params->iSpoofStartTimer)
   {
/* reset state machine to spoof preamble and configure DP to low speed 
   reception to detect signal drop */
        ConfigDpInHdlcModeRx(ptT38Params,RATE_V21300);
        InitHdlcVarRx(ptT38Params);
        ptT38Params->iNoCarCounter = 1;
        /* Reset data buffer pointers */
        ptT38Params->piFrameDataPtr = ptT38Params->piDataBufPtr;
        ptT38Params->piReadDataPtr = ptT38Params->piDataBufPtr;
        ptT38Params->fpStateFnPtr = (fpPtrIntFn)(WaitForNoCarMsg);
#ifdef T30_T38
        ptT38Params->fpStateFnPtr = (fpPtrIntFn)(WaitForSettingMode);
#endif

#if 1 //added on 270804
        if(((ptT38Params->iPrevFrameRx & 0x7F) == MPSID) || ((ptT38Params->iPrevFrameRx & 0x7F) == EOPID))
        {
           //Buffer the MCF/EOP message during the MPS/EOP/MCF collision
           //for Toshiba TF-231 profile
           ptT38Params->fpStateFnPtr = (fpPtrIntFn)(WaitForNoCarMsg);
        }
        else
#endif
        {
           //Waiting for CSI/DIS
           //for profiles ignoring the first CSI/DIS message
#if 1
//Dhruva 130704
        ptT38Params->fpStateFnPtr = (fpPtrIntFn)(WaitForSettingMode);
        printf("\r\n WaitForMCF_DCS: Next Function WaitForSettingMode");
#endif
        }

   }
   else
   {
        ptT38Params->iWaitForPreambleTxFlgTim++; 
   }

   return;
}



/*-------------------------------------------------------------------------

Function Name: WaitForNoCarMsg

Description : 
    State machine is set to wait for no carrier message from DP
    After getting which, the frame queued by T38 is sent to DP for
    transmission

Input:
    ptT38Params = relay channel pointer

Return Value: 
None

---------------------------------------------------------------------------*/

void WaitForNoCarMsg(tFAXRELAY_ESL_T38 *ptT38Params)
{
   /* Disable the data sent by DP to be written into data buffer, since
   it has to be ignored */
   ptT38Params->iDisableData = 1;
   ptT38Params->fpStateFnPtr = (fpPtrIntFn)(WaitForNoCarMsg);
   //printf("\r\nWaitForNoCarMsg \r\n");
   if(ptT38Params->iNoCarStat)
   {
	   printf("\r\n WaitForNoCarMsg : iNoCarStat set");
	   ptT38Params->iNoCarStat = 0;

/* If DCS and TCF are repeated by the fax terminal wait for both the sequences
   to end */
		if(ptT38Params->iNoCarCounter == 2)
		{
	                printf("\r\n WaitForNoCarMsg : NoCarCounter(2) now (3)");
			ptT38Params->iNoCarCounter = 3;


#ifndef T30_T38
			SendTxRxMsgToDp(ptT38Params,mDP_DATAMODE_SELECT,DATA_M2L);
			if(ptT38Params->iV17ModeRx)
			{
		/* Message to configure the DP in fax reception mode at the given rate 
		   in Long train mode */
				SendTxRxMsgToDp(ptT38Params,mDP_START_FAXRX,ptT38Params->idRateRx+1);
			}
			else    
			{
		/* Message to configure the DP in fax reception mode at the given rate 
		   in Short train mode */
				SendTxRxMsgToDp(ptT38Params,mDP_START_FAXRX,ptT38Params->idRateRx);
			}

#endif

			ptT38Params->fpStateFnPtr = (fpPtrIntFn)(WaitForNoCarMsg);
//////////dhruva
    ptT38Params->iTrnRxFlg = 0;
    //ptT38Params->iNoCarStat = 0;
    ptT38Params->iNumBytes = 0;
    ptT38Params->iPackedBytes = 0;
    ptT38Params->iByteCnt = 0;
    ptT38Params->iZeroesInTCF = 0;

////

			return;
		}

/* Set the DP to transmit HDLC frmaes in V21 mode, on reception of preamble
   indiator packet */
		if(ptT38Params->iPreambleTxFlg)
		{
                   printf("\r\n WaitForNoCarMsg : V21 Preamble Ind Rcvd");
        /* Enable the data sent by DP to be written into data buffer */
			ptT38Params->iDisableData = 0;
			ptT38Params->iPreambleTxFlg = 0;
	                //ptT38Params->iPreambleTim = 5;	
	                ptT38Params->iPreambleTim = 85; 
			ConfigDpInHdlcModeTx(ptT38Params,RATE_V21300);
			TxPreamble(ptT38Params);
		}

/* If the awaited ( preamble indicator)Packet is not recived from the 
   IP network, set the state machine to wait for another No carrier message */
		else
		{
                    printf("\r\nWaitForNoCarMsg:WaitForAnother NoCar Fptr->WaitForPreamble");

#if 0 //dhruva, changed on 09/jan/03
#ifndef T30_T38
			ConfigDpInHdlcModeRx(ptT38Params,RATE_V21300);
		    InitHdlcVarRx(ptT38Params);
#endif
			if(ptT38Params->iNoCarCounter == 3)
			{
	            ptT38Params->iNoCarCounter = 2;
		    }
			else
			{
	            ptT38Params->iNoCarCounter = 1;
			}
        /* Reset data buffer pointers */
	        ptT38Params->piFrameDataPtr = ptT38Params->piDataBufPtr;
		    ptT38Params->piReadDataPtr = ptT38Params->piDataBufPtr;
			ptT38Params->fpStateFnPtr = (fpPtrIntFn)(WaitForNoCarMsg);
#endif//dhruva, changed on 09/jan/03

			ptT38Params->iSpoofStartTimer = 270;
			if(ptT38Params->iFrameLength == FL_5MS)      //Frame length = 5ms
			{
				ptT38Params->iSpoofStartTimer = 540;
			}
			if(ptT38Params->iFrameLength == FL_20MS)      //Frame length = 20ms
			{
				ptT38Params->iSpoofStartTimer = 135;
			}
			ptT38Params->iWaitForPreambleTxFlgTim = 0;
			ptT38Params->fpStateFnPtr = (fpPtrIntFn)(WaitForPreamble);
		}
   }
   return;
}

void WaitForPreamble(tFAXRELAY_ESL_T38 *ptT38Params)
{
/* Set the DP to transmit HDLC frmaes in V21 mode, on reception of preamble
   indiator packet */
	if(ptT38Params->iPreambleTxFlg)
	{
    /* Enable the data sent by DP to be written into data buffer */
		ptT38Params->iDisableData = 0;
		ptT38Params->iPreambleTxFlg = 0;
	    //ptT38Params->iPreambleTim = 5;		//dhruva
	    //BF 260804
	    ptT38Params->iPreambleTim = 100;		//dhruva
		ConfigDpInHdlcModeTx(ptT38Params,RATE_V21300);
		TxPreamble(ptT38Params);
	}

/* wait for the spoofing timer to elapse */
   else if(ptT38Params->iWaitForPreambleTxFlgTim == ptT38Params->iSpoofStartTimer)
   {
#ifndef T30_T38
		ConfigDpInHdlcModeRx(ptT38Params,RATE_V21300);
		InitHdlcVarRx(ptT38Params);
#endif
		if(ptT38Params->iNoCarCounter == 3)
		{
			ptT38Params->iNoCarCounter = 2;
		}
		else
		{
			ptT38Params->iNoCarCounter = 1;
		}
	/* Reset data buffer pointers */
		ptT38Params->piFrameDataPtr = ptT38Params->piDataBufPtr;
		ptT38Params->piReadDataPtr = ptT38Params->piDataBufPtr;

		ptT38Params->fpStateFnPtr = (fpPtrIntFn)(WaitForNoCarMsg);

#ifdef T30_T38
//dhruva, added on 13/01/03
		ptT38Params->fpStateFnPtr = (fpPtrIntFn)(WaitForSettingMode);
	   ptT38Params->iDisableData = 0;

#endif
   
   }
   else
        ptT38Params->iWaitForPreambleTxFlgTim++; 
}

/*-------------------------------------------------------------------------

Function Name: WaitForCFR_FTT

Description : 
    State machine is set to wait for response for training (CFR or FTT)
    If response is not obtained before the timer elapses, set the state
    machine to wait for no carrier message from DP, then the queued
    frame (CFR or FTT) is sent to DP.


Input:
    ptT38Params = relay channel pointer


Return Value: 
None

---------------------------------------------------------------------------*/

void WaitForCFR_FTT(tFAXRELAY_ESL_T38 *ptT38Params)
{
   ptT38Params->fpStateFnPtr = (fpPtrIntFn)(WaitForCFR_FTT);

/* Set the DP to transmit HDLC frmaes in V21 mode, on reception of preamble
   indiator packet */
   if(ptT38Params->iPreambleTxFlg)
   {
    printf("\r\n WaitForCFR_FTT: Preamble Rcvd port[%d]",ptT38Params->iPortNo);
    ptT38Params->iPreambleTxFlg = 0;
    ConfigDpInHdlcModeTx(ptT38Params,RATE_V21300);
    /* Reset data buffer pointers */
    ptT38Params->piFrameDataPtr = ptT38Params->piDataBufPtr;
    ptT38Params->piReadDataPtr = ptT38Params->piDataBufPtr;


#ifdef T30_T38
	ptT38Params->iPreambleTim = 1;
#endif

    TxPreamble(ptT38Params);
   }
/* wait for the spoofing timer to elapse */
   else if(ptT38Params->iWaitForPreambleTxFlgTim == ptT38Params->iSpoofStartTimer)
   {
/* reset state machine to spoof preamble and configure DP to low speed 
   reception to detect signal drop */

    printf("\r\nWaitForCFR_FTT: Spoof Timer Elapsed\n");
    ConfigDpInHdlcModeRx(ptT38Params,RATE_V21300);
    InitHdlcVarRx(ptT38Params);
// The counter is set to 2 to incate that DCS and TCF frames are to be ignored
    ptT38Params->iNoCarCounter = 2;
    ptT38Params->piFrameDataPtr = ptT38Params->piDataBufPtr;
    ptT38Params->piReadDataPtr = ptT38Params->piDataBufPtr;

	ptT38Params->fpStateFnPtr = (fpPtrIntFn)(WaitForNoCarMsg);
//	ptT38Params->fpStateFnPtr = (fpPtrIntFn)(WaitForEndOfTCF);
   }
   else
   {
        ptT38Params->iWaitForPreambleTxFlgTim++; 
   }

    return;
}


/*-------------------------------------------------------------------------

Function Name: SpoofTCF

Description : 
    State machine is set to spoof TCF data, data pump is set to transmit
    in high speed mode.

Input:
    ptT38Params = relay channel pointer


Return Value: 
None

---------------------------------------------------------------------------*/

void SpoofTCF(tFAXRELAY_ESL_T38 *ptT38Params)
{
    if( ((ptT38Params->iPrevFrameTx & 0x7F) == DCSID)
//BF 260804 dhruva 140203       && ((ptT38Params->iPrevFrameRx & 0x7F) != CFRID)
       )
    {
        ptT38Params->iImageDataTx = 0;    //Since the high speed data is TCF
                                            //and not image
//set up DP to transmit in high speed mode
        SetDpModeHSTx(ptT38Params);
        ptT38Params->iConnected = 0;
// Set TCF flag, used to validate the TCF packets received from the IP network
        ptT38Params->iTCFdata = 1;
    }
    return;
}


/*-------------------------------------------------------------------------

Function Name: SendFTTPkt

Description : 
    FTT ( Failure To Train) frame is packetized and sent to the network

Input:
    ptT38Params = relay channel pointer


Return Value: 
None

---------------------------------------------------------------------------*/

void SendFTTPkt(tFAXRELAY_ESL_T38 *ptT38Params)
{
    ptT38Params->iCRCStatusRx = 1;        //Set CRC information
    ptT38Params->iByteCnt = 5;            //number of bytes in FTT frame

/* Reset pointers to data buffer */
    ptT38Params->piFrameDataPtr = ptT38Params->piDataBufPtr;
    ptT38Params->piReadDataPtr = ptT38Params->piDataBufPtr;

    *ptT38Params->piFrameDataPtr++ = 0xff;    //adress filed
    *ptT38Params->piFrameDataPtr++ = 0xc8;    //control field
    *ptT38Params->piFrameDataPtr++ = 0x22;    //information feild
    if(ptT38Params->piFrameDataPtr == (ptT38Params->piDataBufPtr + ptT38Params->iSizeOfDataBuf))
        ptT38Params->piFrameDataPtr = ptT38Params->piDataBufPtr ;

/* State machine is set to packetize the above frmae and sent to the 
   IP network */
//dhruva
    ptT38Params->iResendPktCnt = 0;
    ptT38Params->fpStateFnPtr = (fpPtrIntFn)(PackHdlcDataPkt);
   return;
}



/*-------------------------------------------------------------------------

Function Name: SendDCN

Description :
    This function is used to spoof the HDLC frame "DCN"
 The bytes in DCN frame are sent as messages to DP which inturn applies the
 low speed modulation technique (V.21) and send to the fax terminal.

Input: 
    ptT38Params = relay channel pointer

Return Value:
None

---------------------------------------------------------------------------*/

void SendDCN(tFAXRELAY_ESL_T38 *ptT38Params)
{
    ptT38Params->tMessage.aiParams[0] = (UCHAR8)(0xff);    //Address field
    ptT38Params->tMessage.aiParams[1] = (UCHAR8)(0x13);    //Control field
    ptT38Params->tMessage.aiParams[2] = (UCHAR8)(0xfb);    //information field
    ptT38Params->tMessage.iLen = 3;                //length of the message

//Message ID mT38MSGTODP indicates host to forward this message to DP
    ptT38Params->tMessage.iId = mDP_MOD_DATA;

    SendMsgToHost(ptT38Params);
/*Also a message indicating that the above HDLC frame is not in error is sent*/
    ptT38Params->tMessage.iLen = 0;
//Message ID mT38MSGTODP indicates host to forward this message to DP
    ptT38Params->tMessage.iId = mDP_FCSOK;

    SendMsgToHost(ptT38Params);

/* Function call to set the state machine for further operation od the relay */
    ptT38Params->iPrevCRCStatus = 3;    //CRP is a final frame
    ProcEndHdlcDataTx(ptT38Params);
    ptT38Params->iPrevCRCStatus = 0;
    return;
}

/*-------------------------------------------------------------------------

Function Name: ResetTCFParams

Description :
        To reset the parameters used in TCF phase.

Input: 
    ptT38Params = relay channel pointer

Return Value:
None

---------------------------------------------------------------------------*/

void ResetTCFParams(tFAXRELAY_ESL_T38 *ptT38Params)
{
    ptT38Params->iZeroesInTCF = 0;
    ptT38Params->iMinReqdZeroesInTCF = 0;
    ptT38Params->iLocalRespOfTCF = 0;
    ptT38Params->iReqdZeroesInTCFForTx = 0;
    ptT38Params->iZeroesInTCFTx = 0;
    ptT38Params->iTCFdata = 0;
}

void CopyChars(UCHAR8 *DestPtr,UCHAR8 *SrcPtr,UINT16 Len)
{
    UINT16 i;
    for(i=0;i<Len;i++) {
        DestPtr[i] = SrcPtr[i];
    }
}

void ClearCRCBuf(tFAXRELAY_ESL_T38 *ptT38Params)
{
    UINT16 i;

    for(i=0;i<CRCBUFSIZE;i++)
        ptT38Params->aiCRCBuf[i] = 0;
    ptT38Params->iCRCRIndex = ptT38Params->iCRCWIndex = 0;
    //printf("\r\n ClearCRCBuf port[%d]",ptT38Params->iPortNo);
}

void ClearSizeBuf(tFAXRELAY_ESL_T38 *ptT38Params)
{
    UINT16 i;
    for(i=0;i<SIZEBUFSIZE;i++)
        ptT38Params->aiSizeBuf[i] = 0;
    ptT38Params->iSizeRIndex = ptT38Params->iSizeWIndex = 0;
}


///////////////////////
/*
void WaitForEndOfTCF(tFAXRELAY_ESL_T38 *ptT38Params)
{
   ptT38Params->iDisableData = 1;
   ptT38Params->fpStateFnPtr = (fpPtrIntFn)(WaitForEndOfTCF);

// Set the DP to transmit HDLC frmaes in V21 mode, on reception of preamble
//   indiator packet
   if(ptT38Params->iPreambleTxFlg && TCFTxEnd)
   {
	   TCFTxEnd = 0;
    // Enable the data sent by DP to be written into data buffer
		ptT38Params->iDisableData = 0;
		ptT38Params->iPreambleTxFlg = 0;
	    ptT38Params->iPreambleTim = 5;		//dhruva
		ConfigDpInHdlcModeTx(ptT38Params,RATE_V21300);
		TxPreamble(ptT38Params);
   }

}
*/
////////////////


/////////////////////// dhruva 310804

void WaitToSendCRPPkt(tFAXRELAY_ESL_T38 *ptT38Params)
{
    UCHAR8 TypeOfMsg,status;

/* Set the DP to transmit HDLC frmaes in V21 mode, on reception of preamble
   indiator packet */
   if(ptT38Params->iPreambleRxFlg)
   {
		ptT38Params->IgnorePkts = 1;
		ptT38Params->iDisablePktData=1;
        ptT38Params->iPreambleRxFlg=0;
        T38PktT30Ind(ptT38Params,V21PREAMBLE);
		if(++ptT38Params->iCurrentRedundancy >= ptT38Params->iLSRedundancy)
			ptT38Params->iCurrentRedundancy = ptT38Params->iLSRedundancy;

        ptT38Params->iResendPktCnt = 2;
        ptT38Params->fpStateFnPtr =(fpPtrIntFn )(PackHdlcDataPkt);
   }
#if 1 //dhruva 010904
   else if (ptT38Params->iPreambleTxFlg)//Dhruva 010804
   {
        ptT38Params->iPreambleTxFlg = 0;
        ConfigDpInHdlcModeTx(ptT38Params,RATE_V21300);
        TxPreamble(ptT38Params);
   }
#endif
/* wait for the spoofing timer to elapse */
   else if(ptT38Params->iWaitForPreambleTxFlgTim == ptT38Params->iSpoofStartTimer)
   {
     //Dhruva 010904
        T38PktT30Ind(ptT38Params,V21PREAMBLE);
		if(++ptT38Params->iCurrentRedundancy >= ptT38Params->iLSRedundancy)
			ptT38Params->iCurrentRedundancy = ptT38Params->iLSRedundancy;

        ptT38Params->iResendPktCnt = 2;
        ptT38Params->fpStateFnPtr =(fpPtrIntFn )(SendCRPPkt);
   }
   else
   {
        ptT38Params->iWaitForPreambleTxFlgTim++; 
   }

}

/////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////Dhruva 010804

void SendCRPPkt(tFAXRELAY_ESL_T38 *ptT38Params)
{

    UCHAR8 TypeOfMsg,status;

    if(ptT38Params->iResendPktCnt > 0) {
		ptT38Params->iResendPktCnt--;
		T38PktT30Ind(ptT38Params,V21PREAMBLE);
		if(++ptT38Params->iCurrentRedundancy >= ptT38Params->iLSRedundancy)
			ptT38Params->iCurrentRedundancy = ptT38Params->iLSRedundancy;
		return;
	}

		ptT38Params->piFrameDataPtr = 
		ptT38Params->piReadDataPtr = ptT38Params->piDataBufPtr;

		ptT38Params->piFrameDataPtr[0] = 0xFF;
		ptT38Params->piFrameDataPtr[1] = 0xC8;
		ptT38Params->piFrameDataPtr[2] = 
	ptT38Params->iPrevFrameRx = 0xD8;
		ptT38Params->piFrameDataPtr += 3;
		
		TypeOfMsg = V21C2;
		status = HDLCFCSOKSIGEND;
		ptT38Params->iNumBytes = 3;
		T38PktT30Data(ptT38Params,TypeOfMsg, HDLCDATA, 2, status,ptT38Params->iCurrentRedundancy);

		if(++ptT38Params->iCurrentRedundancy >= ptT38Params->iLSRedundancy)
			ptT38Params->iCurrentRedundancy = ptT38Params->iLSRedundancy;

/* reset state machine to spoof preamble and configure DP to low speed
transmision to send flags*/
        ptT38Params->iWaitForPreambleTxFlgTim = 0;
        ConfigDpInHdlcModeTx(ptT38Params,RATE_V21300);
        ptT38Params->iPreambleTim=100;
        if(ptT38Params->iFrameLength == FL_5MS)      //Frame length = 5ms
        {
            ptT38Params->iPreambleTim=200;
        }
        if(ptT38Params->iFrameLength == FL_20MS)      //Frame length = 20ms
        {
            ptT38Params->iPreambleTim=50;
        }
    SendCRPtoDP[ptT38Params->iPortNo] = FALSE;
        ptT38Params->fpStateFnPtr = (fpPtrIntFn)(SpoofPreamble);

}


///////////////////////////////////////////////////////////////



/**************************************************/
#endif


/*--------------------------------------------------------------------
             END    OF   FILE
---------------------------------------------------------------------*/
